using System;

namespace NetTopologySuite.Geometries.Implementation
{
    /// <summary>
    /// Builds packed array coordinate sequences. The array data type can be either
    /// double or float, and defaults to double.
    /// </summary>
    [Serializable]
    public class PackedCoordinateSequenceFactory : CoordinateSequenceFactory
    {
        /// <summary>
        ///
        /// </summary>
        public enum PackedType
        {
            /// <summary>
            ///
            /// </summary>
            Double = 0,

            /// <summary>
            ///
            /// </summary>
            Float = 1,
        }

        public static readonly PackedCoordinateSequenceFactory DoubleFactory =
            new PackedCoordinateSequenceFactory(PackedType.Double);

        public static readonly PackedCoordinateSequenceFactory FloatFactory =
            new PackedCoordinateSequenceFactory(PackedType.Float);

        private PackedType type = PackedType.Double;

        /// <summary>
        /// Initializes a new instance of the <see cref="PackedCoordinateSequenceFactory"/> class,
        /// using double values.
        /// </summary>
        public PackedCoordinateSequenceFactory() : this(PackedType.Double) { }

        /// <summary>
        /// Initializes a new instance of the <see cref="PackedCoordinateSequenceFactory"/> class.
        /// </summary>
        /// <param name="type">The type.</param>
        public PackedCoordinateSequenceFactory(PackedType type)
        {
            Type = type;
        }

        /// <summary>
        ///
        /// </summary>
        public PackedType Type
        {
            get => type;
            set
            {
                if (value != PackedType.Double && value != PackedType.Float)
                    throw new ArgumentException("Unknown type " + value);
                this.type = value;
            }
        }

        /// <summary>
        /// Returns a CoordinateSequence based on the given array; whether or not the
        /// array is copied is implementation-dependent.
        /// </summary>
        /// <param name="coordinates">Coordinates array, which may not be null nor contain null elements</param>
        /// <returns></returns>
        public override CoordinateSequence Create(Coordinate[] coordinates)
        {
            int dimension = 2;
            int measures = 0;
            if (coordinates != null && coordinates.Length > 0 && coordinates[0] != null)
            {
                var first = coordinates[0];
                dimension = Coordinates.Dimension(first);
                measures = Coordinates.Measures(first);
            }

            if (type == PackedType.Double)
                return new PackedDoubleCoordinateSequence(coordinates, dimension, measures);
            return new PackedFloatCoordinateSequence(coordinates, dimension, measures);
        }

        /// <summary>
        /// Returns a CoordinateSequence based on the given coordinate sequence; whether or not the
        /// array is copied is implementation-dependent.
        /// </summary>
        /// <param name="coordSeq"></param>
        /// <returns></returns>
        public override CoordinateSequence Create(CoordinateSequence coordSeq)
        {
            int dimension = coordSeq.Dimension;
            int measures = coordSeq.Measures;
            if (type == PackedType.Double)
                return new PackedDoubleCoordinateSequence(coordSeq.ToCoordinateArray(), dimension, measures);
            return new PackedFloatCoordinateSequence(coordSeq.ToCoordinateArray(), dimension, measures);
        }

        /// <summary>
        /// Create a packed coordinate sequence from the provided array. 
        /// </summary>
        /// <param name="packedCoordinates"></param>
        /// <param name="dimension"></param>
        /// <returns>Packaged coordinate seqeunce of the requested type</returns>
        public CoordinateSequence Create(double[] packedCoordinates, int dimension)
        {
            return Create(packedCoordinates, dimension, 0);
        }

        /// <summary>
        /// Create a packed coordinate sequence from the provided array. 
        /// </summary>
        /// <param name="packedCoordinates"></param>
        /// <param name="dimension"></param>
        /// <param name="measures"></param>
        /// <returns>Packaged coordinate seqeunce of the requested type</returns>
        public CoordinateSequence Create(double[] packedCoordinates, int dimension, int measures)
        {
            if (type == PackedType.Double)
                return new PackedDoubleCoordinateSequence(packedCoordinates, dimension, measures);
            return new PackedFloatCoordinateSequence(packedCoordinates, dimension, measures);
        }

        /// <summary>
        /// Create a packed coordinate sequence from the provided array. 
        /// </summary>
        /// <param name="packedCoordinates"></param>
        /// <param name="dimension"></param>
        /// <returns>Packaged coordinate seqeunce of the requested type</returns>
        public CoordinateSequence Create(float[] packedCoordinates, int dimension)
        {
            return Create(packedCoordinates, dimension, 0);
        }

        /// <summary>
        ///
        /// </summary>
        /// <param name="packedCoordinates"></param>
        /// <param name="dimension"></param>
        /// <param name="measures"></param>
        /// <returns>Packaged coordinate seqeunce of the requested type</returns>
        public CoordinateSequence Create(float[] packedCoordinates, int dimension, int measures)
        {
            if (type == PackedType.Double)
                return new PackedDoubleCoordinateSequence(packedCoordinates, dimension, measures);
            else return new PackedFloatCoordinateSequence(packedCoordinates, dimension, measures);
        }

        /// <inheritdoc />
        public override CoordinateSequence Create(int size, int dimension, int measures)
        {
            if (type == PackedType.Double)
                return new PackedDoubleCoordinateSequence(size, dimension, measures);
            else return new PackedFloatCoordinateSequence(size, dimension, measures);
        }
    }
}
