﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using NetTopologySuite.Algorithm;
using NetTopologySuite.Geometries;
using NetTopologySuite.Geometries.Implementation;
using NetTopologySuite.Geometries.Utilities;
using NetTopologySuite.Index.KdTree;
using NetTopologySuite.IO;
using NetTopologySuite.Operation.Polygonize;
using NetTopologySuite.Operation.Valid;
using NetTopologySuite.Precision;
using NetTopologySuite.SnapRound;
using NUnit.Framework;

namespace NetTopologySuite.Samples.Tests.Github
{
    [TestFixture, Category("GitHub Issue")]
    public class Issues
    {
        [Test(Description = "GitHub pull request #97")]
        public void TestNearestNeighbor2()
        {
            var kd = new KdTree<string>();
            const int Count = 8;

            for (int row = 0; row < Count; ++row)
            {
                for (int column = 0; column < Count; ++column)
                {
                    kd.Insert(new Coordinate(column, row), (column * 100 + row).ToString());
                }
            }

            var testCoordinate = new Coordinate(Count / 2, Count / 2);
            var res = kd.NearestNeighbor(testCoordinate);

            Assert.AreEqual(testCoordinate, res.Coordinate);
        }

        [Test(Description = "GitHub issue request #114")]
        public void Should_Find_Correct_Number_Of_Polygons_From_Lines()
        {
            var paths = new List<Geometry>();
            var factory = GeometryFactory.Default;

            for (int x = 1; x < 10; x++)
            {
                var startPoint = new Coordinate(x * 40, 30);
                var endPoint = new Coordinate(x * 40, 440);
                paths.Add(factory.CreateLineString(new[] { startPoint, endPoint }));
            }

            for (int y = 1; y < 10; y++)
            {
                var startPoint = new Coordinate(30, y * 40);
                var endPoint = new Coordinate(450, y * 40);

                paths.Add(factory.CreateLineString(new[] { startPoint, endPoint }));
            }

            var noder = new Noding.Snapround.GeometryNoder(new PrecisionModel(1.0d));

            var geomFactory = GeometryFactory.Default;
            var nodedLines = noder.Node(paths);

            var nodedDedupedLinework = geomFactory.BuildGeometry(nodedLines.ToArray()).Union();
            // WKT committed by FObermaier in 01/31/2017 => test fail
            string expected = @"MULTILINESTRING ((40 30,  40 40),  (40 40,  40 80),  (40 80,  40 120),  (40 120,  40 160),  (40 160,  40 200),  (40 200,  40 240),  (40 240,  40 280),  (40 280,  40 320),  (40 320,  40 360),  (40 360,  40 440),  (80 30,  80 40),  (80 40,  80 80),  (80 80,  80 120),  (80 120,  80 160),  (80 160,  80 200),  (80 200,  80 240),  (80 240,  80 280),  (80 280,  80 320),  (80 320,  80 360),  (80 360,  80 440),  (120 30,  120 40),  (120 40,  120 80),  (120 80,  120 120),  (120 120,  120 160),  (120 160,  120 200),  (120 200,  120 240),  (120 240,  120 280),  (120 280,  120 320),  (120 320,  120 360),  (120 360,  120 440),  (160 30,  160 40),  (160 40,  160 80),  (160 80,  160 120),  (160 120,  160 160),  (160 160,  160 200),  (160 200,  160 240),  (160 240,  160 280),  (160 280,  160 320),  (160 320,  160 360),  (160 360,  160 440),  (200 30,  200 40),  (200 40,  200 80),  (200 80,  200 120),  (200 120,  200 160),  (200 160,  200 200),  (200 200,  200 240),  (200 240,  200 280),  (200 280,  200 320),  (200 320,  200 360),  (200 360,  200 440),  (240 30,  240 40),  (240 40,  240 80),  (240 80,  240 120),  (240 120,  240 160),  (240 160,  240 200),  (240 200,  240 240),  (240 240,  240 280),  (240 280,  240 320),  (240 320,  240 360),  (240 360,  240 440),  (280 30,  280 40),  (280 40,  280 80),  (280 80,  280 120),  (280 120,  280 160),  (280 160,  280 200),  (280 200,  280 240),  (280 240,  280 280),  (280 280,  280 320),  (280 320,  280 360),  (280 360,  280 440),  (320 30,  320 40),  (320 40,  320 80),  (320 80,  320 120),  (320 120,  320 160),  (320 160,  320 200),  (320 200,  320 240),  (320 240,  320 280),  (320 280,  320 320),  (320 320,  320 360),  (320 360,  320 440),  (360 30,  360 40),  (360 40,  360 80),  (360 80,  360 120),  (360 120,  360 160),  (360 160,  360 200),  (360 200,  360 240),  (360 240,  360 280),  (360 280,  360 320),  (360 320,  360 360),  (360 360,  360 440),  (30 40,  40 40),  (40 40,  80 40),  (80 40,  120 40),  (120 40,  160 40),  (160 40,  200 40),  (200 40,  240 40),  (240 40,  280 40),  (280 40,  320 40),  (320 40,  360 40),  (360 40,  450 40),  (30 80,  40 80),  (40 80,  80 80),  (80 80,  120 80),  (120 80,  160 80),  (160 80,  200 80),  (200 80,  240 80),  (240 80,  280 80),  (280 80,  320 80),  (320 80,  360 80),  (360 80,  450 80),  (30 120,  40 120),  (40 120,  80 120),  (80 120,  120 120),  (120 120,  160 120),  (160 120,  200 120),  (200 120,  240 120),  (240 120,  280 120),  (280 120,  320 120),  (320 120,  360 120),  (360 120,  450 120),  (30 160,  40 160),  (40 160,  80 160),  (80 160,  120 160),  (120 160,  160 160),  (160 160,  200 160),  (200 160,  240 160),  (240 160,  280 160),  (280 160,  320 160),  (320 160,  360 160),  (360 160,  450 160),  (30 200,  40 200),  (40 200,  80 200),  (80 200,  120 200),  (120 200,  160 200),  (160 200,  200 200),  (200 200,  240 200),  (240 200,  280 200),  (280 200,  320 200),  (320 200,  360 200),  (360 200,  450 200),  (30 240,  40 240),  (40 240,  80 240),  (80 240,  120 240),  (120 240,  160 240),  (160 240,  200 240),  (200 240,  240 240),  (240 240,  280 240),  (280 240,  320 240),  (320 240,  360 240),  (360 240,  450 240),  (30 280,  40 280),  (40 280,  80 280),  (80 280,  120 280),  (120 280,  160 280),  (160 280,  200 280),  (200 280,  240 280),  (240 280,  280 280),  (280 280,  320 280),  (320 280,  360 280),  (360 280,  450 280),  (30 320,  40 320),  (40 320,  80 320),  (80 320,  120 320),  (120 320,  160 320),  (160 320,  200 320),  (200 320,  240 320),  (240 320,  280 320),  (280 320,  320 320),  (320 320,  360 320),  (360 320,  450 320),  (30 360,  40 360),  (40 360,  80 360),  (80 360,  120 360),  (120 360,  160 360),  (160 360,  200 360),  (200 360,  240 360),  (240 360,  280 360),  (280 360,  320 360),  (320 360,  360 360),  (360 360,  450 360))";
            // WKT from previous (to 01/31/2017) commit => test ok
            expected = @"MULTILINESTRING ((40 30, 40 40), (40 40, 40 80), (40 80, 40 120), (40 120, 40 160), (40 160, 40 200), (40 200, 40 240), (40 240, 40 280), (40 280, 40 320), (40 320, 40 360), (40 360, 40 440), (80 30, 80 40), (80 40, 80 80), (80 80, 80 120), (80 120, 80 160), (80 160, 80 200), (80 200, 80 240), (80 240, 80 280), (80 280, 80 320), (80 320, 80 360), (80 360, 80 440), (120 30, 120 40), (120 40, 120 80), (120 80, 120 120), (120 120, 120 160), (120 160, 120 200), (120 200, 120 240), (120 240, 120 280), (120 280, 120 320), (120 320, 120 360), (120 360, 120 440), (160 30, 160 40), (160 40, 160 80), (160 80, 160 120), (160 120, 160 160), (160 160, 160 200), (160 200, 160 240), (160 240, 160 280), (160 280, 160 320), (160 320, 160 360), (160 360, 160 440), (200 30, 200 40), (200 40, 200 80), (200 80, 200 120), (200 120, 200 160), (200 160, 200 200), (200 200, 200 240), (200 240, 200 280), (200 280, 200 320), (200 320, 200 360), (200 360, 200 440), (240 30, 240 40), (240 40, 240 80), (240 80, 240 120), (240 120, 240 160), (240 160, 240 200), (240 200, 240 240), (240 240, 240 280), (240 280, 240 320), (240 320, 240 360), (240 360, 240 440), (280 30, 280 40), (280 40, 280 80), (280 80, 280 120), (280 120, 280 160), (280 160, 280 200), (280 200, 280 240), (280 240, 280 280), (280 280, 280 320), (280 320, 280 360), (280 360, 280 440), (320 30, 320 40), (320 40, 320 80), (320 80, 320 120), (320 120, 320 160), (320 160, 320 200), (320 200, 320 240), (320 240, 320 280), (320 280, 320 320), (320 320, 320 360), (320 360, 320 440), (360 30, 360 40), (360 40, 360 80), (360 80, 360 120), (360 120, 360 160), (360 160, 360 200), (360 200, 360 240), (360 240, 360 280), (360 280, 360 320), (360 320, 360 360), (360 360, 360 440), (30 40, 40 40), (40 40, 80 40), (80 40, 120 40), (120 40, 160 40), (160 40, 200 40), (200 40, 240 40), (240 40, 280 40), (280 40, 320 40), (320 40, 360 40), (360 40, 450 40), (30 80, 40 80), (40 80, 80 80), (80 80, 120 80), (120 80, 160 80), (160 80, 200 80), (200 80, 240 80), (240 80, 280 80), (280 80, 320 80), (320 80, 360 80), (360 80, 450 80), (30 120, 40 120), (40 120, 80 120), (80 120, 120 120), (120 120, 160 120), (160 120, 200 120), (200 120, 240 120), (240 120, 280 120), (280 120, 320 120), (320 120, 360 120), (360 120, 450 120), (30 160, 40 160), (40 160, 80 160), (80 160, 120 160), (120 160, 160 160), (160 160, 200 160), (200 160, 240 160), (240 160, 280 160), (280 160, 320 160), (320 160, 360 160), (360 160, 450 160), (30 200, 40 200), (40 200, 80 200), (80 200, 120 200), (120 200, 160 200), (160 200, 200 200), (200 200, 240 200), (240 200, 280 200), (280 200, 320 200), (320 200, 360 200), (360 200, 450 200), (30 240, 40 240), (40 240, 80 240), (80 240, 120 240), (120 240, 160 240), (160 240, 200 240), (200 240, 240 240), (240 240, 280 240), (280 240, 320 240), (320 240, 360 240), (360 240, 450 240), (30 280, 40 280), (40 280, 80 280), (80 280, 120 280), (120 280, 160 280), (160 280, 200 280), (200 280, 240 280), (240 280, 280 280), (280 280, 320 280), (320 280, 360 280), (360 280, 450 280), (30 320, 40 320), (40 320, 80 320), (80 320, 120 320), (120 320, 160 320), (160 320, 200 320), (200 320, 240 320), (240 320, 280 320), (280 320, 320 320), (320 320, 360 320), (360 320, 450 320), (30 360, 40 360), (40 360, 80 360), (80 360, 120 360), (120 360, 160 360), (160 360, 200 360), (200 360, 240 360), (240 360, 280 360), (280 360, 320 360), (320 360, 360 360), (360 360, 450 360))";
            Assert.AreEqual(expected, nodedDedupedLinework.ToString());
            var polygonizer = new Polygonizer();
            polygonizer.Add(nodedDedupedLinework);

            var polygons = polygonizer.GetPolygons();
            Assert.AreEqual(64, polygons.Count);
        }

        [Test(Description = "GitHub Issue #122/1")]
        public void Polygon_intersection_Error1()
        {
            //Arrange
            var reader = new WKTReader();
            var g1 = reader.Read(
@"LINEARRING (0.0000000000000000 -6.1026585365860800,  3.8000000000000000 -5.7690000000000000,
7.9000000000000000 -5.4090000000000000,  12.0000000000000000 -5.3690000000000000,
16.1000000000000000 -5.0990000000000000,  20.3000000000000000 -5.1390000000000000,
24.5000000000000000 -4.9090000000000000,  28.5000000000000000 -4.6390000000000000,
32.6000000000000000 -4.1890000000000000,  36.6000000000000000 -3.2890000000000000,
40.8000000000000000 -3.1890000000000000,  44.8000000000000000 -2.9490000000000000,
49.0000000000000000 -2.8200000000000000,  53.3000000000000000 -2.3600000000000000,
57.3000000000000000 -2.1400000000000000,  61.3000000000000000 -1.5200000000000000,
65.5000000000000000 -1.3900000000000000,  69.6000000000000000 -1.1700000000000000,
73.6000000000000000 -1.0100000000000000,  77.9000000000000000 -0.9700000000000000,
82.0000000000000000 -0.8000000000000000,  86.2000000000000000 -0.7400000000000000,
90.2000000000000000 -0.6700000000000000,  94.5000000000000000 -0.6000000000000000,
98.6000000000000000 -0.5000000000000000,  102.6000000000000000 -0.3900000000000000,
106.9000000000000000 -0.2400000000000000,  111.1000000000000000 -0.1500000000000000,
115.1000000000000000 -0.1510000000000000,  119.2000000000000000 -0.1510000000000000,
123.1000000000000000 -0.0510000000000000,  127.2000000000000000 0.0590000000000000,
131.2000000000000000 0.2090000000000000,  132.2000000000000000 1.0250000000000000,
132.6000000000000000 2.6400000000000000,  132.6000000000000000 2.6400000000000000,
132.6000000000000000 0.5000000000000000,  128.6000000000000000 0.5000000000000000,
106.2000000000000000 -2.3000000000000000,  49.9000000000000000 -2.3000000000000000,
31.0000000000000000 -5.4500000000000000,  0.0000000000000000 -5.4500000000000000,
0.0000000000000000 -6.1026585365860800)");
            var g2 = reader.Read(
@"LINEARRING (0.0000000000000000 2.6400000000000000,  0.0000000000000000 -6.0030697674418800,
0.3200000000000000 -5.9800000000000000,  4.6200000000000000 -5.6700000000000000,
8.8100000000000000 -5.1600000000000000,  13.1000000000000000 -5.2100000000000000,
17.1200000000000000 -4.8400000000000000,  21.2800000000000000 -5.1700000000000000,
25.5000000000000000 -5.1600000000000000,  29.7100000000000000 -4.6700000000000000,
34.1300000000000000 -3.9000000000000000,  38.2800000000000000 -3.4800000000000000,
42.6400000000000000 -3.3000000000000000,  46.9800000000000000 -3.2600000000000000,
51.1600000000000000 -2.8900000000000000,  55.3200000000000000 -2.8300000000000000,
59.3000000000000000 -2.9200000000000000,  63.6800000000000000 -2.3600000000000000,
67.4800000000000000 -2.5700000000000000,  70.8900000000000000 -2.4800000000000000,
74.8900000000000000 -2.5500000000000000,  78.8800000000000000 -2.6900000000000000,
82.4400000000000000 -2.4600000000000000,  87.5300000000000000 -2.8100000000000000,
91.6200000000000000 -2.7500000000000000,  95.9000000000000000 -2.1000000000000000,
100.2200000000000000 -2.2100000000000000,  104.4400000000000000 -2.3000000000000000,
109.1500000000000000 -1.9500000000000000,  113.2400000000000000 -2.8200000000000000,
117.2400000000000000 -1.9600000000000000,  120.9400000000000000 -1.3000000000000000,
125.0300000000000000 -0.8200000000000000,  126.6800000000000000 0.1300000000000000,
132.5500000000000000 1.5400000000000000,  132.6000000000000000 2.6400000000000000,
132.6000000000000000 2.6400000000000000,  132.6000000000000000 2.6400000000000000,
0.0000000000000000 2.6400000000000000)");

            //Act
            var res = g1.Intersection(g2);

            //
            ToImage(1, g1, g2, res);

            // Assert
            Assert.That(res, Is.Not.Null);
            Debug.WriteLine(res.AsText());

        }

        [Test(Description = "GitHub Issue #122/2")]
        public void Polygon_intersection_Error2()
        {
            //Arrange
            var reader = new WKTReader();
            var g1 = reader.Read(
@"POLYGON ((0.0000000000000000 -6.1026585365860800,  3.8000000000000000 -5.7690000000000000,
7.9000000000000000 -5.4090000000000000,  12.0000000000000000 -5.3690000000000000,
16.1000000000000000 -5.0990000000000000,  20.3000000000000000 -5.1390000000000000,
24.5000000000000000 -4.9090000000000000,  28.5000000000000000 -4.6390000000000000,
32.6000000000000000 -4.1890000000000000,  36.6000000000000000 -3.2890000000000000,
40.8000000000000000 -3.1890000000000000,  44.8000000000000000 -2.9490000000000000,
49.0000000000000000 -2.8200000000000000,  53.3000000000000000 -2.3600000000000000,
57.3000000000000000 -2.1400000000000000,  61.3000000000000000 -1.5200000000000000,
65.5000000000000000 -1.3900000000000000,  69.6000000000000000 -1.1700000000000000,
73.6000000000000000 -1.0100000000000000,  77.9000000000000000 -0.9700000000000000,
82.0000000000000000 -0.8000000000000000,  86.2000000000000000 -0.7400000000000000,
90.2000000000000000 -0.6700000000000000,  94.5000000000000000 -0.6000000000000000,
98.6000000000000000 -0.5000000000000000,  102.6000000000000000 -0.3900000000000000,
106.9000000000000000 -0.2400000000000000,  111.1000000000000000 -0.1500000000000000,
115.1000000000000000 -0.1510000000000000,  119.2000000000000000 -0.1510000000000000,
123.1000000000000000 -0.0510000000000000,  127.2000000000000000 0.0590000000000000,
131.2000000000000000 0.2090000000000000,  132.2000000000000000 1.0250000000000000,
132.6000000000000000 2.6400000000000000,  132.6000000000000000 2.6400000000000000,
132.6000000000000000 0.5000000000000000,  128.6000000000000000 0.5000000000000000,
106.2000000000000000 -2.3000000000000000,  49.9000000000000000 -2.3000000000000000,
31.0000000000000000 -5.4500000000000000,  0.0000000000000000 -5.4500000000000000,
0.0000000000000000 -6.1026585365860800))");
            var isValidOp = new IsValidOp(g1);
            if (!isValidOp.IsValid)
            {
                Debug.WriteLine("g1 is not valid:" + isValidOp.ValidationError);
                g1 = g1.Buffer(0);
                Debug.WriteLine(g1.AsText());
            }
            var g2 = reader.Read(
@"POLYGON ((0.0000000000000000 2.6400000000000000,  0.0000000000000000 -6.0030697674418800,
0.3200000000000000 -5.9800000000000000,  4.6200000000000000 -5.6700000000000000,
8.8100000000000000 -5.1600000000000000,  13.1000000000000000 -5.2100000000000000,
17.1200000000000000 -4.8400000000000000,  21.2800000000000000 -5.1700000000000000,
25.5000000000000000 -5.1600000000000000,  29.7100000000000000 -4.6700000000000000,
34.1300000000000000 -3.9000000000000000,  38.2800000000000000 -3.4800000000000000,
42.6400000000000000 -3.3000000000000000,  46.9800000000000000 -3.2600000000000000,
51.1600000000000000 -2.8900000000000000,  55.3200000000000000 -2.8300000000000000,
59.3000000000000000 -2.9200000000000000,  63.6800000000000000 -2.3600000000000000,
67.4800000000000000 -2.5700000000000000,  70.8900000000000000 -2.4800000000000000,
74.8900000000000000 -2.5500000000000000,  78.8800000000000000 -2.6900000000000000,
82.4400000000000000 -2.4600000000000000,  87.5300000000000000 -2.8100000000000000,
91.6200000000000000 -2.7500000000000000,  95.9000000000000000 -2.1000000000000000,
100.2200000000000000 -2.2100000000000000,  104.4400000000000000 -2.3000000000000000,
109.1500000000000000 -1.9500000000000000,  113.2400000000000000 -2.8200000000000000,
117.2400000000000000 -1.9600000000000000,  120.9400000000000000 -1.3000000000000000,
125.0300000000000000 -0.8200000000000000,  126.6800000000000000 0.1300000000000000,
132.5500000000000000 1.5400000000000000,  132.6000000000000000 2.6400000000000000,
132.6000000000000000 2.6400000000000000,  132.6000000000000000 2.6400000000000000,
0.0000000000000000 2.6400000000000000))");
            isValidOp = new IsValidOp(g1);
            if (!isValidOp.IsValid)
            {
                Debug.WriteLine("g2 is not valid:" + isValidOp.ValidationError);
                g2 = g2.Buffer(0);
                Debug.WriteLine(g2.AsText());
            }
            //Act
            var res = g1.Intersection(g2);

            //
            ToImage(1, g1, g2, res);

            // Assert
            Assert.That(res, Is.Not.Null);
            Debug.WriteLine(res.AsText());

        }

        [Test(Description = "GitHub Issue #123")]
        public void LineString_Intersection_with_Polygon_return_null1()
        {
            //Arrange
            var reader = new WKTReader();
            var g1 = reader.Read(@"LINESTRING (0 -5.15,  30 -5.15,  48.9 -2,  105.2 -2,  127.6 0.8,  132.6 0.8)");
            var g2 = reader.Read(@"LINEARRING (15.325555555555553 0.8,  42.309375 0.8,  42.309375 -5.15,  15.325555555555553 -5.15,  15.325555555555553 0.8)");

            //Act
            var res = g1.Intersection(g2);

            //
            ToImage(1, g1, g2, res);

            // Assert
            Assert.That(res, Is.Not.Null);
            Debug.WriteLine(res.AsText());
        }

        [Test(Description = "GitHub Issue #125")]
        public void Fixing_invalid_polygon_with_Buffer_0_returns_empty_polygon()
        {
            //arrange
            var rdr = new WKTReader();
            var poly = rdr.Read(
@"POLYGON ((1.4749999999994841 -5.15,  30 -5.15,  48.9 -2,
108.1997 -2,  130.25148787313435 0.75647348414179227,
130.25148787313435 0.75647348414179161,  130 0.75,
126.3 0.72,  122.9 0.7,  119.2 0.42,  115.2 0.45,
111 0.29,  106.9 0.23,  102.8 0.2,  98.8 0.12,  94.8 0.04,
90.7 -0.08,  86.5 -0.2,  82.4 -0.42,  78.3 -0.57,  74.1 -0.69,
69.9 -0.78,  65.8 -0.87,  61.7 -1.07,  57.7 -1.09,  53.7 -1.229,
49.5 -1.289,  45.3 -1.369,  41.2 -1.719,  37 -2.409,  32.8 -3.219,
28.6 -3.769,  24.5 -4.089,  20.4 -4.429,  16.3 -4.599,  12.1 -4.759,
8 -4.889,  4 -5.049,  1.4749999999994841 -5.15))");

            //act
            var gpr = new GeometryPrecisionReducer(new PrecisionModel(1e10));
            var poly1 = gpr.Reduce(poly);
            var poly2 = poly.Buffer(0);
            var shellCoords = ((Polygon)poly).Shell.CoordinateSequence.Copy();
            CoordinateSequences.Reverse(shellCoords);
            var shell = poly.Factory.CreatePolygon(shellCoords).Buffer(0);

            ToImage(0, poly, poly1, poly2);

            var isValidOp = new IsValidOp(poly);
            if (!isValidOp.IsValid)
            {
                Debug.WriteLine(isValidOp.ValidationError);

            }
            Debug.WriteLine(poly1.AsText());
            // assert
            //Assert.That(poly.IsValid,  Is.True,  "poly.IsValid");
            Assert.That(poly1.IsValid, Is.True, "poly1.IsValid");
            Assert.That(poly2, Is.Not.Null, "poly2 != null");

            //Known to fail
            //Assert.That(poly2.IsEmpty,  Is.False,  "poly2.IsEmpty");
        }

        [Test(Description = "GitHub Issue #126")]

        [TestCase(Ordinates.XY)]
        [TestCase(Ordinates.XYZ)]
        [TestCase(Ordinates.XYM)]
        [TestCase(Ordinates.XYZM)]
        public void GeometryFactory_CreatePoint_does_not_have_Measure_value(Ordinates ordinates)
        {
            var csf = new DotSpatialAffineCoordinateSequenceFactory(ordinates);
            var gf = new GeometryFactory(csf);
            var cs1 = csf.Create(1, Ordinates.XYZM);
            Assert.That(cs1.Ordinates == ordinates);

            cs1.SetOrdinate(0, Ordinate.X, 1);
            cs1.SetOrdinate(0, Ordinate.Y, 2);
            cs1.SetOrdinate(0, Ordinate.Z, 3);
            cs1.SetOrdinate(0, Ordinate.M, 4);

            var cs2 = csf.Create(new[] { new Coordinate(1, 2) });
            Assert.That(cs1.Ordinates == ordinates);

            var pt1 = gf.CreatePoint(cs1);
            var pt2 = gf.CreatePoint(cs2);
            var pt3 = gf.CreatePoint(new Coordinate(1, 2));

            Assert.That(pt1, Is.EqualTo(pt2));
            Assert.That(pt2, Is.EqualTo(pt3));
            Assert.That(pt1, Is.EqualTo(pt3));
        }

        [Test, Category("GitHub Issue"), Description("Polygon / rectangle intersection returns empty polygon"), Ignore("Claim not valid")]
        public void TestIssue149()
        {
            var wktreader = new WKTReader();
            var polygon = wktreader.Read(@"POLYGON ((-62.327500000000008 77.777166176470587,
 -62.327500000000008 191.08,
 -39.425176714633878 191.08,
 30.449401394787785 111.06902734290135,
 34.240621449176217 85.17199763498823,
 32.474968591813237 66.628747810197453,
 29.724407935551895 56.755281980420541,
 10.963366463822565 52.502595315049426,
 -62.327500000000008 77.777166176470587))");

            Assert.IsTrue(polygon.IsValid);
            Assert.IsFalse(polygon.IsEmpty);

            var boundingbox = wktreader.Read(@"POLYGON ((-52.5 -34, -52.5 34, 52.5 34, 52.5 -34, -52.5 -34))");
            Assert.IsTrue(boundingbox.IsValid);
            Assert.IsFalse(boundingbox.IsEmpty);

            //ToImage(1, polygon, boundingbox, boundingbox.Intersection(polygon));
            //Assert.IsTrue(polygon.Intersects(boundingbox));

            var result = boundingbox.Intersection(polygon); // =>{ POLYGON EMPTY }
            //Assert.IsFalse(result.IsEmpty, "result.IsEmpty");
            var resultInverted = polygon.Intersection(boundingbox); // => { POLYGON EMPTY }
            //Assert.IsFalse(resultInverted.IsEmpty, "result.IsEmpty");
        }

        [Test, Description("TopologyException when generating a VoronoiDiagram"), Ignore("Known to fail, waiting for fix from JTS.")]
        public void TestIssue151()
        {
            var wktreader = new WKTReader();
            var polygon = wktreader.Read("POLYGON((14.7119 201.6703, 74.2154 201.6703, 74.2154 166.6391, 14.7119 166.6391, 14.7119 201.6703))");

            Assert.IsTrue(polygon.IsValid);
            var vdb = new Triangulate.VoronoiDiagramBuilder();
            vdb.SetSites(polygon);
            Geometry result = null;
            Assert.DoesNotThrow(() => result= vdb.GetDiagram(polygon.Factory));
            Assert.IsNotNull(result);
        }

        [Test, Description("SnapRoundOverlayFunctions, Casting error in GeometryEditorEx.EditGeometryCollection")]
        public void TestIssue177()
        {
            var reader = new WKTReader();
            var g1 = reader.Read("MULTIPOLYGON (((30 20, 45 40, 10 40, 30 20)),((15 5, 40 10, 10 20, 5 10, 15 5)))");
            var g2 = reader.Read("MULTIPOLYGON (((40 40, 20 45, 45 30, 40 40)),((20 35, 10 30, 10 10, 30 5, 45 20, 20 35),(30 20, 20 15, 20 25, 30 20)))");

            Geometry res = null;
            //                                                                                     |
            //                                                             Some cruel scale factor V
            Assert.DoesNotThrow(() => res = SnapRoundOverlayFunctions.SnappedIntersection(g1, g2, 0.1));
            Assert.That(res, Is.Not.Null);
            Assert.That(res.IsValid, Is.True);

            ToImage(0, g1, g2, res);

            // To show that the result is correct:
            var r1 = new GeometryPrecisionReducer(new PrecisionModel(0.1)).Reduce(g1);
            var r2 = new GeometryPrecisionReducer(new PrecisionModel(0.1)).Reduce(g2);
            ToImage(0, r1, r2, res);
        }

        [Test(Description = "Bug in difference method #141"), Category("GitHub Issue")]
        public void TestDifference()
        {
            byte[] buffer1 = WKBReader.HexToBytes(
               //         1         2         3         4         5         6         7         8
                "010300000001000000050000000793adf1033458c0a8a08e06b1b34740f6df183c0b3458c0b9d84f"+
                "27e2b247403559f264963458c0f0eeec1ee5b24740fd9c0635913458c067f1cffdb6b347400793ad"+
                "f1033458c0a8a08e06b1b34740");
            byte[] buffer2 = WKBReader.HexToBytes(
                "010300000001000000b5000000f5a3caf8a33458c0f66779fd4db34740e8b86ed8a33458c09207fc"+
                "c351b3474086533f9fa33458c0f5565f8855b347407dda504da33458c04fb0764959b347401eedba"+
                "e2a23458c0929317065db3474080e79f5fa23458c0ec5a15bd60b34740208427c4a13458c011ac47"+
                "6d64b347405f278310a13458c07452881568b34740e4a6ea44a03458c00d65b5b46bb347409a499c"+
                "619f3458c011d5aa496fb347403313e1669e3458c0fb044bd372b34740064005559d3458c0c6a27c"+
                "5076b347402dc95f2c9c3458c0338228c079b3474049f34bed9a3458c081c23b217db34740fd992d"+
                "98993458c038f4a97280b34740161d702d983458c0dda766b383b34740e02783ad963458c0417674"+
                "e286b34740cde9dd18953458c0ef86cffe89b34740383c0170933458c0fce380078db34740590b6f"+
                "b3913458c001e395fb8fb3474053edb2e38f3458c09a7024da92b34740a8d65d018e3458c0e9c446"+
                "a295b34740431a060d8c3458c014af1f5398b347406d6947078a3458c00424d4eb9ab34740b9e6c3"+
                "f0873458c0ac469a6b9db34740390022ca853458c04014a6d19fb34740836f0c94833458c03c923a"+
                "1da2b347405372354f813458c0a0119f4da4b34740266c51fc7e3458c034a02562a6b34740fc0b19"+
                "9c7c3458c0c5e2285aa8b3474019724b2f7a3458c064ef0935aab3474084e4a9b6773458c0ec9834"+
                "f2abb34740e5e1f832753458c0fe6e1f91adb34740474702a5723458c03e984911afb34740781792"+
                "0d703458c09aac3872b0b34740077b766d6d3458c084267fb3b1b3474005e681c56a3458c0393dba"+
                "d4b2b3474066df8716683458c0fabe8fd5b3b34740bd265f61653458c00f11aeb5b4b34740828edf"+
                "a6623458c0477bd074b5b34740ea0ee3e75f3458c079dcba12b6b347400db344255d3458c043d03b"+
                "8fb6b34740c6abe15f5a3458c049892aeab6b347400c179698573458c03a686f23b7b347407a4b41"+
                "d0543458c087f3f43ab7b34740e979c007523458c0aedfb630b7b34740f5f8f23f4f3458c0ae2cb5"+
                "04b7b3474077f9b5794c3458c011bdfcb6b6b3474047ace6b5493458c0a97ba947b6b347403f4262"+
                "f5463458c08ec4ddb6b5b3474059d90439443458c0df8ac404b5b34740aa7ca981413458c042a495"+
                "31b4b34740caeb26d03e3458c024a3933db3b347400c0c56253c3458c0fbb00929b2b34740278a0c"+
                "82393458c085ff51f4b0b347406ab41ae7363458c07fc0c69fafb3474006ec5155343458c0399fd7"+
                "2baeb347408a5980cd313458c0fe46f498acb3474041b46d502f3458c029919de7aab3474056c6e2"+
                "de2c3458c056c85a18a9b34740afe8a1792a3458c069a8b92ba7b34740913b6a21283458c0cccf54"+
                "22a5b34740bd93f6d6253458c0b399d1fca2b3474034a0ff9a233458c01087d7bba0b34740ef7835"+
                "6e213458c023211d609eb34740e93548511f3458c0ad3c5dea9bb347403445de441d3458c0376b5d"+
                "5b99b3474004029d491b3458c08fafe9b396b3474085302160193458c088a3d6f493b34740415c04"+
                "89173458c0772cfd1e91b347409f8cd8c4153458c03d1243338eb347405f902c14143458c07a4290"+
                "328bb34740608c8577123458c0141cd31d88b34740db6c65ef103458c0f59402f684b34740048745"+
                "7c0f3458c08bee19bc81b34740abd1991e0e3458c04b011d717eb347405dd2cfd60c3458c06bcb11"+
                "167bb347409f774ca50b3458c0a29602ac77b347405577718a0a3458c0b043023474b34740bbb794"+
                "86093458c052b323af70b3474089d3079a083458c04e5d821e6db3474033f415c5073458c0a69337"+
                "8369b347406686ff07073458c0256567de65b347406b98ff62063458c01295313162b3474046c74a"+
                "d6053458c0f957bc7c5eb34740f6180d62053458c0e92d32c25ab3474092e96806053458c06e4bb9"+
                "0257b34740b0497cc3043458c058567e3f53b34740deb25c99043458c0b7ceab794fb34740e4e114"+
                "88043458c0dca572b24bb34740e75aac8f043458c09a81ffea47b3474064d21eb0043458c0bf077f"+
                "2444b3474036c464e9043458c01ede1d6040b34740aec9693b053458c046d00a9f3cb34740bc1d14"+
                "a6053458c087386ee238b34740e89c4429063458c071e2762b35b34740f766d1c4063458c052284d"+
                "7b31b34740662a8a78073458c033f312d32db34740ceeb3444083458c0619df0332ab34740247794"+
                "27093458c025ea059f26b3474096db60220a3458c041516e1523b347400eb74b340b3458c0ff9549"+
                "981fb347403136005d0c3458c01b99aa281cb347409eee209c0d3458c09415a2c718b3474092174b"+
                "f10e3458c028ec427615b347402164145c103458c04af5903512b347405ef009dc113458c0f25494"+
                "060fb3474095b2b670133458c08f4c48ea0bb347406fd19919153458c00ad2a3e108b34740d14d30"+
                "d6163458c050db9ded05b347409991eea5183458c001561e0f03b3474004ce45881a3458c0fd090b"+
                "4700b3474089779c7c1c3458c0de4d4396fdb247407e155a821e3458c038e19dfdfab24740905fda"+
                "98203458c0dac6e67df8b24740ade776bf223458c09101ea17f6b247401e0786f5243458c0e08b64"+
                "ccf3b247402980553a273458c004ef0c9cf1b247406fdc2f8d293458c0fa429387efb24740d27f5d"+
                "ed2b3458c0b3089f8fedb247402b371e5a2e3458c09cdecab4ebb2474037e2b2d2303458c0dcf1aa"+
                "f7e9b24740acc95356333458c054fecc58e8b247401e493ae4353458c0da91add8e6b24740c25d9a"+
                "7b383458c0463ac977e5b2474047b9a31b3b3458c062578b36e4b247407bfa84c33d3458c0f1b156"+
                "15e3b247404dad6b72403458c036c78914e2b2474047ff7f27433458c066e67134e1b24740f41dea"+
                "e1453458c0b1c75375e0b24740fd23d1a0483458c0c3d76fd7dfb247406af358634b3458c0ba09f1"+
                "5adfb24740426ea5284e3458c0b5500200dfb24740ab63d9ef503458c08597bfc6deb24740cea217"+
                "b8533458c0380c3aafdeb24740efe78180563458c0112078b9deb2474073dc3848593458c04fad77"+
                "e5deb24740824f5f0e5c3458c06ad12b33dfb24740232319d25e3458c04fc77aa2dfb24740bb0087"+
                "92613458c0e7324233e0b2474013f0ce4e643458c051fb54e5e0b24740f3f81606673458c0aa707d"+
                "b8e1b24740262385b7693458c0c2da76ace2b2474016af42626c3458c02310f6c0e3b24740d0157c"+
                "056f3458c0942aa5f5e4b2474081bd5ca0713458c0d2ac254ae6b24740ba6a1532743458c051110a"+
                "bee7b24740cc07d9b9763458c00387e050e9b24740acb7dd36793458c04f5a2a02ebb24740efd55c"+
                "a87b3458c0994060d1ecb24740af09940d7e3458c0fd7df4bdeeb24740e60cc265803458c04f4e4a"+
                "c7f0b24740b41d2db0823458c01f7cbeecf2b24740dab21eec843458c07786a92df5b247409c8ee4"+
                "18873458c019e45489f7b247400099ce35893458c045c005fff9b24740d47637428b3458c07089f6"+
                "8dfcb24740e4cc793d8d3458c00c175935ffb2474044b1f6268f3458c08b405ff401b347402abe16"+
                "fe903458c08f8927ca04b3474050d946c2923458c07f9bd2b507b34740b459fa72943458c0b98878"+
                "b60ab34740d8e1a80f963458c0d5a626cb0db3474025bed397973458c06b4beaf210b34740c460fe"+
                "0a993458c08be9c32c14b34740c4e5b5689a3458c041f4b37717b347407cda8db09b3458c05a6db4"+
                "d21ab34740a42a1fe29c3458c09abfb63c1eb3474019460afd9d3458c0877baeb421b34740bf33f8"+
                "009f3458c01d4f823925b34740de5897ed9f3458c01b0e1bca28b34740018c9cc2a03458c0bd405d"+
                "652cb347407c60c77fa13458c0f9fd260a30b3474045a2da24a23458c0063754b733b34740f9eca4"+
                "b1a23458c09c28c56b37b34740b827f925a33458c029074b263bb34740abd0b281a33458c0e3c3c1"+
                "e53eb34740fdfcb5c4a33458c03793faa842b347401f33edeea33458c017f5ca6e46b34740a87d4a"+
                "00a43458c0f21d04364ab34740f5a3caf8a33458c0f66779fd4db34740");

            NtsGeometryServices.Instance = new NtsGeometryServices();
            var reader = new WKBReader();
            var geom1 = reader.Read(buffer1);
            var geom2 = reader.Read(buffer2);

            var gu = geom1.Difference(geom2);
            Assert.IsNotNull(gu);
            Assert.IsTrue(gu is IPolygonal);

            ToImage(141, geom1, geom2, gu);
            ToImage(141, geom1, geom2, null);
            ToImage(141, null, null, gu);
            ToImage(141, null, null, geom2.Difference(geom1));

            bool isCCW = Orientation.IsCCW(((Polygon) gu.GetGeometryN(0)).ExteriorRing.CoordinateSequence);
            for (int i = 1; i < gu.NumGeometries; i++)
            {
                var p = (Polygon) gu.GetGeometryN(1);
                Assert.That(Orientation.IsCCW(p.ExteriorRing.CoordinateSequence), Is.EqualTo(isCCW));
            }
            Console.WriteLine("Orientation CCW = {0}", isCCW);
        }

        [Test]
        public void CrossAndIntersectionTest()
        {
            // Arrange
            var gf = new GeometryFactory(new PrecisionModel(100000000));

            var closestCoordinate = new Coordinate(152608, 594957);
            var closestLine = gf.CreateLineString(new[]
            {
                new Coordinate(152348, 595130),
                new Coordinate(152421, 595061),
                new Coordinate(152455, 595033),
                new Coordinate(152524, 595001),
                new Coordinate(152593, 594973),
                new Coordinate(152622, 594946),
                new Coordinate(152634, 594930),
                new Coordinate(152641, 594921),
                new Coordinate(152649, 594910),
                new Coordinate(152863, 594623),
                new Coordinate(152873, 594607)
            });
            var indexedLine = new NetTopologySuite.LinearReferencing.LengthIndexedLine(closestLine);
            double projectedIndex = indexedLine.Project(closestCoordinate);
            var coordinateToAdd = indexedLine.ExtractPoint(projectedIndex);
            gf.PrecisionModel.MakePrecise(coordinateToAdd);

            var line = gf.CreateLineString(new[] { new Coordinate(152503, 594904), coordinateToAdd });

            ToImage(0, closestLine, line, GeometryFactory.Default.CreatePoint(coordinateToAdd));

            // act
            var intersectionPt = line.Intersection(closestLine).Coordinate;
            gf.PrecisionModel.MakePrecise(intersectionPt);

            // assert intersection point is equal to coordinate to add
            Assert.AreEqual(coordinateToAdd, intersectionPt);

            // act insertion of coordinate to add
            var lip = new NetTopologySuite.LinearReferencing.LocationIndexOfPoint(closestLine);
            var ll = lip.IndexOf(coordinateToAdd);
            if (!ll.IsVertex)
            {
                var cl = (LineString) closestLine;
                var cls = cl.Factory.CoordinateSequenceFactory.Create(cl.CoordinateSequence.Count + 1, cl.CoordinateSequence.Ordinates);
                CoordinateSequences.Copy(cl.CoordinateSequence, 0, cls, 0, ll.SegmentIndex+1);
                cls.SetOrdinate(ll.SegmentIndex+1, Ordinate.X, coordinateToAdd.X);
                cls.SetOrdinate(ll.SegmentIndex+1, Ordinate.Y, coordinateToAdd.Y);
                CoordinateSequences.Copy(cl.CoordinateSequence, ll.SegmentIndex+1, cls, ll.SegmentIndex + 2, cl.CoordinateSequence.Count-ll.SegmentIndex-1);
                closestLine = gf.CreateLineString(cls);
            }

            ToImage(1, closestLine, line, GeometryFactory.Default.CreatePoint(coordinateToAdd));

            Assert.IsTrue(line.Touches(closestLine));
            Assert.IsFalse(line.Crosses(closestLine));
        }

        [Test(Description = "Polygonizer is creating polygons with duplicate points. #253"), Category("GitHub Issue")]
        public void PolygonizerTest()
        {
            const string wkt = @"MULTILINESTRING(
(595435.592 663003.43, 595329.952 663070.947), (595435.592 663003.43, 595422.128 662982.365), 
(595422.128 662982.365, 595408.755 662961.281), (595408.755 662961.281, 595324.534 663015.109), 
(595324.534 663015.109, 595324.534 663015.109), (595326.792 663018.656, 595324.534 663015.109), 
(595324.534 663015.109, 595324.534 663015.109), (595324.534 663015.109, 595311.111 662994.018), 
(595311.111 662994.018, 595299.143 662975.215), (595299.143 662975.215, 595297.687 662972.927), 
(595297.687 662972.927, 595284.264 662951.837), (595284.264 662951.837, 595270.782 662930.654), 
(595299.143 662975.215, 595302.092 663071.747), (595302.092 663071.747, 595302.098 663071.926), 
(595329.289 663070.966, 595327.933 663042.567), (595327.933 663042.567, 595327.933 663042.567), 
(595327.933 663042.567, 595326.792 663018.656), (595327.933 663042.567, 595422.128 662982.365), 
(595355.101 662876.893, 595341.661 662855.813), (595341.661 662855.813, 595232.906 662925.154), 
(595232.906 662925.154, 595257.509 662928.727), (595257.509 662928.727, 595257.509 662928.727), 
(595257.509 662928.727, 595270.782 662930.654), (595270.782 662930.654, 595355.101 662876.893), 
(595301.269 662789.679, 595290.846 662792.619), (595290.846 662792.619, 595229.883 662829.806), 
(595229.883 662829.806, 595231.001 662867.129), (595314.822 662813.628, 595314.822 662813.628), 
(595314.822 662813.628, 595231.001 662867.129), (595301.269 662789.679, 595314.822 662813.628), 
(595314.822 662813.628, 595314.822 662813.628), (595237.314 663073.714, 595237.311 663073.607), 
(595237.311 663073.607, 595234.348 662973.758), (595234.348 662973.758, 595232.906 662925.154), 
(595232.906 662925.154, 595231.955 662896.179), (595231.955 662896.179, 595231.955 662896.179), 
(595231.955 662896.179, 595231.001 662867.129), (595395.292 662940.216, 595311.111 662994.018), 
(595381.829 662919.151, 595297.687 662972.927), (595368.365 662898.086, 595284.264 662951.837), 
(595328.272 662834.701, 595231.955 662896.179), (595231.955 662896.179, 595231.955 662896.179), 
(595408.755 662961.281, 595395.292 662940.216), (595395.292 662940.216, 595395.292 662940.216), 
(595395.292 662940.216, 595381.829 662919.151), (595381.829 662919.151, 595381.829 662919.151), 
(595381.829 662919.151, 595368.365 662898.086), (595368.365 662898.086, 595355.101 662876.893), 
(595341.661 662855.813, 595328.272 662834.701), (595328.272 662834.701, 595328.272 662834.701), 
(595328.272 662834.701, 595314.822 662813.628), (595314.822 662813.628, 595314.822 662813.628), 
(595229.059 662774.178, 595229.883 662829.806), (595301.269 662789.679, 595292.309 662772.342), 
(595210.495 662868.477, 595216.033 662860.79), (595217.835 662853.417, 595216.033 662860.79), 
(595208.232 662974.501, 595234.348 662973.758), (595212.331 663074.324, 595237.311 663073.607),
(595237.311 663073.607, 595262.33 663072.888), (595262.33 663072.888, 595302.092 663071.747), 
(595302.092 663071.747, 595329.289 663070.966), (595329.289 663070.966, 595329.952 663070.947),
(595329.952 663070.947, 595404.571 663068.804), (595404.571 663068.804, 595455.568 663034.686),
(595455.568 663034.686, 595435.592 663003.43), (595257.509 662928.727, 595257.509 662928.727), 
(595257.509 662928.727, 595262.33 663072.888), (595262.33 663072.888, 595262.333 663073.001),
(595177.856 662775.665, 595229.059 662774.178), (595229.059 662774.178, 595229.059 662774.178),
(595229.059 662774.178, 595292.309 662772.342), (595217.835 662853.417, 595215.34 662774.577))";

            // Arrange
            var geom = new WKTReader(new GeometryFactory(new PrecisionModel(1000))).Read(wkt);

            var op = new Polygonizer();
            op.Add(geom);

            // Act
            var res = op.GetGeometry();

            // Assert
            Assert.That(res, Is.Not.Null, "geom != null");
            
            Assert.That(res.IsValid, Is.True, "IsValid");
            Assert.That(res.IsSimple, Is.True, "IsSimple");
            Assert.That(new RepeatedPointTester().HasRepeatedPoint(res), Is.False, "No repeated points");

            Assert.That(CoordinateArrays.HasRepeatedPoints(res.Coordinates), Is.False, "No repeated points");

        }

        [Test, Category("GitHub Issue"), Description("Buffer(0) throws exception with a particular valid geometry")]
        public void TestBuffer0()
        {
            const string WKT = @"GEOMETRYCOLLECTION(POLYGON((-56.4539804 6, -56 6, -56 5.9343781,
  -56.0766319 5.9452482, -56.3499378 5.9902815, -56.4539804 6)), 
  POLYGON((-56 5.9343781, -56 6, -55.8922665 6,
    -56 5.9343781)), 
  POLYGON((-57.579945444686231 6.5, -57 6.5, -57 6.0725838,
    -57.0752357 6.0660014, -57.1527016 6.2093133, -57.2301675 6.2829058,
    -57.4199588 6.4029779, -57.501298 6.4107245, -57.579945444686231 6.5)), 
  POLYGON((-57 6.0725838, -57 6.5, -56 6.5,
    -56 6, -56.4539804 6, -56.7738725 6.0555022,
    -56.9571117 6.0725838, -57 6.0725838)), 
  POLYGON((-55.8922665 6, -56 6, -56 6.5,
    -55 6.5, -55 6.0671396, -55.1593285 6.0193392,
    -55.2355814 6.0113725, -55.4540971 6.0409632, -55.5895313 6.050068,
    -55.6851319 6.0409632, -55.8376376 6.0159249, -55.8547091 6.0125105,
    -55.8922665 6)), 
  POLYGON((-58.3240277 7, -58 7, -58 6.8484066,
    -58.1452741 6.8840592, -58.2608099 6.9385573, -58.3240277 7)), 
  POLYGON((-58 6.8484066, -58 7, -57 7,
    -57 6.5, -57.579945444686231 6.5, -57.6446098 6.5734028,
    -57.7801751 6.6973482, -57.9234869 6.8019271, -58 6.8484066)), 
  POLYGON((-57 6.5, -57 7, -56 7,
    -56 6.5, -57 6.5)), 
  POLYGON((-56 6.5, -56 7, -55 7,
    -55 6.5, -56 6.5)), 
  POLYGON((-58.477437209077074 7.5, -58 7.5, -58 7,
    -58.3240277 7, -58.3937851 7.0737124, -58.395965 7.3135037,
    -58.4003249 7.3963408, -58.477437209077074 7.5)), 
  POLYGON((-58 7, -58 7.5, -57 7.5,
    -57 7, -58 7)), 
  POLYGON((-57 7, -57 7.5, -56 7.5,
    -56 7, -57 7)), 
  POLYGON((-56 7, -56 7.5, -55 7.5,
    -55 7, -56 7)), 
  POLYGON((-59 8, -58 8, -58 7.5,
    -58.477437209077074 7.5, -58.5333001 7.5750944, -58.6793548 7.7102495,
    -58.8297694 7.865024, -59 8)), 
  POLYGON((-58 7.5, -58 8, -57 8,
    -57 7.5, -58 7.5)), 
  POLYGON((-57 7.5, -57 8, -56 8,
    -56 7.5, -57 7.5)), 
  POLYGON((-56 7.5, -56 8, -55 8,
    -55 7.5, -56 7.5)), 
  POLYGON((-77.143756579869134 8.5, -77 8.5, -77 8.3265824,
    -77.0284837 8.3617687, -77.0449773 8.3815611, -77.0900599 8.4618301,
    -77.143756579869134 8.5)), 
  POLYGON((-77 8.3265824, -77 8.5, -76.987830149074028 8.5,
    -76.9877996 8.4992159, -76.9421935 8.4643622, -76.8701449 8.4354405,
    -76.8426555 8.3826608, -76.8338589 8.3342795, -76.8316597 8.3298811999999991,
    -76.8173652 8.2320189, -76.8635474 8.2221227, -76.9119287 8.2419151,
    -76.916327 8.2935952, -76.965808 8.3309807, -76.9794195 8.3321976,
    -77 8.3265824)), 
  POLYGON((-59.829685705333247 8.5, -59 8.5, -59 8,
    -59.0665614 8.0897117, -59.2439938 8.2199785, -59.5247413 8.3794431,
    -59.7201416 8.4760203, -59.8099808 8.4827582, -59.829685705333247 8.5)), 
  POLYGON((-59 8, -59 8.5, -58 8.5,
    -58 8, -59 8)), 
  POLYGON((-58 8, -58 8.5, -57 8.5,
    -57 8, -58 8)), 
  POLYGON((-57 8, -57 8.5, -56 8.5,
    -56 8, -57 8)), 
  POLYGON((-56 8, -56 8.5, -55 8.5,
    -55 8, -56 8)), 
  POLYGON((-81.608027885531257 9, -81 9, -81 8.8985004510513583,
    -81.0564867 8.8745364, -81.213844999999992 8.8492996, -81.448398 8.8760209,
    -81.5775505 8.9680606, -81.608027885531257 9)), 
  POLYGON((-81 8.8985004510513583, -81 9, -80.8174804 9,
    -80.90952 8.9368859, -81 8.8985004510513583)), 
  POLYGON((-77.6896136 9, -77 9, -77 8.5,
    -77.143756579869134 8.5, -77.1813248 8.5267051000000009, -77.2363036 8.6289657,
    -77.2978799 8.7191309999999991, -77.4313084 8.7839817, -77.5500694 8.8730524,
    -77.6317176 8.9591541, -77.6896136 9)), 
  POLYGON((-76.987830149074028 8.5, -77 8.5, -77 9,
    -76.354442863119516 9, -76.3544438 8.9999983, -76.4775961 8.9357478,
    -76.5556661 8.859877, -76.6788186 8.7763092, -76.7019097 8.7268282,
    -76.8349585 8.6982393000000009, -76.9372191 8.6597541, -76.9910983 8.5838833,
    -76.987830149074028 8.5)), 
  POLYGON((-60.6583383 9, -60 9, -60 8.6377308,
    -60.1673668 8.6801964, -60.3681521 8.7080425, -60.4736744 8.6523502,
    -60.5835933 8.7080425, -60.5997148 8.7769251, -60.5381602 8.8296862,
    -60.595318 8.9000344, -60.5997148 8.9571922, -60.6583383 9)), 
  POLYGON((-60 8.6377308, -60 9, -59 9,
    -59 8.5, -59.829685705333247 8.5, -59.9357556 8.5928112,
    -60 8.6377308)), 
  POLYGON((-59 8.5, -59 9, -58 9,
    -58 8.5, -59 8.5)), 
  POLYGON((-58 8.5, -58 9, -57 9,
    -57 8.5, -58 8.5)), 
  POLYGON((-57 8.5, -57 9, -56 9,
    -56 8.5, -57 8.5)), 
  POLYGON((-56 8.5, -56 9, -55 9,
    -55 8.5, -56 8.5)), 
  POLYGON((-82.290544470855764 9.5, -82 9.5, -82 9.2382419,
    -82.0525943 9.3213744, -82.1847159 9.4238057, -82.2782401 9.4980313,
    -82.290544470855764 9.5)), 
  POLYGON((-82 9.2382419, -82 9.5, -81 9.5,
    -81 9, -81.608027885531257 9, -81.6446251 9.0383528,
    -81.70076499999999 9.1061203, -81.7794442 9.1773769, -81.8417937 9.2367573,
    -81.9160193 9.2486334, -81.9679773 9.2233967, -82 9.2382419), 
    (-81.5901977 9.0738145, -81.5702179 9.0565233, -81.541932 9.0514464, 
      -81.4998659 9.0630508, -81.4955142 9.0848092, -81.5034923 9.1152709, 
      -81.5172725 9.1326776, -81.5412067 9.1413809, -81.5750995 9.1448373, 
      -81.59768 9.1265991, -81.5901977 9.0738145)), 
  POLYGON((-80.8174804 9, -81 9, -81 9.5,
    -80 9.5, -80 9.4148991, -80.1346047 9.3020761,
    -80.2533657 9.2619943, -80.46862 9.1981602, -80.6244938 9.1506558,
    -80.7328632 9.0793993, -80.8174804 9)), 
  POLYGON((-80 9.4148991, -80 9.5, -79.7976289791633 9.5,
    -79.8080122 9.4861557, -79.8440242 9.4622269, -79.9238041 9.4178681,
    -80 9.4148991)), 
  POLYGON((-78.687276365583244 9.5, -78 9.5, -78 9.2679326,
    -78.1646572 9.3228595, -78.3650664 9.4104457, -78.503126 9.4772487,
    -78.6634534 9.4965474, -78.687276365583244 9.5)), 
  POLYGON((-78 9.2679326, -78 9.5, -77 9.5,
    -77 9, -77.6896136 9, -77.7742307 9.1046363,
    -77.8929917 9.2040986, -78 9.2679326)), 
  POLYGON((-76.354442863119516 9, -77 9, -77 9.5,
    -76 9.5, -76 9.4393254892335214, -76.0835344 9.3896906,
    -76.1988955 9.2752751, -76.2357732 9.1986829, -76.263179 9.1292733,
    -76.3080028 9.0842669999999988, -76.354442863119516 9), 
    (-76.2038542 9.3677355, -76.1826169 9.3698591, -76.1664768 9.3783541, 
      -76.1588314 9.3881232, -76.1545839 9.3953438, -76.1545839 9.403414, 
      -76.1567077 9.4131831, -76.1635036 9.4221027, -76.1834665 9.4242264, 
      -76.20258 9.4102099, -76.2089511 9.3919459, -76.2038542 9.3677355)), 
  POLYGON((-76 9.4393254892335214, -76 9.5, -75.666006674259648 9.5,
    -75.6859015 9.4723413, -75.7661706 9.4899346, -75.8607342 9.4954324,
    -75.9684927 9.4613456, -75.9999537 9.4393529999999988, -76 9.4393254892335214)), 
  POLYGON((-60.840525824339004 9.5, -60 9.5, -60 9,
    -60.6583383 9, -60.630492 9.109613, -60.655407 9.2356535,
    -60.7272207 9.347038, -60.840525824339004 9.5)), 
  POLYGON((-60 9, -60 9.5, -59 9.5,
    -59 9, -60 9)), 
  POLYGON((-59 9, -59 9.5, -58 9.5,
    -58 9, -59 9)), 
  POLYGON((-58 9, -58 9.5, -57 9.5,
    -57 9, -58 9)), 
  POLYGON((-57 9, -57 9.5, -56 9.5,
    -56 9, -57 9)), 
  POLYGON((-56 9, -56 9.5, -55 9.5,
    -55 9, -56 9)), 
  POLYGON((-83 10, -82 10, -82 9.5,
    -82.290544470855764 9.5, -82.3524657 9.5099074, -82.517246599999993 9.6316374,
    -82.56772 9.6776573, -82.6612443 9.705862999999999, -82.7146867 9.7147700999999991,
    -82.7592221 9.7652435, -82.8200871 9.8172015, -83 10)), 
  POLYGON((-82 9.5, -82 10, -81 10,
    -81 9.5, -82 9.5)), 
  POLYGON((-81 9.5, -81 10, -80 10,
    -80 9.5, -81 9.5)), 
  POLYGON((-79.7976289791633 9.5, -80 9.5, -80 10,
    -79 10, -79 9.6409923673935953, -79.0002494 9.6409901,
    -79.1370126 9.6153082, -79.2053002 9.6078857, -79.3789881 9.6449985,
    -79.4962646 9.6717197, -79.5823663 9.6865648, -79.6892512 9.640545,
    -79.7456627 9.5692884, -79.7976289791633 9.5)), 
  POLYGON((-79 9.6409923673935953, -79 10, -78 10,
    -78 9.5, -78.687276365583244 9.5, -78.7658847 9.5113925,
    -78.7945923 9.5455881, -78.8345812 9.5901471, -78.8745701 9.6421327,
    -79 9.6409923673935953)), 
  POLYGON((-78 9.5, -78 10, -77 10,
    -77 9.5, -78 9.5)), 
  POLYGON((-77 9.5, -77 10, -76 10,
    -76 9.5, -77 9.5)), 
  POLYGON((-75.666006674259648 9.5, -76 9.5, -76 10,
    -75.628247290380884 10, -75.6282474 9.9999988, -75.6457162 9.9391671,
    -75.6837024 9.8165087, -75.7342829 9.7626295, -75.7661706 9.6911569999999987,
    -75.7562744 9.6647672, -75.6727066 9.6526719, -75.6342214 9.5998922,
    -75.6408189 9.5350172, -75.666006674259648 9.5), 
    (-75.7937462 9.8171595, -75.8110727 9.8255912, -75.8298511 9.8205837, 
      -75.8486294 9.8205837, -75.8795094 9.812655, -75.8895245 9.7947113, 
      -75.8849342 9.7725945, -75.8911937 9.7621621, -75.9087202 9.7467221, 
      -75.9195699 9.7333686, -75.9212391 9.7187632, -75.9212391 9.6991503, 
      -75.9204045 9.6866314, -75.9083029 9.6753644, -75.8774229 9.6803719, 
      -75.8757537 9.7004022, -75.8840996 9.7296129999999987, -75.8945321 9.7417146, 
      -75.8807613 9.7567372, -75.8582272 9.7750983, -75.8373624 9.7838615, 
      -75.8169149 9.7805232, -75.8018922 9.7838615, -75.7922944 9.7993015, 
      -75.7937462 9.8171595)), 
  POLYGON((-62 10, -61 10, -61 9.6108436,
    -61.0643055 9.6254993, -61.1317225 9.6474831, -61.3017306 9.747143,
    -61.3615283 9.7693048, -61.4131151 9.8482685, -61.5347588 9.9303413,
    -61.6432122 9.9611186, -61.7033013 9.9435316, -61.707698 9.9024952,
    -61.6886454 9.8643899, -61.7472689 9.8907705, -61.7956332 9.8863737,
    -61.8821028 9.9274101, -62 10)), 
  POLYGON((-61 9.6108436, -61 10, -60 10,
    -60 9.5, -60.840525824339004 9.5, -60.8444676 9.5053214,
    -60.9426905 9.5684637, -61 9.6108436)), 
  POLYGON((-60 9.5, -60 10, -59 10,
    -59 9.5, -60 9.5)), 
  POLYGON((-59 9.5, -59 10, -58 10,
    -58 9.5, -59 9.5)), 
  POLYGON((-58 9.5, -58 10, -57 10,
    -57 9.5, -58 9.5)), 
  POLYGON((-57 9.5, -57 10, -56 10,
    -56 9.5, -57 9.5)), 
  POLYGON((-56 9.5, -56 10, -55 10,
    -55 9.5, -56 9.5)), 
  POLYGON((-83.372826949887909 10.5, -83 10.5, -83 10,
    -83.1155049 10.1393403, -83.361933999999991 10.4763245, -83.372826949887909 10.5)), 
  POLYGON((-83 10, -83 10.5, -82 10.5,
    -82 10, -83 10)), 
  POLYGON((-82 10, -82 10.5, -81 10.5,
    -81 10, -82 10)), 
  POLYGON((-81 10, -81 10.5, -80 10.5,
    -80 10, -81 10)), 
  POLYGON((-80 10, -80 10.5, -79 10.5,
    -79 10, -80 10)), 
  POLYGON((-79 10, -79 10.5, -78 10.5,
    -78 10, -79 10)), 
  POLYGON((-78 10, -78 10.5, -77 10.5,
    -77 10, -78 10)), 
  POLYGON((-77 10, -77 10.5, -76 10.5,
    -76 10, -77 10)), 
  POLYGON((-75.628247290380884 10, -76 10, -76 10.5,
    -75.610231445087209 10.5, -75.611550999999992 10.4952105, -75.6516226 10.4106149,
    -75.6646807 10.3630263, -75.6683191 10.2948524, -75.7088108 10.187259300000001,
    -75.7082692 10.1664978, -75.7128402 10.1365786, -75.6999583 10.1141391,
    -75.6554949 10.1199567, -75.6172647 10.1000106, -75.6226819 10.0399524,
    -75.6260212 10.024369, -75.628247290380884 10), 
    (-75.8133381 10.1704573, -75.8103872 10.1575807, -75.8028758 10.1500693, 
      -75.8001932 10.1460454, -75.7929501 10.1455089, -75.7760496 10.1452406, 
      -75.7688065 10.1489963, -75.7591491 10.1559711, -75.7497599 10.1575807, 
      -75.7352795 10.1623423, -75.7265531 10.1669133, -75.7199866 10.1796414, 
      -75.7181049 10.1881626, -75.7256163 10.1935278, -75.7360785 10.1988931, 
      -75.7470773 10.1988931, -75.7548569 10.1962105, -75.7634413 10.1884308, 
      -75.7647826 10.1758225, -75.7663922 10.169115999999999, -75.7779274 10.1675064, 
      -75.7851705 10.1766273, -75.7964375 10.1849434, -75.8068998 10.1836021, 
      -75.8133381 10.1704573)), 
  POLYGON((-65.941795159426221 10.5, -65 10.5, -65 10.2064641,
    -65.0725608 10.1729443, -65.2401586 10.1934285, -65.5511457 10.2809518,
    -65.7820582 10.3684751, -65.8979095 10.4475282, -65.941795159426221 10.5)), 
  POLYGON((-65 10.2064641, -65 10.5, -64.299334152286434 10.5,
    -64.2848511 10.4671718, -64.4356891 10.43179, -64.5436966 10.3591643,
    -64.6405309 10.3610265, -64.716881 10.3424045, -64.8081287 10.3256448,
    -64.8015991 10.2572751, -64.772747 10.1952909, -64.7988178 10.1543225,
    -64.8565459 10.1654957, -64.8969606 10.1896883, -65 10.2064641)), 
  POLYGON((-62.272928364165637 10.5, -62 10.5, -62 10,
    -62.0829471 10.0505919, -62.1801821 10.089119, -62.2462286 10.1496616,
    -62.2865903 10.1074652, -62.3232827 10.0340803, -62.3691483 10.0432534,
    -62.4186831 10.1680078, -62.4608794 10.2597389, -62.536099 10.2964314,
    -62.5837992 10.3147776, -62.6259955 10.395501, -62.6700264 10.4487051,
    -62.6498456 10.4982399, -62.5342643 10.4798937, -62.3985022 10.4523743,
    -62.3306212 10.4560436, -62.272928364165637 10.5)), 
  POLYGON((-61 10.092982720726486, -61 10, -62 10,
    -62 10.5, -61.530669051024127 10.5, -61.5287448 10.4781914,
    -61.5405069 10.4605483, -61.5463879 10.4237918, -61.5655013 10.385565,
    -61.6301927 10.3091115, -61.8330887 10.1959014, -61.9698229 10.0797508,
    -61.9712932 10.0518159, -61.9639419 10.0341728, -61.9315962 10.0135891,
    -61.8360292 10.0194702, -61.8110348 10.0444646, -61.7478136 10.0327025,
    -61.6860626 10.0488754, -61.6213712 10.0194702, -61.5875552 10.0179999,
    -61.5552095 10.0238809, -61.5140422 10.035643, -61.4611128 10.0268215,
    -61.4096537 10.0224107, -61.3787782 10.0282917, -61.3449622 10.0282917,
    -61.2832113 10.0415241, -61.2405737 10.0444646, -61.2038172 10.029762,
    -61.146477 10.029762, -61.0935477 10.0444646, -61.0479696 10.0621077,
    -61 10.092982720726486)), 
  POLYGON((-60.918101877048876 10.5, -60 10.5, -60 10,
    -61 10, -61 10.092982720726486, -60.9999991 10.0929833,
    -60.9332892 10.1194478, -60.9185866 10.1356207, -60.9068245 10.1885501,
    -60.9053542 10.2400092, -60.8950624 10.2605928, -60.9024137 10.3017601,
    -60.9024137 10.3149925, -60.8685977 10.3502787, -60.8671275 10.385565,
    -60.8803598 10.4384944, -60.918101877048876 10.5)), 
  POLYGON((-60 10, -60 10.5, -59 10.5,
    -59 10, -60 10)), 
  POLYGON((-59 10, -59 10.5, -58 10.5,
    -58 10, -59 10)), 
  POLYGON((-58 10, -58 10.5, -57 10.5,
    -57 10, -58 10)), 
  POLYGON((-57 10, -57 10.5, -56 10.5,
    -56 10, -57 10)), 
  POLYGON((-56 10, -56 10.5, -55 10.5,
    -55 10, -56 10)), 
  POLYGON((-83.681103999999991 11, -83 11, -83 10.5,
    -83.372826949887909 10.5, -83.5074162 10.7925257, -83.5415599 10.8712048,
    -83.5697656 10.9394923, -83.6410222 10.9855122, -83.681103999999991 11)), 
  POLYGON((-83 10.5, -83 11, -82 11,
    -82 10.5, -83 10.5)), 
  POLYGON((-82 10.5, -82 11, -81 11,
    -81 10.5, -82 10.5)), 
  POLYGON((-81 10.5, -81 11, -80 11,
    -80 10.5, -81 10.5)), 
  POLYGON((-80 10.5, -80 11, -79 11,
    -79 10.5, -80 10.5)), 
  POLYGON((-79 10.5, -79 11, -78 11,
    -78 10.5, -79 10.5)), 
  POLYGON((-78 10.5, -78 11, -77 11,
    -77 10.5, -78 10.5)), 
  POLYGON((-77 10.5, -77 11, -76 11,
    -76 10.5, -77 10.5)), 
  POLYGON((-75.610231445087209 10.5, -76 10.5, -76 11,
    -75.0850546 11, -75.1429357 10.9393376, -75.2787339 10.865872999999999,
    -75.3399545 10.821349, -75.3755737 10.7801643, -75.4512645 10.7233961,
    -75.5814973 10.6042944, -75.610231445087209 10.5)), 
  POLYGON((-68.242115866498253 11, -68 11, -68 10.5584192,
    -68.0781484 10.5826277, -68.1693961 10.6850486, -68.2215376 10.7632609,
    -68.2010534 10.8284378, -68.1842937 10.8898904, -68.215951 10.9532051,
    -68.242115866498253 11)), 
  POLYGON((-68 10.5584192, -68 11, -67 11,
    -67 10.6906354, -67.208502 10.6198718, -67.4096193 10.6142852,
    -67.6423941 10.582627800000001, -67.8137163 10.5733168, -67.9179994 10.5528326,
    -68 10.5584192)), 
  POLYGON((-67 10.6906354, -67 11, -66 11,
    -66 10.5695925, -66.0409038 10.6440804, -66.1526357 10.7092574,
    -66.2643675 10.7297415, -66.3760994 10.7167062, -66.5064533 10.7260171,
    -66.767161 10.714844, -66.9496564 10.6943598, -67 10.6906354)), 
  POLYGON((-66 10.5695925, -66 11, -65 11,
    -65 10.5, -65.941795159426221 10.5, -66 10.5695925), 
    (-65.4787417 10.9229522, -65.4567094 10.8922507, -65.4334333 10.8431122, 
      -65.3849414 10.8334138, -65.2853713 10.8411725, -65.2200689 10.8625089, 
      -65.1806288 10.8838454, -65.1515337 10.9051818, -65.1398956 10.9446219, 
      -65.1450681 10.9659584, -65.2058446 10.9847086, -65.2795523 10.9885879, 
      -65.3332167 10.9860017, -65.3907604 10.9821223, -65.447011 10.9775964, 
      -65.4702872 10.9575531, -65.4787417 10.9229522)), 
  POLYGON((-64.299334152286434 10.5, -65 10.5, -65 11,
    -64.4598450896023 11, -64.4505867 10.9569299, -64.377961 10.9103749,
    -64.2308473 10.8954774, -64.2401583 10.8377492, -64.2587803 10.7539503,
    -64.161946 10.7278795, -64.1973278 10.6999466, -64.3258194 10.6999466,
    -64.3630634 10.593801299999999, -64.3127841 10.5304865, -64.299334152286434 10.5)), 
  POLYGON((-63.718742999999996 11, -63 11, -63 10.753949666924289,
    -63.0800091 10.750226, -63.1395994 10.7632614, -63.2382959 10.7390528,
    -63.2855135 10.7270549, -63.370512 10.731603999999999, -63.4654841 10.7185686,
    -63.5362476 10.6813247, -63.7038454 10.6999467, -63.7839199 10.7334662,
    -63.8397859 10.7856078, -63.8341993 10.8489225, -63.807899 10.8928641,
    -63.755987 10.9420324, -63.718742999999996 11)), 
  GEOMETRYCOLLECTION(
  POLYGON((-62 10.5826326, -62 10.5, -62.272928364165637 10.5,
    -62.253567 10.5147515, -62.2278823 10.5477747, -62.0774433 10.5642863,
    -62 10.5826326)), 
    POLYGON((-63 10.753949666924289, -63 11, -62 11,
      -62 10.75532804311367, -62.0000013 10.7553279, -62.0762844 10.7390529,
      -62.1619455 10.7316041, -62.2550554 10.7390529, -62.3500275 10.7558126,
      -62.5045899 10.768848, -62.608872999999996 10.7856078, -62.6889475 10.7893322,
      -62.7652977 10.7949188, -62.8546832 10.7483639, -62.9422065 10.7446395,
      -62.9999821 10.7539505, -63 10.753949666924289))), 
  POLYGON((-61.530669051024127 10.5, -62 10.5, -62 10.5826326,
    -61.8397641 10.6076759, -61.8368919 10.6141141, -61.8322017 10.6319543,
    -61.8368783 10.6817424, -61.8219689 10.7394565, -61.8291831 10.7697565,
    -61.8686211 10.7760088, -61.9070972 10.764465999999999, -61.9345115 10.753404100000001,
    -61.9912637 10.7562898, -62 10.75532804311367, -62 11,
    -61 11, -61 10.8883941, -61.1200123 10.8663402,
    -61.2023469 10.831054, -61.2979139 10.831054, -61.3861295 10.8325242,
    -61.4125942 10.8060595, -61.4243563 10.7987082, -61.4478804 10.7957677,
    -61.4655236 10.803118999999999, -61.490517999999994 10.7840056, -61.5434474 10.7825354,
    -61.5875552 10.7840056, -61.6287225 10.7751841, -61.6787113 10.7266655,
    -61.7007652 10.707552100000001, -61.7419325 10.7119629, -61.772808 10.7060818,
    -61.7816296 10.6810874, -61.7698675 10.6649145, -61.7463433 10.6502119,
    -61.7007652 10.6443309, -61.6537169 10.6413904, -61.605731 10.6310274,
    -61.5478582 10.588460999999999, -61.5463879 10.5458234, -61.5331555 10.5281803,
    -61.530669051024127 10.5)), 
  POLYGON((-61 10.8883941, -61 11, -60 11,
    -60 10.5, -60.918101877048876 10.5, -60.9200568 10.5031858,
    -60.9288784 10.5472936, -60.9450513 10.6031635, -60.9303487 10.6443308,
    -60.8391925 10.7237249, -60.8215494 10.8075298, -60.8391925 10.8795725,
    -60.9230268 10.9060463, -61 10.8883941)), 
  POLYGON((-60 10.5, -60 11, -59 11,
    -59 10.5, -60 10.5)), 
  POLYGON((-59 10.5, -59 11, -58 11,
    -58 10.5, -59 10.5)), 
  POLYGON((-58 10.5, -58 11, -57 11,
    -57 10.5, -58 10.5)), 
  POLYGON((-57 10.5, -57 11, -56 11,
    -56 10.5, -57 10.5)), 
  POLYGON((-56 10.5, -56 11, -55 11,
    -55 10.5, -56 10.5)), 
  POLYGON((-83.658957799752471 11.5, -83 11.5, -83 11,
    -83.681103999999991 11, -83.7612676 11.1517774, -83.7894734 11.3210118,
    -83.7271238 11.4471953, -83.658957799752471 11.5)), 
  POLYGON((-83 11, -83 11.5, -82 11.5,
    -82 11, -83 11)), 
  POLYGON((-82 11, -82 11.5, -81 11.5,
    -81 11, -82 11)), 
  POLYGON((-81 11, -81 11.5, -80 11.5,
    -80 11, -81 11)), 
  POLYGON((-80 11, -80 11.5, -79 11.5,
    -79 11, -80 11)), 
  POLYGON((-79 11, -79 11.5, -78 11.5,
    -78 11, -79 11)), 
  POLYGON((-78 11, -78 11.5, -77 11.5,
    -77 11, -78 11)), 
  POLYGON((-77 11, -77 11.5, -76 11.5,
    -76 11, -77 11)), 
  POLYGON((-75.0850546 11, -76 11, -76 11.5,
    -75 11.5, -75 11.1378938, -75.0850546 11)), 
  POLYGON((-75 11.1378938, -75 11.5, -74 11.5,
    -74 11.4957227, -74.2599435 11.4193323, -74.3363339 11.3268596,
    -74.3524161 11.1821198, -74.3725189 11.0856266, -74.4810737 11.1338732,
    -74.6499368 11.198202, -74.8308615 11.2585103, -74.9635397 11.2062431,
    -75 11.1378938)), 
  POLYGON((-74 11.4957227, -74 11.5, -73.24427473128371 11.5,
    -73.2909911 11.4715994, -73.5201624 11.4233528, -73.890053 11.4635583,
    -74 11.4957227)), 
  POLYGON((-71.864637225180985 11.5, -71 11.5, -71 11.1971526,
    -71.1861563 11.0984561, -71.3574786 11.0425901, -71.4692104 11.0556255,
    -71.6125997 11.1077671, -71.7820597 11.2530185, -71.8726125 11.4218745,
    -71.864637225180985 11.5)), 
  POLYGON((-71 11.1971526, -71 11.5, -70.21604026071428 11.5,
    -70.2289866 11.4597225, -70.3612026 11.39827, -70.5436981 11.3535772,
    -70.6517055 11.2995735, -70.8025436 11.2735027, -70.9366219 11.2232234,
    -71 11.1971526)), 
  POLYGON((-68.860745177833209 11.5, -68 11.5, -68 11,
    -68.242115866498253 11, -68.3127852 11.1263895, -68.338855999999993 11.2139128,
    -68.4170683 11.2586056, -68.5679064 11.3591643, -68.7410908 11.4634474,
    -68.860745177833209 11.5)), 
  POLYGON((-68 11, -68 11.5, -67 11.5,
    -67 11, -68 11)), 
  POLYGON((-67 11, -67 11.5, -66 11.5,
    -66 11, -67 11)), 
  POLYGON((-66 11, -66 11.5, -65 11.5,
    -65 11, -66 11)), 
  POLYGON((-64.4598450896023 11, -65 11, -65 11.5,
    -64 11.5, -64 11.146854980717478, -64.0427654 11.079835,
    -64.076285 11.0574886, -64.1414619 11.1375631, -64.3109219 11.1748071,
    -64.4487246 11.0947326, -64.4599007 11.0002587, -64.4598450896023 11)), 
  POLYGON((-64 11.146854980717478, -64 11.5, -63 11.5,
    -63 11, -63.718742999999996 11, -63.7616593 11.0673351,
    -63.7876443 11.1319766, -63.8174395 11.1897047, -63.8770298 11.2362597,
    -63.9291713 11.2139133, -63.9757263 11.1710827, -63.9999878 11.1468741,
    -64 11.146854980717478), 
    (-63.7468192 11.2117939, -63.7593615 11.1995184, -63.7481535 11.1736331, 
      -63.7324355 11.1736404, -63.7254705 11.1744337, -63.7070572 11.1856417, 
      -63.716930999999995 11.2128613, -63.7468192 11.2117939), 
    (-63.1055824 11.3892058, -63.1106175 11.394674, -63.1283623 11.394674, 
      -63.1501575 11.3919737, -63.1511219 11.3813654, -63.1461071 11.3721073, 
      -63.1349201 11.3628491, -63.124119 11.3564841, -63.1191041 11.348768999999999, 
      -63.1069528 11.3479975, -63.1007807 11.3505049, -63.0963445 11.3543625, 
      -63.0832288 11.3591844, -63.0735849 11.3663209, -63.0724543 11.3714827, 
      -63.0724277 11.3761577, -63.0735849 11.3796295, -63.080779 11.3821293, 
      -63.0888223 11.3831013, -63.0915226 11.380400999999999, -63.0994306 11.3690212, 
      -63.1025166 11.3692141, -63.1086887 11.3736503, -63.1079172 11.3829084, 
      -63.1055824 11.3892058)), 
  POLYGON((-63 11, -63 11.5, -62 11.5,
    -62 11, -63 11)), 
  POLYGON((-62 11, -62 11.5, -61 11.5,
    -61 11, -62 11)), 
  POLYGON((-61 11, -61 11.5, -60 11.5,
    -60 11, -61 11), 
    (-60.8761739 11.0566268, -60.8554785 11.0334967, -60.8238267 11.0322793, 
      -60.7800011 11.0237577, -60.7459145 11.0176708, -60.6619155 11.070018, 
      -60.5961771 11.0870613, -60.5608731 11.0943656, -60.5292213 11.1296695, 
      -60.5206997 11.134539, -60.512178 11.1381911, -60.512178 11.1430607, 
      -60.5000043 11.1491475, -60.4768741 11.1613213, -60.4306137 11.1564518, 
      -60.4099183 11.2112338, -60.3721796 11.3013197, -60.4196573 11.334188900000001, 
      -60.4269616 11.3475801, -60.4367006 11.3621886, -60.4525265 11.3816667, 
      -60.4695698 11.3877535, -60.5645253 11.447405, -60.6339158 11.4242749, 
      -60.7020889 11.3974926, -60.7678273 11.3841014, -60.8554785 11.3293194, 
      -60.8810434 11.2647984, -60.9041736 11.2124512, -60.907825700000004 11.179582, 
      -60.9163474 11.1747125, -60.9236516 11.1515823, -60.939477600000004 11.1260174, 
      -60.9163474 11.0809744, -60.8980867 11.0809744, -60.8761739 11.0566268)), 
  POLYGON((-60 11, -60 11.5, -59 11.5,
    -59 11, -60 11)), 
  POLYGON((-59 11, -59 11.5, -58 11.5,
    -58 11, -59 11)), 
  POLYGON((-58 11, -58 11.5, -57 11.5,
    -57 11, -58 11)), 
  POLYGON((-57 11, -57 11.5, -56 11.5,
    -56 11, -57 11)), 
  POLYGON((-56 11, -56 11.5, -55 11.5,
    -55 11, -56 11)), 
  POLYGON((-83.6143009 12, -83 12, -83 11.5,
    -83.658957799752471 11.5, -83.6217235 11.5288435, -83.5757036 11.588224,
    -83.5979713 11.7262836, -83.6365686 11.846529, -83.6736814 11.8999715,
    -83.6380531 11.9578674, -83.6143009 12)), 
  POLYGON((-83 11.5, -83 12, -82 12,
    -82 11.5, -83 11.5)), 
  POLYGON((-82 11.5, -82 12, -81 12,
    -81 11.5, -82 11.5)), 
  POLYGON((-81 11.5, -81 12, -80 12,
    -80 11.5, -81 11.5)), 
  POLYGON((-80 11.5, -80 12, -79 12,
    -79 11.5, -80 11.5)), 
  POLYGON((-79 11.5, -79 12, -78 12,
    -78 11.5, -79 11.5)), 
  POLYGON((-78 11.5, -78 12, -77 12,
    -77 11.5, -78 11.5)), 
  POLYGON((-77 11.5, -77 12, -76 12,
    -76 11.5, -77 11.5)), 
  POLYGON((-76 11.5, -76 12, -75 12,
    -75 11.5, -76 11.5)), 
  POLYGON((-75 11.5, -75 12, -74 12,
    -74 11.5, -75 11.5)), 
  POLYGON((-73.24427473128371 11.5, -74 11.5, -74 12,
    -73 12, -73 11.6485036, -73.24427473128371 11.5)), 
  POLYGON((-73 11.6485036, -73 12, -72.4386346 12,
    -72.8567718 11.7731407, -73 11.6485036)), 
  POLYGON((-71.0427671 12, -71 12, -71 11.5,
    -71.864637225180985 11.5, -71.8639964 11.5062775, -71.746678 11.558418999999999,
    -71.5064544 11.5975252, -71.3537542 11.6645643, -71.2606443 11.7707096,
    -71.1396014 11.9010634, -71.0427671 12)), 
  POLYGON((-70.21604026071428 11.5, -71 11.5, -71 12,
    -70.3258211 12, -70.3556161 11.8582329, -70.3342344 11.7822765,
    -70.2978879 11.6384936, -70.2811281 11.56773, -70.2122268 11.5118641,
    -70.21604026071428 11.5)), 
  POLYGON((-69.7504019107487 12, -69 12, -69 11.519313277593675,
    -69.0000002 11.5193133, -69.0651131 11.5416596, -69.2587817 11.6031122,
    -69.3574782 11.5770414, -69.402171 11.5379352, -69.534387 11.573317,
    -69.580942 11.552832800000001, -69.6889495 11.6813245, -69.7299178 11.8433357,
    -69.746203 11.9200219, -69.7504019107487 12)), 
  POLYGON((-69 11.519313277593675, -69 12, -68 12,
    -68 11.5, -68.860745177833209 11.5, -68.879935 11.5058622,
    -69 11.519313277593675), 
    (-68.6602802 11.9942459, -68.6572239 11.9884981, -68.6557545 11.9830406, 
      -68.652605999999992 11.9726503, -68.6468336 11.9680324, -68.6351839 11.9691869, 
      -68.6310908 11.9783177, -68.6313007 11.9835653, -68.6332948 11.9925912, 
      -68.6364433 11.9980487, -68.6422157 11.9987834, -68.6532357 11.9998329, 
      -68.6568041 11.9983636, -68.6602802 11.9942459)), 
  POLYGON((-68 11.5, -68 12, -67 12,
    -67 11.5, -68 11.5), 
    (-67.7092738 11.9650642, -67.6781114 11.9412246, -67.6513977 11.9424388, 
      -67.6240768 11.95276, -67.6101128 11.9624741, -67.587649 11.979473800000001, 
      -67.6234697 11.994652, -67.6750758 11.9958663, -67.6999681 11.9885807, 
      -67.7139321 11.9776524, -67.7092738 11.9650642)), 
  POLYGON((-67 11.5, -67 12, -66 12,
    -66 11.5, -67 11.5), 
    (-66.9183744 11.9279817, -66.9326429 11.9126329, -66.9595492 11.8756367, 
      -66.9651547 11.8588203, -66.9718813 11.8464882, -66.9741235 11.8341561, 
      -66.9741235 11.8229452, -66.9629125 11.8005232, -66.9595492 11.7960388, 
      -66.9315218 11.7994021, -66.9169475 11.8038865, -66.8922834 11.8150975, 
      -66.8642559 11.8285506, -66.8463184 11.8397616, -66.8384707 11.8431249, 
      -66.8295019 11.8408827, -66.8238964 11.8375194, -66.8104433 11.833034999999999, 
      -66.7667205 11.8397616, -66.7555095 11.844246, -66.738693 11.844246, 
      -66.7286031 11.8420038, -66.726361 11.8218241, -66.7510251 11.8016443, 
      -66.8059589 11.8016443, -66.8261386 11.7994021, -66.841833999999992 11.8005232, 
      -66.8620137 11.8005232, -66.8844357 11.7814646, -66.8889201 11.763527, 
      -66.8889201 11.7534371, -66.8541661 11.7029877, -66.8048378 11.7041088, 
      -66.7375719 11.706351, -66.6960913 11.7074721, -66.6310677 11.7186831, 
      -66.5828605 11.739984, -66.5559542 11.7736169, -66.5413799 11.8038865, 
      -66.5604385 11.8532148, -66.5940715 11.8913321, -66.5974348 11.9103908, 
      -66.6355521 11.9103908, -66.6523685 11.9182384, -66.6422787 11.9294494, 
      -66.6411576 11.9686878, -66.6904858 11.9720511, -66.703938999999991 11.9496292, 
      -66.703938999999991 11.9294494, -66.7476618 11.9272072, -66.7790525 11.913753999999999, 
      -66.7869002 11.901422, -66.8328652 11.9003009, -66.8339863 11.913753999999999, 
      -66.8799513 11.9384182, -66.9183744 11.9279817), 
    (-66.2252305 11.8363984, -66.2263516 11.8072499, -66.2229883 11.7803435, 
      -66.2140195 11.7568005, -66.1646912 11.7186832, -66.1433904 11.7130777, 
      -66.104152 11.7130777, -66.0772456 11.7220465, -66.0559448 11.7377418, 
      -66.0458549 11.7568005, -66.0290385 11.7769802, -66.0279174 11.7949178, 
      -66.0279174 11.8117342, -66.0256752 11.8263085, -66.0267963 11.8487305, 
      -66.0267963 11.8655469, -66.0436127 11.8879689, -66.0530137 11.8932323, 
      -66.0783667 11.8991798, -66.1108786 11.9025431, -66.1299372 11.8801212, 
      -66.1321794 11.8610625, -66.1456326 11.8476094, -66.1546014 11.8453672, 
      -66.1680545 11.8476094, -66.1949609 11.8464883, -66.2252305 11.8363984)), 
  POLYGON((-66 11.5, -66 12, -65 12,
    -65 11.5, -66 11.5)), 
  POLYGON((-65 11.5, -65 12, -64 12,
    -64 11.5, -65 11.5), 
    (-64.4088715 11.7166753, -64.3907255 11.7061459, -64.381109999999993 11.7073479, 
      -64.3763022 11.7151605, -64.3763022 11.7187663, -64.3781052 11.7373963, 
      -64.3835139 11.7536225, -64.3919274 11.7878777, -64.3991391 11.8047049, 
      -64.419572 11.8107145, -64.4305423 11.8092321, -64.4434704 11.7993297, 
      -64.4520244 11.7734545, -64.4412069 11.7560264, -64.4310003 11.7501341, 
      -64.4088715 11.7166753), 
    (-64.6401277 11.918287999999999, -64.6491422 11.9074706, -64.6503442 11.8924463, 
      -64.6563539 11.8846338, -64.6761858 11.8635998, -64.6791907 11.8485756, 
      -64.6773878 11.8245368, -64.6611616 11.7980942, -64.6323151 11.782468999999999, 
      -64.6064734 11.7734545, -64.5806318 11.7716516, -64.5487804 11.783069999999999, 
      -64.5392332 11.8085029, -64.5325542 11.8281427, -64.5301503 11.8467727, 
      -64.5271455 11.862998900000001, -64.537362 11.8834318, -64.5511843 11.9020619, 
      -64.5734201 11.9206919, -64.6184928 11.9206919, -64.6401277 11.918287999999999)), 
  POLYGON((-64 11.5, -64 12, -63 12,
    -63 11.5, -64 11.5)), 
  POLYGON((-63 11.5, -63 12, -62 12,
    -62 11.5, -63 11.5)), 
  POLYGON((-61.6118472 12, -61 12, -61 11.5,
    -62 11.5, -62 12, -61.8490889 12,
    -61.8392038 11.9238838, -61.8053492 11.9190475, -61.730467999999995 11.9041137,
    -61.6513875 11.9139987, -61.6118472 12)), 
  POLYGON((-61 11.5, -61 12, -60 12,
    -60 11.5, -61 11.5)), 
  POLYGON((-60 11.5, -60 12, -59 12,
    -59 11.5, -60 11.5)), 
  POLYGON((-59 11.5, -59 12, -58 12,
    -58 11.5, -59 11.5)), 
  POLYGON((-58 11.5, -58 12, -57 12,
    -57 11.5, -58 11.5)), 
  POLYGON((-57 11.5, -57 12, -56 12,
    -56 11.5, -57 11.5)), 
  POLYGON((-56 11.5, -56 12, -55 12,
    -55 11.5, -56 11.5)), 
  POLYGON((-83.425248088835289 12.5, -83 12.5, -83 12,
    -83.6143009 12, -83.6202387 12.1078031, -83.589064 12.2414092,
    -83.5653118 12.3037587, -83.4866326 12.3067277, -83.4257676 12.3200883,
    -83.4138915 12.4388493, -83.425248088835289 12.5), 
    (-83.0557021 12.1993792, -83.0617005 12.1973798, -83.1033556 12.1670549, 
      -83.1556743 12.1170688, -83.0550357 12.0747472, -83.0230446 12.0580852, 
      -83.0107147 12.1004068, -83.0030501 12.1557247, -83.0193789 12.1840502, 
      -83.0470379 12.1983795, -83.0557021 12.1993792)), 
  POLYGON((-83 12, -83 12.5, -82 12.5,
    -82 12, -83 12), 
    (-82.9985602 12.2845222, -82.99080099999999 12.2712208, -82.9871985 12.2590278, 
      -82.9697404 12.2601363, -82.9650294 12.2654014, -82.9605956 12.2698352, 
      -82.9575474 12.276486, -82.9653066 12.2997635, -82.9711259 12.3069684, 
      -82.9794393 12.3066913, -82.9885841 12.3075226, -82.9946806 12.2989321, 
      -82.9991144 12.2914501, -82.9985602 12.2845222)), 
  POLYGON((-81.640356329438291 12.5, -81 12.5, -81 12,
    -82 12, -82 12.5, -81.82024214725827 12.5,
    -81.8189186 12.4845589, -81.8065208 12.4622429, -81.7581695 12.4324883,
    -81.722216 12.433727999999999, -81.6713852 12.4473656, -81.6465896 12.4820793,
    -81.640356329438291 12.5)), 
  POLYGON((-81 12, -81 12.5, -80 12.5,
    -80 12, -81 12)), 
  POLYGON((-80 12, -80 12.5, -79 12.5,
    -79 12, -80 12)), 
  POLYGON((-79 12, -79 12.5, -78 12.5,
    -78 12, -79 12)), 
  POLYGON((-78 12, -78 12.5, -77 12.5,
    -77 12, -78 12)), 
  POLYGON((-77 12, -77 12.5, -76 12.5,
    -76 12, -77 12)), 
  POLYGON((-76 12, -76 12.5, -75 12.5,
    -75 12, -76 12)), 
  POLYGON((-75 12, -75 12.5, -74 12.5,
    -74 12, -75 12)), 
  POLYGON((-74 12, -74 12.5, -73 12.5,
    -73 12, -74 12)), 
  POLYGON((-72.4386346 12, -73 12, -73 12.5,
    -72 12.5, -72 12.4244698, -72.1571961 12.4164287,
    -72.2938948 12.3199355, -72.3984291 12.199319, -72.3944086 12.0867436,
    -72.4386346 12)), 
  GEOMETRYCOLLECTION(
  POLYGON((-71.228703226208651 12.5, -71 12.5, -71 12,
    -71.0427671 12, -71.0394834 12.235503999999999, -71.08773 12.4003465,
    -71.228703226208651 12.5)), 
    POLYGON((-72 12.4244698, -72 12.5, -71.937747906018146 12.5,
      -72 12.4244698))), 
  POLYGON((-70.3258211 12, -71 12, -71 12.5,
    -70.111512183597242 12.5, -70.1077686 12.4925128, -70.0792399 12.4639841,
    -70 12.3955152, -70 12.2641923, -70.1153927 12.2306728,
    -70.2196758 12.1654959, -70.2960259 12.1077677, -70.3258211 12)), 
  POLYGON((-69 12.1259438, -69 12, -69.7504019107487 12,
    -69.750402 12.0000017, -69.8006812 12.1040433, -69.8472362 12.1915666,
    -69.9105509 12.2306728, -70 12.2641923, -70 12.3955152,
    -69.9365963 12.355575, -69.8852446 12.355575, -69.8453044 12.3612808,
    -69.8167757 12.3841037, -69.784719 12.4201962, -69.831178975662482 12.5,
    -69 12.5, -69 12.324491712613035, -69.038947999999991 12.3526765,
    -69.0611538 12.3854008, -69.1020592 12.3947507, -69.1382897 12.4251375,
    -69.2049071 12.4239688, -69.2014009 12.3877383, -69.2177631 12.36787,
    -69.2130882 12.3211209, -69.1885449 12.2720345, -69.1651704 12.2159356,
    -69.1227052 12.1796125, -69 12.1259438)), 
  POLYGON((-69 12.324491712613035, -69 12.5, -68 12.5,
    -68 12, -69 12, -69 12.1259438,
    -68.9922549 12.0966569, -68.9699932 12.0710137, -68.9477874 12.0534828,
    -68.9127257 12.0359519, -68.8636392 12.0172523, -68.8297462 12.0149148,
    -68.7993593 12.0149148, -68.7607914 12.0137461, -68.7397543 12.0160836,
    -68.6469089 12.0155979, -68.6419305 12.0579149, -68.6867367 12.1201457,
    -68.7912845 12.2022904, -68.9207246 12.262032, -68.9431278 12.262032,
    -68.9996997 12.3242744, -69 12.324491712613035), 
    (-68.4187492 12.3295904, -68.4400595 12.3141996, -68.446570999999992 12.3005847, 
      -68.4489388 12.2875618, -68.4483468 12.2632918, -68.4501227 12.2508608, 
      -68.4477549 12.2301425, -68.445979 12.2188954, -68.4400595 12.2082402, 
      -68.421708999999993 12.1964012, -68.408686 12.1916656, -68.3773126 12.1839702, 
      -68.3542265 12.1798265, -68.3412035 12.176866799999999, -68.3376518 12.173907, 
      -68.3329162 12.1668036, -68.3323242 12.1561485, -68.334692 12.149637, 
      -68.3370598 12.1395738, -68.335284 12.1312865, -68.3317323 12.1271428, 
      -68.3287725 12.1224072, -68.3258128 12.1188555, -68.3216691 12.1111601, 
      -68.3181174 12.1052406, -68.3181174 12.0868901, -68.3151576 12.0744591, 
      -68.3086462 12.0572925, -68.3098301 12.0490052, -68.3121979 12.0430857, 
      -68.3074623 12.0330225, -68.3033186 12.025919, -68.2902956 12.0075685, 
      -68.2778647 12.002241, -68.2624739 12.002241, -68.2186695 12.0057927, 
      -68.2044088 12.0142611, -68.1932156 12.0525569, -68.2044627 12.0626201, 
      -68.2032788 12.0726833, -68.1825605 12.0945855, -68.16421 12.1218153, 
      -68.1695375 12.1437175, -68.176641 12.162068, -68.1754571 12.1774588, 
      -68.1665778 12.199361, -68.1612502 12.2177115, -68.1736812 12.2407976, 
      -68.1932156 12.2502689, -68.2305086 12.2508608, -68.2559625 12.2585562, 
      -68.2849682 12.273354999999999, -68.3086463 12.2893377, -68.3660655 12.3319582, 
      -68.3808643 12.3355099, -68.4003987 12.334918, -68.4187492 12.3295904)), 
  POLYGON((-68 12, -68 12.5, -67 12.5,
    -67 12, -68 12)), 
  POLYGON((-67 12, -67 12.5, -66 12.5,
    -66 12, -67 12)), 
  POLYGON((-66 12, -66 12.5, -65 12.5,
    -65 12, -66 12)), 
  POLYGON((-65 12, -65 12.5, -64 12.5,
    -64 12, -65 12)), 
  POLYGON((-64 12, -64 12.5, -63 12.5,
    -63 12, -64 12)), 
  POLYGON((-63 12, -63 12.5, -62 12.5,
    -62 12, -63 12)), 
  GEOMETRYCOLLECTION(
  POLYGON((-61.8490889 12, -62 12, -62 12.5,
    -61.572764012542812 12.5, -61.5970196 12.418137399999999, -61.592076999999996 12.3736546,
    -61.6316173 12.3390568, -61.6909278 12.3044591, -61.7897785 12.2599763,
    -61.8688591 12.2056084, -61.8787441 12.082044999999999, -61.8490889 12)), 
    POLYGON((-61.3449502 12.5, -61 12.5, -61 12,
      -61.6118472 12, -61.5772494 12.0573323, -61.5278241 12.082044999999999,
      -61.5228815 12.1067576, -61.5228815 12.1512405, -61.5031114 12.2599763,
      -61.453686 12.358827, -61.3746054 12.4280225, -61.3449502 12.4873329,
      -61.3449502 12.5))), 
  POLYGON((-61 12, -61 12.5, -60 12.5,
    -60 12, -61 12)), 
  POLYGON((-60 12, -60 12.5, -59 12.5,
    -59 12, -60 12)), 
  POLYGON((-59 12, -59 12.5, -58 12.5,
    -58 12, -59 12)), 
  POLYGON((-58 12, -58 12.5, -57 12.5,
    -57 12, -58 12)), 
  POLYGON((-57 12, -57 12.5, -56 12.5,
    -56 12, -57 12)), 
  POLYGON((-56 12, -56 12.5, -55 12.5,
    -55 12, -56 12)), 
  POLYGON((-83.467333801760645 13, -83 13, -83 12.5,
    -83.425248088835289 12.5, -83.4331902 12.5427651, -83.473272 12.6318359,
    -83.4480353 12.8248225, -83.4465508 12.8485747, -83.4465507 12.9153777,
    -83.467333801760645 13)), 
  POLYGON((-83 12.5, -83 13, -82 13,
    -82 12.5, -83 12.5)), 
  POLYGON((-81.82024214725827 12.5, -82 12.5, -82 13,
    -81 13, -81 12.5, -81.640356329438291 12.5,
    -81.6267532 12.5391091, -81.6304725 12.6271333, -81.6825431 12.6345719,
    -81.7246956 12.6345719, -81.7564432 12.6106976, -81.8040413 12.5750626,
    -81.822638 12.5279511, -81.82024214725827 12.5)), 
  POLYGON((-81 12.5, -81 13, -80 13,
    -80 12.5, -81 12.5)), 
  POLYGON((-80 12.5, -80 13, -79 13,
    -79 12.5, -80 12.5)), 
  POLYGON((-79 12.5, -79 13, -78 13,
    -78 12.5, -79 12.5)), 
  POLYGON((-78 12.5, -78 13, -77 13,
    -77 12.5, -78 12.5)), 
  POLYGON((-77 12.5, -77 13, -76 13,
    -76 12.5, -77 12.5)), 
  POLYGON((-76 12.5, -76 13, -75 13,
    -75 12.5, -76 12.5)), 
  POLYGON((-75 12.5, -75 13, -74 13,
    -74 12.5, -75 12.5)), 
  POLYGON((-74 12.5, -74 13, -73 13,
    -73 12.5, -74 12.5)), 
  POLYGON((-73 12.5, -73 13, -72 13,
    -72 12.5, -73 12.5)), 
  POLYGON((-71.937747906018146 12.5, -72 12.5, -72 13,
    -71 13, -71 12.5, -71.228703226208651 12.5,
    -71.3209218 12.565189, -71.6144219 12.6094151, -71.9039015 12.5410657,
    -71.937747906018146 12.5)), 
  POLYGON((-70.111512183597242 12.5, -71 12.5, -71 13,
    -70 13, -70 12.672333960037857, -70.0008431 12.6731686,
    -70.0343052 12.7029515, -70.0820927 12.7121839, -70.1248858 12.6950667,
    -70.1305915 12.6836552, -70.1277387 12.6465679, -70.1277387 12.6123334,
    -70.1334444 12.5895104, -70.1305915 12.5381588, -70.111512183597242 12.5)), 
  POLYGON((-70 12.672333960037857, -70 13, -69 13,
    -69 12.5, -69.831178975662482 12.5, -69.8567159 12.5438645,
    -69.9023619 12.5809518, -69.9451549 12.6180392, -70 12.672333960037857)), 
  POLYGON((-69 12.5, -69 13, -68 13,
    -68 12.5, -69 12.5)), 
  POLYGON((-68 12.5, -68 13, -67 13,
    -67 12.5, -68 12.5)), 
  POLYGON((-67 12.5, -67 13, -66 13,
    -66 12.5, -67 12.5)), 
  POLYGON((-66 12.5, -66 13, -65 13,
    -65 12.5, -66 12.5)), 
  POLYGON((-65 12.5, -65 13, -64 13,
    -64 12.5, -65 12.5)), 
  POLYGON((-64 12.5, -64 13, -63 13,
    -63 12.5, -64 12.5)), 
  POLYGON((-63 12.5, -63 13, -62 13,
    -62 12.5, -63 12.5)), 
  GEOMETRYCOLLECTION(
  POLYGON((-61.572764012542812 12.5, -62 12.5, -62 13,
    -61.5179391 13, -61.3696629 12.9371037, -61.3498928 12.902505999999999,
    -61.3103525 12.8728507, -61.3202375 12.7789426, -61.3844905 12.7344597,
    -61.4388584 12.7048045, -61.4882838 12.635608999999999, -61.5080539 12.5960687,
    -61.5574793 12.5515859, -61.572764012542812 12.5)), 
    POLYGON((-61.087983540697891 13, -61 13, -61 12.5,
      -61.3449502 12.5, -61.3449502 12.571356, -61.3251801 12.6207814,
      -61.2955249 12.6454941, -61.2856398 12.6751493, -61.2559846 12.7146896,
      -61.1867891 12.7987127, -61.1472488 12.8481381, -61.1225361 12.9173336,
      -61.087983540697891 13))), 
  POLYGON((-61 12.5, -61 13, -60 13,
    -60 12.5, -61 12.5)), 
  POLYGON((-60 12.5, -60 13, -59 13,
    -59 12.5, -60 12.5)), 
  POLYGON((-59 12.5, -59 13, -58 13,
    -58 12.5, -59 12.5)), 
  POLYGON((-58 12.5, -58 13, -57 13,
    -57 12.5, -58 12.5)), 
  POLYGON((-57 12.5, -57 13, -56 13,
    -56 12.5, -57 12.5)), 
  POLYGON((-56 12.5, -56 13, -55 13,
    -55 12.5, -56 12.5)), 
  POLYGON((-83.47306255690836 13.5, -83 13.5, -83 13,
    -83.467333801760645 13, -83.4673339 13.0000004, -83.4970241 13.2716605,
    -83.47306255690836 13.5)), 
  POLYGON((-83 13, -83 13.5, -82 13.5,
    -82 13, -83 13)), 
  POLYGON((-82 13, -82 13.5, -81 13.5,
    -81 13, -82 13), 
    (-81.3993878 13.4055187, -81.4025966 13.3900094, -81.3972485 13.3777088, 
      -81.4025966 13.3712912, -81.418106 13.3648735, -81.4517987 13.3472249, 
      -81.4624948 13.3327852, -81.4571468 13.3129974, -81.4304065 13.2990924, 
      -81.3919005 13.2910703, -81.3568231 13.2937071, -81.3309326 13.3124625, 
      -81.3202365 13.3343896, -81.3202365 13.3589906, -81.3261194 13.3755696, 
      -81.3346763 13.3932182, -81.3598121 13.4007055, -81.382274 13.4060535, 
      -81.3993878 13.4055187)), 
  POLYGON((-81 13, -81 13.5, -80 13.5,
    -80 13, -81 13)), 
  POLYGON((-80 13, -80 13.5, -79 13.5,
    -79 13, -80 13)), 
  POLYGON((-79 13, -79 13.5, -78 13.5,
    -78 13, -79 13)), 
  POLYGON((-78 13, -78 13.5, -77 13.5,
    -77 13, -78 13)), 
  POLYGON((-77 13, -77 13.5, -76 13.5,
    -76 13, -77 13)), 
  POLYGON((-76 13, -76 13.5, -75 13.5,
    -75 13, -76 13)), 
  POLYGON((-75 13, -75 13.5, -74 13.5,
    -74 13, -75 13)), 
  POLYGON((-74 13, -74 13.5, -73 13.5,
    -73 13, -74 13)), 
  POLYGON((-73 13, -73 13.5, -72 13.5,
    -72 13, -73 13)), 
  POLYGON((-72 13, -72 13.5, -71 13.5,
    -71 13, -72 13)), 
  POLYGON((-71 13, -71 13.5, -70 13.5,
    -70 13, -71 13)), 
  POLYGON((-70 13, -70 13.5, -69 13.5,
    -69 13, -70 13)), 
  POLYGON((-69 13, -69 13.5, -68 13.5,
    -68 13, -69 13)), 
  POLYGON((-68 13, -68 13.5, -67 13.5,
    -67 13, -68 13)), 
  POLYGON((-67 13, -67 13.5, -66 13.5,
    -66 13, -67 13)), 
  POLYGON((-66 13, -66 13.5, -65 13.5,
    -65 13, -66 13)), 
  POLYGON((-65 13, -65 13.5, -64 13.5,
    -64 13, -65 13)), 
  POLYGON((-64 13, -64 13.5, -63 13.5,
    -63 13, -64 13)), 
  POLYGON((-63 13, -63 13.5, -62 13.5,
    -62 13, -63 13)), 
  POLYGON((-61.5179391 13, -62 13, -62 13.5,
    -61 13.5, -61 13, -61.087983540697891 13,
    -61.0879333 13.0001202, -61.1126511 13.0754949, -61.1122546 13.0802528,
    -61.1107932 13.1471014, -61.1002905 13.1739415, -61.0827862 13.2112842,
    -61.0699496 13.2404581, -61.0722835 13.273133, -61.0722835 13.3606549,
    -61.1131271 13.4190029, -61.1656403 13.4481769, -61.2194529 13.4555851,
    -61.2426596 13.4575126, -61.2811692 13.444676, -61.3290146 13.4038324,
    -61.3535207 13.3548201, -61.4188704 13.2929713, -61.4527122 13.2241207,
    -61.4678827 13.1681067, -61.4678827 13.1412666, -61.4830532 13.1120926,
    -61.5203959 13.0537447, -61.5179391 13)), 
  POLYGON((-61 13, -61 13.5, -60 13.5,
    -60 13, -61 13)), 
  POLYGON((-60 13, -60 13.5, -59 13.5,
    -59 13, -60 13), 
    (-59.6580476 13.0849171, -59.6386794 13.0655489, -59.622979 13.0486408, 
      -59.6072785 13.0305249, -59.5758777 13.0184476, -59.5118683 13.0184476, 
      -59.442074 13.021127, -59.3975912 13.0408972, -59.3481659 13.0754949, 
      -59.3382808 13.1298628, -59.3629935 13.2040009, -59.4074763 13.2682538, 
      -59.4816143 13.3226217, -59.5112695 13.3720471, -59.57058 13.4165299, 
      -59.6809497 13.3843881, -59.6833652 13.3095092, -59.6845729 13.2805238, 
      -59.679742 13.2261762, -59.6749111 13.1694132, -59.6700802 13.1186888, 
      -59.6628339 13.0897034, -59.6580476 13.0849171)), 
  POLYGON((-59 13, -59 13.5, -58 13.5,
    -58 13, -59 13)), 
  POLYGON((-58 13, -58 13.5, -57 13.5,
    -57 13, -58 13)), 
  POLYGON((-57 13, -57 13.5, -56 13.5,
    -56 13, -57 13)), 
  POLYGON((-56 13, -56 13.5, -55 13.5,
    -55 13, -56 13)), 
  POLYGON((-83.318881786712865 14, -83 14, -83 13.5,
    -83.47306255690836 13.5, -83.4717874 13.5121515, -83.4317055 13.7660031,
    -83.3990462 13.8535893, -83.3188825 13.9999994, -83.318881786712865 14)), 
  POLYGON((-83 13.5, -83 14, -82 14,
    -82 13.5, -83 13.5)), 
  POLYGON((-82 13.5, -82 14, -81 14,
    -81 13.5, -82 13.5)), 
  POLYGON((-81 13.5, -81 14, -80 14,
    -80 13.5, -81 13.5)), 
  POLYGON((-80 13.5, -80 14, -79 14,
    -79 13.5, -80 13.5)), 
  POLYGON((-79 13.5, -79 14, -78 14,
    -78 13.5, -79 13.5)), 
  POLYGON((-78 13.5, -78 14, -77 14,
    -77 13.5, -78 13.5)), 
  POLYGON((-77 13.5, -77 14, -76 14,
    -76 13.5, -77 13.5)), 
  POLYGON((-76 13.5, -76 14, -75 14,
    -75 13.5, -76 13.5)), 
  POLYGON((-75 13.5, -75 14, -74 14,
    -74 13.5, -75 13.5)), 
  POLYGON((-74 13.5, -74 14, -73 14,
    -73 13.5, -74 13.5)), 
  POLYGON((-73 13.5, -73 14, -72 14,
    -72 13.5, -73 13.5)), 
  POLYGON((-72 13.5, -72 14, -71 14,
    -71 13.5, -72 13.5)), 
  POLYGON((-71 13.5, -71 14, -70 14,
    -70 13.5, -71 13.5)), 
  POLYGON((-70 13.5, -70 14, -69 14,
    -69 13.5, -70 13.5)), 
  POLYGON((-69 13.5, -69 14, -68 14,
    -68 13.5, -69 13.5)), 
  POLYGON((-68 13.5, -68 14, -67 14,
    -67 13.5, -68 13.5)), 
  POLYGON((-67 13.5, -67 14, -66 14,
    -66 13.5, -67 13.5)), 
  POLYGON((-66 13.5, -66 14, -65 14,
    -65 13.5, -66 13.5)), 
  POLYGON((-65 13.5, -65 14, -64 14,
    -64 13.5, -65 13.5)), 
  POLYGON((-64 13.5, -64 14, -63 14,
    -63 13.5, -64 13.5)), 
  POLYGON((-63 13.5, -63 14, -62 14,
    -62 13.5, -63 13.5)), 
  POLYGON((-61 13.6438863, -61 13.5, -62 13.5,
    -62 14, -61.1175934 14, -61.122536 13.9256109,
    -61.122536 13.9206684, -61.132420999999994 13.8860706, -61.132420999999994 13.8663005,
    -61.1423061 13.8415878, -61.1472486 13.7723923, -61.1423061 13.7130818,
    -61.1367379 13.7079419, -61.0780531 13.6537714, -61 13.6438863)), 
  POLYGON((-60.7716159 14, -60 14, -60 13.5,
    -61 13.5, -61 13.6438863, -60.9050644 13.6537714,
    -60.8852942 13.668598999999999, -60.855639 13.7279094, -60.8358689 13.8218176,
    -60.8062137 13.9206684, -60.7716159 14)), 
  POLYGON((-60 13.5, -60 14, -59 14,
    -59 13.5, -60 13.5)), 
  POLYGON((-59 13.5, -59 14, -58 14,
    -58 13.5, -59 13.5)), 
  POLYGON((-58 13.5, -58 14, -57 14,
    -57 13.5, -58 13.5)), 
  POLYGON((-57 13.5, -57 14, -56 14,
    -56 13.5, -57 13.5)), 
  POLYGON((-56 13.5, -56 14, -55 14,
    -55 13.5, -56 13.5)), 
  POLYGON((-83.157781366666669 14.5, -83 14.5, -83 14,
    -83.318881786712865 14, -83.2807321 14.0320906, -83.149648 14.3033963,
    -83.1466789 14.4666926, -83.157781366666669 14.5)), 
  POLYGON((-83 14, -83 14.5, -82 14.5,
    -82 14, -83 14), 
    (-82.8257002 14.270157, -82.8099393 14.2699656, -82.7618281 14.279416, 
      -82.7412091 14.292302999999999, -82.7102804 14.3077673, -82.6595919 14.3051899, 
      -82.6441276 14.2940212, -82.5925799 14.2880073, -82.5650878 14.2819934, 
      -82.5479052 14.3051899, -82.546187 14.3258089, -82.5539191 14.3498645, 
      -82.5805521 14.3593149, -82.6114807 14.3636106, -82.6767745 14.3618923, 
      -82.7162943 14.3618923, -82.7162943 14.3670471, -82.7180126 14.3885253, 
      -82.7180126 14.4091444, -82.7283221 14.4349182, -82.7412091 14.4409321, 
      -82.7575325 14.4443686, -82.7747151 14.4443686, -82.7755742 14.4435095, 
      -82.8101472 14.4464497, -82.8357131 14.4624103, -82.854614 14.4658468, 
      -82.8846834 14.4658468, -82.9104573 14.4529599, -82.9259216 14.4263269, 
      -82.9199077 14.4039896, -82.9113164 14.3842297, -82.8924156 14.3644697, 
      -82.8589096 14.3318228, -82.8257002 14.270157)), 
  POLYGON((-82 14, -82 14.5, -81 14.5,
    -81 14, -82 14)), 
  POLYGON((-81 14, -81 14.5, -80 14.5,
    -80 14, -81 14)), 
  POLYGON((-80 14, -80 14.5, -79 14.5,
    -79 14, -80 14)), 
  POLYGON((-79 14, -79 14.5, -78 14.5,
    -78 14, -79 14)), 
  POLYGON((-78 14, -78 14.5, -77 14.5,
    -77 14, -78 14)), 
  POLYGON((-77 14, -77 14.5, -76 14.5,
    -76 14, -77 14)), 
  POLYGON((-76 14, -76 14.5, -75 14.5,
    -75 14, -76 14)), 
  POLYGON((-75 14, -75 14.5, -74 14.5,
    -74 14, -75 14)), 
  POLYGON((-74 14, -74 14.5, -73 14.5,
    -73 14, -74 14)), 
  POLYGON((-73 14, -73 14.5, -72 14.5,
    -72 14, -73 14)), 
  POLYGON((-72 14, -72 14.5, -71 14.5,
    -71 14, -72 14)), 
  POLYGON((-71 14, -71 14.5, -70 14.5,
    -70 14, -71 14)), 
  POLYGON((-70 14, -70 14.5, -69 14.5,
    -69 14, -70 14)), 
  POLYGON((-69 14, -69 14.5, -68 14.5,
    -68 14, -69 14)), 
  POLYGON((-68 14, -68 14.5, -67 14.5,
    -67 14, -68 14)), 
  POLYGON((-67 14, -67 14.5, -66 14.5,
    -66 14, -67 14)), 
  POLYGON((-66 14, -66 14.5, -65 14.5,
    -65 14, -66 14)), 
  POLYGON((-65 14, -65 14.5, -64 14.5,
    -64 14, -65 14)), 
  POLYGON((-64 14, -64 14.5, -63 14.5,
    -63 14, -64 14)), 
  POLYGON((-63 14, -63 14.5, -62 14.5,
    -62 14, -63 14)), 
  POLYGON((-61.1175934 14, -62 14, -62 14.5,
    -61.128256013707521 14.5, -61.132420999999994 14.4791749, -61.132420999999994 14.4643473,
    -61.1274785 14.4346921, -61.1175934 14.4050369, -61.0978233 14.3803242,
    -61.0831468 14.3761309, -61.0632255 14.3704391, -61 14.2814735,
    -61 14.1529676, -61.0434554 14.073887, -61.1175934 14)), 
  POLYGON((-60.771634330004147 14.5, -60 14.5, -60 14,
    -60.7716159 14, -60.8062137 14.0788295, -60.8358689 14.1233123,
    -60.8605816 14.148024999999999, -60.8902368 14.1727377, -60.9446047 14.1727377,
    -61 14.1529676, -61 14.2814735, -60.9396621 14.2913586,
    -60.9149494 14.3061862, -60.8655241 14.3308989, -60.8309263 14.350669,
    -60.7864435 14.3951518, -60.7666734 14.4099794, -60.7567883 14.4445772,
    -60.7617308 14.4643473, -60.771634330004147 14.5)), 
  POLYGON((-60 14, -60 14.5, -59 14.5,
    -59 14, -60 14)), 
  POLYGON((-59 14, -59 14.5, -58 14.5,
    -58 14, -59 14)), 
  POLYGON((-58 14, -58 14.5, -57 14.5,
    -57 14, -58 14)), 
  POLYGON((-57 14, -57 14.5, -56 14.5,
    -56 14, -57 14)), 
  POLYGON((-56 14, -56 14.5, -55 14.5,
    -55 14, -56 14)), 
  POLYGON((-83.0908029231754 15, -83 15, -83 14.5,
    -83.157781366666669 14.5, -83.216450999999992 14.6760089, -83.2312961 14.729451300000001,
    -83.1837917 14.8719644, -83.1319545 14.9444714, -83.1133052 14.9718152,
    -83.0888017 14.9955543, -83.0908029231754 15)), 
  POLYGON((-83 14.5, -83 15, -82 15,
    -82 14.5, -83 14.5), 
    (-82.6725742 14.5252931, -82.646705 14.5182536, -82.6303815 14.5234084, 
      -82.620072 14.527704, -82.6028894 14.547464, -82.6037486 14.5689422, 
      -82.6432685 14.5818291, -82.6656058 14.5749561, -82.6853657 14.5594918, 
      -82.6879431 14.550900500000001, -82.6725742 14.5252931)), 
  POLYGON((-82 14.5, -82 15, -81 15,
    -81 14.5, -82 14.5)), 
  POLYGON((-81 14.5, -81 15, -80 15,
    -80 14.5, -81 14.5)), 
  POLYGON((-80 14.5, -80 15, -79 15,
    -79 14.5, -80 14.5)), 
  POLYGON((-79 14.5, -79 15, -78 15,
    -78 14.5, -79 14.5)), 
  POLYGON((-78 14.5, -78 15, -77 15,
    -77 14.5, -78 14.5)), 
  POLYGON((-77 14.5, -77 15, -76 15,
    -76 14.5, -77 14.5)), 
  POLYGON((-76 14.5, -76 15, -75 15,
    -75 14.5, -76 14.5)), 
  POLYGON((-75 14.5, -75 15, -74 15,
    -74 14.5, -75 14.5)), 
  POLYGON((-74 14.5, -74 15, -73 15,
    -73 14.5, -74 14.5)), 
  POLYGON((-73 14.5, -73 15, -72 15,
    -72 14.5, -73 14.5)), 
  POLYGON((-72 14.5, -72 15, -71 15,
    -71 14.5, -72 14.5)), 
  POLYGON((-71 14.5, -71 15, -70 15,
    -70 14.5, -71 14.5)), 
  POLYGON((-70 14.5, -70 15, -69 15,
    -69 14.5, -70 14.5)), 
  POLYGON((-69 14.5, -69 15, -68 15,
    -68 14.5, -69 14.5)), 
  POLYGON((-68 14.5, -68 15, -67 15,
    -67 14.5, -68 14.5)), 
  POLYGON((-67 14.5, -67 15, -66 15,
    -66 14.5, -67 14.5)), 
  POLYGON((-66 14.5, -66 15, -65 15,
    -65 14.5, -66 14.5)), 
  POLYGON((-65 14.5, -65 15, -64 15,
    -64 14.5, -65 14.5)), 
  POLYGON((-64 14.5, -64 15, -63 15,
    -63 14.5, -64 14.5)), 
  POLYGON((-63 14.5, -63 15, -62 15,
    -62 14.5, -63 14.5)), 
  POLYGON((-61.128256013707521 14.5, -62 14.5, -62 15,
    -61 15, -61 14.894347999999999, -61.132420999999994 14.9190607,
    -61.2213867 14.9190607, -61.260926999999995 14.9091756, -61.3103523 14.8844629,
    -61.3054098 14.830095, -61.2905822 14.7806696, -61.2708121 14.7361868,
    -61.2411568 14.7263018, -61.2312718 14.7065316, -61.2263292 14.6521637,
    -61.2263292 14.617566, -61.2016165 14.5780257, -61.1769039 14.5631981,
    -61.132420999999994 14.553313, -61.122536 14.5483704, -61.122536 14.5286003,
    -61.128256013707521 14.5)), 
  POLYGON((-61 14.894347999999999, -61 15, -60 15,
    -60 14.5, -60.771634330004147 14.5, -60.7864435 14.553313,
    -60.791385999999996 14.5631981, -60.8210413 14.6373361, -60.8457539 14.6818189,
    -60.8605816 14.7114741, -60.8408114 14.7312443, -60.8309263 14.755956999999999,
    -60.8358689 14.7954973, -60.8605816 14.8202099, -60.9149494 14.8399801,
    -60.9050644 14.8498652, -61 14.894347999999999)), 
  POLYGON((-60 14.5, -60 15, -59 15,
    -59 14.5, -60 14.5)), 
  POLYGON((-59 14.5, -59 15, -58 15,
    -58 14.5, -59 14.5)), 
  POLYGON((-58 14.5, -58 15, -57 15,
    -57 14.5, -58 14.5)), 
  POLYGON((-57 14.5, -57 15, -56 15,
    -56 14.5, -57 14.5)), 
  POLYGON((-56 14.5, -56 15, -55 15,
    -55 14.5, -56 14.5)), 
  POLYGON((-83.714661886187528 15.5, -83 15.5, -83 15,
    -83.0908029231754 15, -83.1261246 15.0784668, -83.2273308 15.161961999999999,
    -83.277934 15.2884698, -83.3867307 15.3643745, -83.6093845 15.4402793,
    -83.69794 15.4934126, -83.714661886187528 15.5)), 
  POLYGON((-83 15, -83 15.5, -82 15.5,
    -82 15, -83 15)), 
  POLYGON((-82 15, -82 15.5, -81 15.5,
    -81 15, -82 15)), 
  POLYGON((-81 15, -81 15.5, -80 15.5,
    -80 15, -81 15)), 
  POLYGON((-80 15, -80 15.5, -79 15.5,
    -79 15, -80 15)), 
  POLYGON((-79 15, -79 15.5, -78 15.5,
    -78 15, -79 15)), 
  POLYGON((-78 15, -78 15.5, -77 15.5,
    -77 15, -78 15)), 
  POLYGON((-77 15, -77 15.5, -76 15.5,
    -76 15, -77 15)), 
  POLYGON((-76 15, -76 15.5, -75 15.5,
    -75 15, -76 15)), 
  POLYGON((-75 15, -75 15.5, -74 15.5,
    -74 15, -75 15)), 
  POLYGON((-74 15, -74 15.5, -73 15.5,
    -73 15, -74 15)), 
  POLYGON((-73 15, -73 15.5, -72 15.5,
    -72 15, -73 15)), 
  POLYGON((-72 15, -72 15.5, -71 15.5,
    -71 15, -72 15)), 
  POLYGON((-71 15, -71 15.5, -70 15.5,
    -70 15, -71 15)), 
  POLYGON((-70 15, -70 15.5, -69 15.5,
    -69 15, -70 15)), 
  POLYGON((-69 15, -69 15.5, -68 15.5,
    -68 15, -69 15)), 
  POLYGON((-68 15, -68 15.5, -67 15.5,
    -67 15, -68 15)), 
  POLYGON((-67 15, -67 15.5, -66 15.5,
    -66 15, -67 15)), 
  POLYGON((-66 15, -66 15.5, -65 15.5,
    -65 15, -66 15)), 
  POLYGON((-65 15, -65 15.5, -64 15.5,
    -64 15, -65 15)), 
  POLYGON((-64 15, -64 15.5, -63 15.5,
    -63 15, -64 15)), 
  POLYGON((-63 15, -63 15.5, -62 15.5,
    -62 15, -63 15)), 
  POLYGON((-61.21611356243163 15.5, -61 15.5, -61 15,
    -62 15, -62 15.5, -61.516772156016472 15.5,
    -61.515328 15.4920571, -61.5187739 15.4713817, -61.5187739 15.4438144,
    -61.4774229 15.3886798, -61.4601934 15.3473289, -61.4636393 15.305978,
    -61.4498557 15.2784107, -61.4119506 15.2542893, -61.4119506 15.2439516,
    -61.4153965 15.2163843, -61.4119506 15.1784793, -61.3843833 15.1715874,
    -61.3568161 15.1750333, -61.3085733 15.1957088, -61.2947897 15.188816999999998,
    -61.2775601 15.188816999999998, -61.243100999999996 15.1991547, -61.205195999999994 15.2336138,
    -61.205195999999994 15.3232075, -61.1983042 15.4265848, -61.2155337 15.498949,
    -61.21611356243163 15.5)), 
  POLYGON((-61 15, -61 15.5, -60 15.5,
    -60 15, -61 15)), 
  POLYGON((-60 15, -60 15.5, -59 15.5,
    -59 15, -60 15)), 
  POLYGON((-59 15, -59 15.5, -58 15.5,
    -58 15, -59 15)), 
  POLYGON((-58 15, -58 15.5, -57 15.5,
    -57 15, -58 15)), 
  POLYGON((-57 15, -57 15.5, -56 15.5,
    -56 15, -57 15)), 
  POLYGON((-56 15, -56 15.5, -55 15.5,
    -55 15, -56 15)), 
  POLYGON((-88 16, -88 15.862527495139863, -88.0721499 15.8132587,
    -88.1275773 15.7541362, -88.1472848 15.7430507, -88.1928584 15.7763072,
    -88.2975545 15.8440517, -88.4355071 15.9265769, -88.5057151 15.9758457,
    -88.5907038 15.9930898, -88.655984899999993 15.9856995, -88.6793876 15.9549065,
    -88.6707655 15.9081011, -88.6165699 15.8699178, -88.6325822 15.8415883,
    -88.6633752 15.8243442, -88.7385101 15.8600641, -88.7853155 15.9019425,
    -88.8481332 15.913027999999999, -88.8703041 15.943821, -88.8493649 15.965992,
    -88.8198036 16, -88 16)), 
  POLYGON((-88 16, -87 16, -87 15.881674651652391,
    -87.0011221 15.880488, -87.0670669 15.8637594, -87.3146425 15.9167233,
    -87.4070215 15.9130282, -87.4452048 15.8908572, -87.5154128 15.9265771,
    -87.5523644 15.9684556, -87.6545971 15.9721507, -87.7666836 15.9783093,
    -87.8726115 15.9536749, -87.9748442 15.8994793, -87.9999997 15.8625277,
    -88 15.862527495139863, -88 16)), 
  POLYGON((-87 15.881674651652391, -87 16, -86.3014713 16,
    -86.4861727 15.9387198, -86.9542518 15.9108881, -86.9795533 15.9032976,
    -87 15.881674651652391)), 
  POLYGON((-84.8289201 16, -84 16, -84 15.677441467506279,
    -84.0011191 15.6780949, -84.1457778 15.8273933, -84.2090317 15.883056700000001,
    -84.3380697 15.9159488, -84.5607235 15.9285995, -84.6897615 15.9665519,
    -84.8289201 16)), 
  POLYGON((-84 15.677441467506279, -84 16, -83 16,
    -83 15.5, -83.714661886187528 15.5, -83.7814352 15.5263046,
    -83.9307145 15.6123299, -83.9534859 15.6502823, -84 15.677441467506279)), 
  POLYGON((-83 15.5, -83 16, -82 16,
    -82 15.5, -83 15.5)), 
  POLYGON((-82 15.5, -82 16, -81 16,
    -81 15.5, -82 15.5)), 
  POLYGON((-81 15.5, -81 16, -80 16,
    -80 15.5, -81 15.5)), 
  POLYGON((-80 15.5, -80 16, -79 16,
    -79 15.5, -80 15.5)), 
  POLYGON((-79 15.5, -79 16, -78 16,
    -78 15.5, -79 15.5)), 
  POLYGON((-78 15.5, -78 16, -77 16,
    -77 15.5, -78 15.5)), 
  POLYGON((-77 15.5, -77 16, -76 16,
    -76 15.5, -77 15.5)), 
  POLYGON((-76 15.5, -76 16, -75 16,
    -75 15.5, -76 15.5)), 
  POLYGON((-75 15.5, -75 16, -74 16,
    -74 15.5, -75 15.5)), 
  POLYGON((-74 15.5, -74 16, -73 16,
    -73 15.5, -74 15.5)), 
  POLYGON((-73 15.5, -73 16, -72 16,
    -72 15.5, -73 15.5)), 
  POLYGON((-72 15.5, -72 16, -71 16,
    -71 15.5, -72 15.5)), 
  POLYGON((-71 15.5, -71 16, -70 16,
    -70 15.5, -71 15.5)), 
  POLYGON((-70 15.5, -70 16, -69 16,
    -69 15.5, -70 15.5)), 
  POLYGON((-69 15.5, -69 16, -68 16,
    -68 15.5, -69 15.5)), 
  POLYGON((-68 15.5, -68 16, -67 16,
    -67 15.5, -68 15.5)), 
  POLYGON((-67 15.5, -67 16, -66 16,
    -66 15.5, -67 15.5)), 
  POLYGON((-66 15.5, -66 16, -65 16,
    -65 15.5, -66 15.5)), 
  POLYGON((-65 15.5, -65 16, -64 16,
    -64 15.5, -65 15.5)), 
  POLYGON((-64 15.5, -64 16, -63 16,
    -63 15.5, -64 15.5)), 
  POLYGON((-63 15.5, -63 16, -62 16,
    -62 15.5, -63 15.5)), 
  POLYGON((-61.516772156016472 15.5, -62 15.5, -62 16,
    -61.7944467 16, -61.7772172 15.9296878, -61.7530958 15.922796,
    -61.746204 15.8952287, -61.6807317 15.884891, -61.6531644 15.8400942,
    -61.6428267 15.8159728, -61.637983999999996 15.8139787, -61.5842462 15.7918514,
    -61.5187739 15.7884055, -61.4808689 15.8021891, -61.4843148 15.8573237,
    -61.4601934 15.9055665, -61.3568161 15.9572551, -61.3568161 15.9538092,
    -61.3430324 15.9090124, -61.3430324 15.8917828, -61.3361406 15.8538778,
    -61.3292488 15.846986, -61.318911 15.8400942, -61.2913437 15.8366482,
    -61.2120878 15.8263105, -61.1948582 15.8366482, -61.167291 15.8745533,
    -61.1603991 15.8883369, -61.1500614 15.9159042, -61.1431696 15.9296878,
    -61.1190482 15.9434715, -61.0845891 16, -61 16,
    -61 15.5, -61.21611356243163 15.5, -61.2706683 15.5988804,
    -61.3154651 15.6367854, -61.3843833 15.6677986, -61.4498557 15.6677986,
    -61.5049902 15.6609068, -61.5463412 15.6023263, -61.5360034 15.5575295,
    -61.5222198 15.5299622, -61.516772156016472 15.5)), 
  POLYGON((-61 15.5, -61 16, -60 16,
    -60 15.5, -61 15.5)), 
  POLYGON((-60 15.5, -60 16, -59 16,
    -59 15.5, -60 15.5)), 
  POLYGON((-59 15.5, -59 16, -58 16,
    -58 15.5, -59 15.5)), 
  POLYGON((-58 15.5, -58 16, -57 16,
    -57 15.5, -58 15.5)), 
  POLYGON((-57 15.5, -57 16, -56 16,
    -56 15.5, -57 15.5)), 
  POLYGON((-56 15.5, -56 16, -55 16,
    -55 15.5, -56 15.5)), 
  LINESTRING(-88.8198036 16, -88 16), 
  POLYGON((-87 16.082168089354784, -87 16, -88 16,
    -87.0023247 16.0019738, -87.001164648783771 16.5, -87 16.5,
    -87 16.0858820597608, -87.0004309 16.083153, -87 16.082168089354784)), 
  POLYGON((-86.3014713 16, -87 16, -87 16.082168089354784,
    -86.9953582 16.0715583, -86.9823141 16.0704713, -86.974705 16.0643116,
    -86.9526025 16.0592388, -86.9435441 16.0624999, -86.9207169 16.0748193,
    -86.9076728 16.078805, -86.9029624 16.0856894, -86.8783236 16.080979,
    -86.8743379 16.0867764, -86.8743379 16.0932985, -86.8757872 16.0980088,
    -86.8765119 16.0990958, -86.8754249 16.1027192, -86.8703522 16.1067049,
    -86.8674535 16.1132269, -86.8689028 16.1204737, -86.8703522 16.121923,
    -86.8739391 16.1298431, -86.89499099999999 16.1356918, -86.9047741 16.1335178,
    -86.9120208 16.1302567, -86.9203546 16.1277204, -86.9236156 16.1298944,
    -86.9312247 16.1298944, -86.9384714 16.1306191, -86.9518778 16.1295321,
    -86.9641972 16.124097, -86.9649219 16.119749, -86.9685453 16.1146763,
    -86.974705 16.1103283, -86.98122699999999 16.1070672, -86.9964452 16.0987335,
    -86.9971698 16.0943855, -86.9993439 16.0900374, -87 16.0858820597608,
    -87 16.5, -86 16.5, -86 16.2010715,
    -86.2205063 16.052577, -86.3014713 16), 
    (-86.6823573 16.2660232, -86.6553899 16.2382993, -86.6152165 16.2483427, 
      -86.5800648 16.2684294, -86.5474239 16.2985594, -86.4369471 16.3010703, 
      -86.3992845 16.351287, -86.3289811 16.3613304, -86.2662101 16.3713737, 
      -86.2335693 16.3864387, -86.1632658 16.3839279, -86.1105382 16.3889496, 
      -86.0879407 16.4165688, -86.1030057 16.4718072, -86.160755 16.4818505, 
      -86.2335693 16.4918939, -86.310406 16.4965904, -86.3666436 16.4768289, 
      -86.3867303 16.474318, -86.4645663 16.459253, -86.5524456 16.4341446, 
      -86.5976406 16.3989929, -86.6403249 16.3713737, -86.6679441 16.3286895, 
      -86.6779874 16.3111136, -86.6823573 16.2660232)), 
  POLYGON((-85.833126304600086 16.5, -85 16.5, -85 16.1006501,
    -85.1603706 16.0500469, -85.2944689 16.0070343, -85.489291 16.019685,
    -85.7878495 16.1133008, -86 16.2010715, -86 16.5,
    -85.9145182 16.5, -85.9173874 16.4990436, -85.9264815 16.4939913,
    -85.9411332 16.4808553, -85.9370914 16.4697403, -85.9391123 16.4621618,
    -85.9562901 16.4510468, -85.9699313 16.4374056, -85.9699313 16.4242696,
    -85.9754888 16.4126494, -85.9608371 16.3979977, -85.9411332 16.3969872,
    -85.9249659 16.4065866, -85.9183979 16.4111337, -85.9027358 16.4116389,
    -85.8895998 16.416186, -85.8830324 16.4238201, -85.8805057 16.4358899,
    -85.8729272 16.4409422, -85.8567599 16.4571095, -85.8557494 16.4656984,
    -85.8466553 16.4687298, -85.8365507 16.4747926, -85.8337324 16.4833336,
    -85.833126304600086 16.5)), 
  POLYGON((-85 16.1006501, -85 16.5, -84 16.5,
    -84 16, -84.8289201 16, -85 16.1006501)), 
  POLYGON((-84 16, -84 16.5, -83 16.5,
    -83 16, -84 16)), 
  POLYGON((-83 16, -83 16.5, -82 16.5,
    -82 16, -83 16)), 
  POLYGON((-82 16, -82 16.5, -81 16.5,
    -81 16, -82 16)), 
  POLYGON((-81 16, -81 16.5, -80 16.5,
    -80 16, -81 16)), 
  POLYGON((-80 16, -80 16.5, -79 16.5,
    -79 16, -80 16)), 
  POLYGON((-79 16, -79 16.5, -78 16.5,
    -78 16, -79 16)), 
  POLYGON((-78 16, -78 16.5, -77 16.5,
    -77 16, -78 16)), 
  POLYGON((-77 16, -77 16.5, -76 16.5,
    -76 16, -77 16)), 
  POLYGON((-76 16, -76 16.5, -75 16.5,
    -75 16, -76 16)), 
  POLYGON((-75 16, -75 16.5, -74 16.5,
    -74 16, -75 16)), 
  POLYGON((-74 16, -74 16.5, -73 16.5,
    -73 16, -74 16)), 
  POLYGON((-73 16, -73 16.5, -72 16.5,
    -72 16, -73 16)), 
  POLYGON((-72 16, -72 16.5, -71 16.5,
    -71 16, -72 16)), 
  POLYGON((-71 16, -71 16.5, -70 16.5,
    -70 16, -71 16)), 
  POLYGON((-70 16, -70 16.5, -69 16.5,
    -69 16, -70 16)), 
  POLYGON((-69 16, -69 16.5, -68 16.5,
    -68 16, -69 16)), 
  POLYGON((-68 16, -68 16.5, -67 16.5,
    -67 16, -68 16)), 
  POLYGON((-67 16, -67 16.5, -66 16.5,
    -66 16, -67 16)), 
  POLYGON((-66 16, -66 16.5, -65 16.5,
    -65 16, -66 16)), 
  POLYGON((-65 16, -65 16.5, -64 16.5,
    -64 16, -65 16)), 
  POLYGON((-64 16, -64 16.5, -63 16.5,
    -63 16, -64 16)), 
  POLYGON((-63 16, -63 16.5, -62 16.5,
    -62 16, -63 16)), 
  GEOMETRYCOLLECTION(
  POLYGON((-61.7944467 16, -62 16, -62 16.5,
    -61.548550476555256 16.5, -61.5566789 16.4810336, -61.5566789 16.4534663,
    -61.5497871 16.4293449, -61.5497871 16.4017776, -61.5428952 16.3948858,
    -61.5428952 16.3673185, -61.5566789 16.3673185, -61.5739084 16.3776563,
    -61.5808003 16.3845481, -61.6187053 16.3845481, -61.6255971 16.3638726,
    -61.629042999999996 16.3569808, -61.6600562 16.3673185, -61.6945153 16.3983317,
    -61.7289744 16.4121154, -61.7530958 16.4224531, -61.784108999999994 16.4362367,
    -61.8254599 16.4327908, -61.8392436 16.387994, -61.8461354 16.350089,
    -61.8461354 16.2880626, -61.8323518 16.2501576, -61.8254599 16.2019148,
    -61.8254599 16.1502261, -61.8289059 16.1123211, -61.8013386 16.0813079,
    -61.7944467 16)), 
    POLYGON((-61 16.1881312, -61 16, -61.0845891 16,
      -61.0432382 16.0606325, -61.0294545 16.0950916, -61 16.1881312)), 
    POLYGON((-61.384385203356636 16.5, -61 16.5, -61 16.4052236,
      -61.0329004 16.4017776, -61.1121564 16.3983317, -61.1535073 16.3707644,
      -61.163844999999995 16.3569808, -61.2189796 16.350089, -61.2396551 16.350089,
      -61.281006 16.3535349, -61.3016815 16.3604267, -61.3533701 16.4017776,
      -61.3671538 16.4569122, -61.3809374 16.4741418, -61.384385203356636 16.5))), 
  POLYGON((-61 16.4052236, -61 16.5, -60 16.5,
    -60 16, -61 16, -61 16.1881312,
    -60.9398608 16.2673871, -60.9260772 16.3190758, -60.932969 16.3811022,
    -60.9398608 16.4190072, -61 16.4052236)), 
  POLYGON((-60 16, -60 16.5, -59 16.5,
    -59 16, -60 16)), 
  POLYGON((-59 16, -59 16.5, -58 16.5,
    -58 16, -59 16)), 
  POLYGON((-58 16, -58 16.5, -57 16.5,
    -57 16, -58 16)), 
  POLYGON((-57 16, -57 16.5, -56 16.5,
    -56 16, -57 16)), 
  POLYGON((-56 16, -56 16.5, -55 16.5,
    -55 16, -56 16)), 
  POLYGON((-87 17, -87 16.5, -87.001164648783771 16.5,
    -87 17)), 
  POLYGON((-87 17, -86 17, -86 16.5,
    -87 16.5, -87 17)), 
  POLYGON((-85.9145182 16.5, -86 16.5, -86 17,
    -85 17, -85 16.5, -85.833126304600086 16.5,
    -85.8330141 16.5030854, -85.8395821 16.5106639, -85.8501919 16.513189999999998,
    -85.8562547 16.5126848, -85.865854 16.5152109, -85.8805057 16.519758,
    -85.8941469 16.5167266, -85.8956625 16.5071272, -85.9022305 16.5040959,
    -85.9145182 16.5)), 
  POLYGON((-85 16.5, -85 17, -84 17,
    -84 16.5, -85 16.5)), 
  POLYGON((-84 16.5, -84 17, -83 17,
    -83 16.5, -84 16.5)), 
  POLYGON((-83 16.5, -83 17, -82 17,
    -82 16.5, -83 16.5)), 
  POLYGON((-82 16.5, -82 17, -81 17,
    -81 16.5, -82 16.5)), 
  POLYGON((-81 16.5, -81 17, -80 17,
    -80 16.5, -81 16.5)), 
  POLYGON((-80 16.5, -80 17, -79 17,
    -79 16.5, -80 16.5)), 
  POLYGON((-79 16.5, -79 17, -78 17,
    -78 16.5, -79 16.5)), 
  POLYGON((-78 16.5, -78 17, -77 17,
    -77 16.5, -78 16.5)), 
  POLYGON((-77 16.5, -77 17, -76 17,
    -76 16.5, -77 16.5)), 
  POLYGON((-76 16.5, -76 17, -75 17,
    -75 16.5, -76 16.5)), 
  POLYGON((-75 16.5, -75 17, -74 17,
    -74 16.5, -75 16.5)), 
  POLYGON((-74 16.5, -74 17, -73 17,
    -73 16.5, -74 16.5)), 
  POLYGON((-73 16.5, -73 17, -72 17,
    -72 16.5, -73 16.5)), 
  POLYGON((-72 16.5, -72 17, -71 17,
    -71 16.5, -72 16.5)), 
  POLYGON((-71 16.5, -71 17, -70 17,
    -70 16.5, -71 16.5)), 
  POLYGON((-70 16.5, -70 17, -69 17,
    -69 16.5, -70 16.5)), 
  POLYGON((-69 16.5, -69 17, -68 17,
    -68 16.5, -69 16.5)), 
  POLYGON((-68 16.5, -68 17, -67 17,
    -67 16.5, -68 16.5)), 
  POLYGON((-67 16.5, -67 17, -66 17,
    -66 16.5, -67 16.5)), 
  POLYGON((-66 16.5, -66 17, -65 17,
    -65 16.5, -66 16.5)), 
  POLYGON((-65 16.5, -65 17, -64 17,
    -64 16.5, -65 16.5)), 
  POLYGON((-64 16.5, -64 17, -63 17,
    -63 16.5, -64 16.5)), 
  POLYGON((-63 16.5, -63 17, -62 17,
    -62 16.5, -63 16.5), 
    (-62.2487618 16.6444818, -62.2424153 16.6395456, -62.2286316 16.6257619, 
      -62.2103694 16.6351167, -62.176942999999994 16.6292079, -62.1528216 16.6292079, 
      -62.1390379 16.6360997, -62.1149166 16.6464374, -62.1080247 16.663667, 
      -62.1080247 16.6877883, -62.1287002 16.739477, -62.135591999999995 16.777382, 
      -62.1424838 16.8049493, -62.1459298 16.8118411, -62.1562675 16.8359625, 
      -62.176942999999994 16.8463002, -62.1838348 16.8463002, -62.214847999999996 16.8497462, 
      -62.2182939 16.8738675, -62.2355234 16.8910971, -62.2561989 16.894543, 
      -62.3078876 16.9427858, -62.309673 16.9588544, -62.3181024 16.9693806, 
      -62.3259991 16.9735173, -62.3744945 16.9725857, -62.3851157 16.9676171, 
      -62.3921697 16.953509, -62.3526844 16.9014348, -62.3354549 16.894543, 
      -62.3009958 16.8394084, -62.2699826 16.8325166, -62.2424153 16.8256248, 
      -62.2251857 16.8049493, -62.2355234 16.7808279, -62.2630907 16.7532607, 
      -62.2665366 16.7325852, -62.2734285 16.7153556, -62.2803203 16.6912343, 
      -62.2734285 16.663667, -62.2487618 16.6444818)), 
  POLYGON((-61.548550476555256 16.5, -62 16.5, -62 17,
    -61.9294318 17, -61.906261 16.982088, -61.8795254 16.9749585,
    -61.8492252 16.9606996, -61.7725834 16.9571349, -61.7315889 16.9678291,
    -61.6692061 17, -61 17, -61 16.5,
    -61.384385203356636 16.5, -61.3878293 16.5258304, -61.3912752 16.5327223,
    -61.4567475 16.5396141, -61.4843148 16.5327223, -61.5394493 16.5120468,
    -61.5463412 16.505155, -61.548550476555256 16.5)), 
  POLYGON((-61 16.5, -61 17, -60 17,
    -60 16.5, -61 16.5)), 
  POLYGON((-60 16.5, -60 17, -59 17,
    -59 16.5, -60 16.5)), 
  POLYGON((-59 16.5, -59 17, -58 17,
    -58 16.5, -59 16.5)), 
  POLYGON((-58 16.5, -58 17, -57 17,
    -57 16.5, -58 16.5)), 
  POLYGON((-57 16.5, -57 17, -56 17,
    -56 16.5, -57 16.5)), 
  POLYGON((-56 16.5, -56 17, -55 17,
    -55 16.5, -56 16.5)), 
  POLYGON((-86.005634629927428 17.5, -86 17.5, -86 17,
    -87 17, -86.0123633 17.0110227, -86.005634629927428 17.5)), 
  POLYGON((-86 17, -86 17.5, -85 17.5,
    -85 17, -86 17)), 
  POLYGON((-85 17, -85 17.5, -84 17.5,
    -84 17, -85 17)), 
  POLYGON((-84 17, -84 17.5, -83 17.5,
    -83 17, -84 17), 
    (-83.9278187 17.4659694, -83.94411199999999 17.4626149, -83.9651974 17.4539891, 
      -83.9752609 17.4333829, -83.9776569 17.4142143, -83.9738232 17.3993587, 
      -83.9613637 17.3859407, -83.9359653 17.3768357, -83.9144007 17.3768357, 
      -83.8890024 17.3830654, -83.8784597 17.3969626, -83.8741467 17.4161312, 
      -83.8765428 17.4343413, -83.8948354 17.4579067, -83.9115254 17.4654902, 
      -83.9278187 17.4659694)), 
  POLYGON((-83 17, -83 17.5, -82 17.5,
    -82 17, -83 17)), 
  POLYGON((-82 17, -82 17.5, -81 17.5,
    -81 17, -82 17)), 
  POLYGON((-81 17, -81 17.5, -80 17.5,
    -80 17, -81 17)), 
  POLYGON((-80 17, -80 17.5, -79 17.5,
    -79 17, -80 17)), 
  POLYGON((-79 17, -79 17.5, -78 17.5,
    -78 17, -79 17)), 
  POLYGON((-78 17, -78 17.5, -77 17.5,
    -77 17, -78 17)), 
  POLYGON((-77 17, -77 17.5, -76 17.5,
    -76 17, -77 17)), 
  POLYGON((-76 17, -76 17.5, -75 17.5,
    -75 17, -76 17)), 
  POLYGON((-75 17, -75 17.5, -74 17.5,
    -74 17, -75 17)), 
  POLYGON((-74 17, -74 17.5, -73 17.5,
    -73 17, -74 17)), 
  POLYGON((-73 17, -73 17.5, -72 17.5,
    -72 17, -73 17)), 
  POLYGON((-71.442826919264064 17.5, -71 17.5, -71 17,
    -72 17, -72 17.5, -71.712531629974478 17.5,
    -71.7164392 17.4570164, -71.6690341 17.4184997, -71.5653353 17.4333139,
    -71.5001532 17.4570164, -71.452748 17.4866447, -71.442826919264064 17.5)), 
  POLYGON((-71 17, -71 17.5, -70 17.5,
    -70 17, -71 17)), 
  POLYGON((-70 17, -70 17.5, -69 17.5,
    -69 17, -70 17)), 
  POLYGON((-69 17, -69 17.5, -68 17.5,
    -68 17, -69 17)), 
  POLYGON((-68 17, -68 17.5, -67 17.5,
    -67 17, -68 17)), 
  POLYGON((-67 17, -67 17.5, -66 17.5,
    -66 17, -67 17)), 
  POLYGON((-66 17, -66 17.5, -65 17.5,
    -65 17, -66 17)), 
  POLYGON((-65 17, -65 17.5, -64 17.5,
    -64 17, -65 17)), 
  POLYGON((-63 17.451353743322375, -63 17, -64 17,
    -64 17.5, -63.049899739095238 17.5, -63.0003941 17.4514468,
    -63 17.451353743322375)), 
  POLYGON((-62.9012327 17.5, -62 17.5, -62 17,
    -63 17, -63 17.451353743322375, -62.9432645 17.4379571,
    -62.9165842 17.4412922, -62.9012327 17.4505193, -62.9012327 17.5), 
    (-62.9374724 17.3906153, -62.9285948 17.3611431, -62.8866757 17.2912778, 
      -62.8342767 17.2773048, -62.7748913 17.2633317, -62.7539317 17.2493587, 
      -62.7399586 17.2353856, -62.7224923 17.2074395, -62.6945462 17.1760001, 
      -62.6735866 17.1445608, -62.6770799 17.0956551, -62.6700934 17.0607225, 
      -62.6142012 17.0432561, -62.5652955 17.0502427, -62.5233763 17.0537359, 
      -62.5128922 17.084654, -62.5006538 17.0997167, -62.4814572 17.1585338, 
      -62.5233763 17.214426, -62.5548157 17.2493587, -62.5792685 17.2563452, 
      -62.6072146 17.2703182, -62.6316675 17.3052509, -62.6596136 17.3541565, 
      -62.6910529 17.3960757, -62.7171803 17.4245064, -62.7567199 17.4405106, 
      -62.7953181 17.4508662, -62.8301506 17.4631046, -62.9035813 17.423565, 
      -62.9374724 17.3906153)), 
  POLYGON((-61.9294318 17, -62 17, -62 17.5,
    -61 17.5, -61 17, -61.6692061 17,
    -61.6690417 17.00069, -61.6460353 17.030212, -61.6192998 17.064076999999997,
    -61.6371235 17.1104186, -61.6513824 17.131807, -61.6905945 17.1496307,
    -61.7012887 17.1674544, -61.7208948 17.1799309, -61.7636716 17.1834957,
    -61.765454 17.185278, -61.7814953 17.199537, -61.8296192 17.1959722,
    -61.8563547 17.1781486, -61.8652666 17.1621073, -61.8830903 17.1585425,
    -61.920519999999996 17.1389365, -61.9294318 17.1015067, -61.9116081 17.0569476,
    -61.9258671 17.0337768, -61.9294318 17)), 
  POLYGON((-61 17, -61 17.5, -60 17.5,
    -60 17, -61 17)), 
  POLYGON((-60 17, -60 17.5, -59 17.5,
    -59 17, -60 17)), 
  POLYGON((-59 17, -59 17.5, -58 17.5,
    -58 17, -59 17)), 
  POLYGON((-58 17, -58 17.5, -57 17.5,
    -57 17, -58 17)), 
  POLYGON((-57 17, -57 17.5, -56 17.5,
    -56 17, -57 17)), 
  POLYGON((-56 17, -56 17.5, -55 17.5,
    -55 17, -56 17)), 
  POLYGON((-86 17.909472614751053, -86 17.5, -86.005634629927428 17.5,
    -86 17.909472614751053)), 
  POLYGON((-85.9987542797838 18, -85 18, -85 17.5,
    -86 17.5, -86 17.909472614751053, -85.9987542797838 18)), 
  POLYGON((-85 17.5, -85 18, -84 18,
    -84 17.5, -85 17.5)), 
  POLYGON((-84 17.5, -84 18, -83 18,
    -83 17.5, -84 17.5)), 
  POLYGON((-83 17.5, -83 18, -82 18,
    -82 17.5, -83 17.5)), 
  POLYGON((-82 17.5, -82 18, -81 18,
    -81 17.5, -82 17.5)), 
  POLYGON((-81 17.5, -81 18, -80 18,
    -80 17.5, -81 17.5)), 
  POLYGON((-80 17.5, -80 18, -79 18,
    -79 17.5, -80 17.5)), 
  POLYGON((-78 17.7753041, -78 17.5, -79 17.5,
    -79 18, -78.1705641 18, -78.1395407 17.934853,
    -78.0907897 17.9171254, -78.0597663 17.8728062, -78.0420386 17.8240552,
    -78 17.7753041)), 
  POLYGON((-77 17.770872252420894, -77 17.5, -78 17.5,
    -78 17.7753041, -77.8603301 17.8151914, -77.7362365 17.779736,
    -77.6475982 17.7930318, -77.5766876 17.8063275, -77.452593999999991 17.7753041,
    -77.3728195 17.7575765, -77.2930028 17.6822041, -77.2354302 17.6423467,
    -77.1556557 17.6334829, -77.066233 17.7097892, -77.0980408 17.779736,
    -77.0581536 17.7885999, -77 17.770872252420894)), 
  POLYGON((-76.074268632448835 18, -76 18, -76 17.5,
    -77 17.5, -77 17.770872252420894, -76.9999995 17.7708721,
    -76.9251961 17.8018956, -76.8498536 17.832919, -76.827694 17.8595105,
    -76.7966706 17.8639424, -76.77451099999999 17.8373509, -76.6814408 17.8107594,
    -76.5750749 17.784167999999998, -76.5130281 17.7664403, -76.3845026 17.7753041,
    -76.2426813 17.8018956, -76.1496111 17.8284871, -76.0654047 17.8949658,
    -76.0476771 17.9259892, -76.0742685 17.9999998, -76.074268632448835 18)), 
  POLYGON((-76 17.5, -76 18, -75 18,
    -75 17.5, -76 17.5)), 
  POLYGON((-75 17.5, -75 18, -74 18,
    -74 17.5, -75 17.5)), 
  POLYGON((-74 17.5, -74 18, -73 18,
    -73 17.5, -74 17.5)), 
  POLYGON((-73 17.5, -73 18, -72 18,
    -72 17.5, -73 17.5)), 
  GEOMETRYCOLLECTION(
  POLYGON((-71.712531629974478 17.5, -72 17.5, -72 18,
    -71.8527291 18, -71.8260637 17.9606963, -71.7680385 17.9221546,
    -71.754023899999993 17.9025341, -71.7500998 17.8930041, -71.7377669 17.875626,
    -71.7282905 17.8569975, -71.7312534 17.8481091, -71.7431046 17.8036667,
    -71.7460675 17.7888526, -71.7223649 17.7266333, -71.6334802 17.6910794,
    -71.6275546 17.6673769, -71.6690341 17.6051576, -71.7105136 17.522198500000002,
    -71.712531629974478 17.5)), 
    POLYGON((-71.0705438 18, -71 18, -71 17.5,
      -71.442826919264064 17.5, -71.3757146 17.5903435, -71.2897927 17.7029307,
      -71.1920196 17.8036667, -71.1594285 17.8955142, -71.0705438 18))), 
  POLYGON((-71 17.5, -71 18, -70 18,
    -70 17.5, -71 17.5)), 
  POLYGON((-70 17.5, -70 18, -69 18,
    -69 17.5, -70 17.5)), 
  POLYGON((-69 17.5, -69 18, -68 18,
    -68 17.5, -69 17.5)), 
  POLYGON((-67 17.883895926398026, -67 17.5, -68 17.5,
    -68 18, -67.902978153284678 18, -67.900887 17.999481,
    -67.896644617594248 18, -67.889023108433733 18, -67.886395 17.999433,
    -67.879572 17.999458, -67.877615415631837 18, -67.87463 18,
    -67.87463 17.999771, -67.873264434622456 18, -67.264282254025034 18,
    -67.265723999999992 17.99456, -67.266565 17.990341, -67.267495 17.983071,
    -67.26744699999999 17.979381, -67.267048 17.975251, -67.266434 17.971522999999998,
    -67.266168999999991 17.968469, -67.2659 17.960483, -67.266500999999991 17.955088999999997,
    -67.26664 17.949824, -67.266002 17.944183, -67.264135 17.936788,
    -67.260680999999991 17.928541, -67.258648 17.925235999999998, -67.25612 17.921893,
    -67.252709 17.918039, -67.249625 17.914037999999998, -67.247676 17.912672999999998,
    -67.24537 17.911503, -67.242871999999991 17.910528, -67.241328 17.910545,
    -67.241036999999992 17.910376, -67.236102 17.902466999999998, -67.233481 17.899511999999998,
    -67.2315 17.897778, -67.22909 17.896003, -67.218516 17.889094,
    -67.214961 17.887225, -67.21059799999999 17.885417, -67.206032 17.884027,
    -67.201416999999992 17.8829, -67.194504999999992 17.881746, -67.189104 17.881442,
    -67.186041 17.881469, -67.179648 17.882042, -67.170384 17.883626,
    -67.166921 17.884557, -67.162832999999992 17.886101999999998, -67.156166 17.889716999999997,
    -67.148544 17.894924, -67.145635 17.89746, -67.142789999999991 17.90057,
    -67.139039 17.899152, -67.135961999999992 17.898274, -67.129907 17.897151,
    -67.123524 17.896698, -67.121059 17.896179, -67.115259 17.895725,
    -67.109748 17.89558, -67.10863599999999 17.895408, -67.106707 17.895407,
    -67.10471 17.89569, -67.101841999999991 17.894956, -67.095421 17.893983,
    -67.090981 17.894199, -67.087724 17.894647, -67.086705 17.894941,
    -67.08279499999999 17.893587999999998, -67.078135 17.892906, -67.06357899999999 17.891908,
    -67.058055 17.891977, -67.054496 17.889947, -67.051129 17.888787999999998,
    -67.044003 17.887159999999998, -67.036058 17.886716, -67.031189 17.887147,
    -67.029849 17.886267, -67.023541 17.884926, -67.019908 17.884355,
    -67.017057999999992 17.884214, -67.012661 17.884428, -67.00906599999999 17.884856,
    -67.006021 17.885796, -67.003886999999992 17.884869, -67.000591 17.883763,
    -67 17.883895926398026)), 
  POLYGON((-66 17.899852897499997, -66 17.5, -67 17.5,
    -67 17.883895926398026, -66.998159 17.88431, -66.996822999999992 17.884074,
    -66.976925 17.878733, -66.971493 17.878466, -66.967826 17.878799,
    -66.966473999999991 17.878083, -66.962332 17.877257999999998, -66.954017999999991 17.876504999999998,
    -66.950006 17.876616, -66.941971999999993 17.877444999999998, -66.938925 17.878172,
    -66.938329 17.878470999999998, -66.93387 17.877201, -66.924336 17.875359,
    -66.920998 17.875177, -66.917796 17.87548, -66.912897 17.876989,
    -66.907817999999992 17.879119, -66.902733 17.881916, -66.897503 17.885206,
    -66.893819999999991 17.888887, -66.891791 17.891583999999998, -66.890562 17.893705,
    -66.888246 17.892574, -66.884963 17.891818, -66.881455 17.891299,
    -66.87799 17.891063, -66.875939 17.891158, -66.87462 17.89142,
    -66.871999 17.891413, -66.865473 17.892339, -66.8572 17.894015,
    -66.851102 17.895706999999998, -66.847016 17.89706, -66.843266 17.898951999999998,
    -66.839157 17.89901, -66.832353 17.899428, -66.829032 17.899793,
    -66.82531 17.90045, -66.821237 17.901775, -66.81795799999999 17.903629,
    -66.812884 17.903686, -66.810295 17.904353, -66.805016 17.906479,
    -66.799748 17.909095999999998, -66.794096 17.912468999999998, -66.791117 17.914519,
    -66.788153999999992 17.916064, -66.782471 17.915046999999998, -66.775628 17.914735,
    -66.770130999999992 17.915042, -66.767316999999991 17.915437, -66.76304 17.917075,
    -66.76189699999999 17.917880999999998, -66.759574 17.917597, -66.749619 17.915627999999998,
    -66.744801999999993 17.915418, -66.741534 17.915771, -66.739958 17.916159,
    -66.73854399999999 17.916797, -66.737965 17.916441, -66.733313 17.914658,
    -66.728653999999992 17.913178, -66.723778 17.912097, -66.721882999999991 17.912067999999998,
    -66.713398 17.912852, -66.712493 17.911437, -66.710066 17.909519,
    -66.70444599999999 17.906437, -66.698996 17.903798, -66.69729 17.902796,
    -66.690467 17.899857, -66.687169 17.898886, -66.682067 17.897109,
    -66.679665 17.896811, -66.672187 17.896902, -66.669237 17.897589999999997,
    -66.666767 17.898412, -66.659351 17.901363, -66.652738 17.904542,
    -66.6491 17.906558999999998, -66.645372999999992 17.908942, -66.641674 17.911973,
    -66.639963 17.911922, -66.636639 17.912164999999998, -66.624617 17.913535,
    -66.62189699999999 17.913591999999998, -66.619225 17.913925, -66.61542 17.914814,
    -66.614274 17.915309, -66.604683 17.911046, -66.601989 17.910211,
    -66.599068 17.90957, -66.59167 17.908718999999998, -66.588662 17.908825999999998,
    -66.585940999999991 17.909145, -66.582624 17.909829, -66.580681 17.910465,
    -66.574068 17.913141, -66.576773 17.909295, -66.581049 17.902192,
    -66.582247 17.899639999999998, -66.584024 17.895011, -66.585188 17.891042,
    -66.585737999999992 17.888503999999998, -66.586187 17.881481, -66.586799 17.878192,
    -66.586907 17.876437, -66.586799 17.874793, -66.585314 17.868320999999998,
    -66.583086 17.862156, -66.580195 17.857032999999998, -66.576596999999992 17.85185,
    -66.571297 17.84548, -66.564204 17.840096, -66.556803 17.836253,
    -66.548560999999992 17.833253, -66.538175 17.831677, -66.530937999999992 17.831509999999998,
    -66.522652 17.832822999999998, -66.513314999999992 17.835414999999998, -66.504265 17.839042,
    -66.499614 17.841558, -66.496980999999991 17.843342, -66.494152 17.846166999999998,
    -66.488953 17.847376999999998, -66.479395 17.852023, -66.476344 17.854048,
    -66.473191 17.856673999999998, -66.465406 17.86414, -66.463004 17.866993,
    -66.461742 17.869671, -66.457279 17.869407, -66.453212999999991 17.869605999999997,
    -66.441667 17.871533, -66.437862 17.873026, -66.424486 17.881992,
    -66.422625 17.880634, -66.421114 17.879068, -66.419099 17.877657,
    -66.415243 17.875563, -66.4111 17.873844, -66.405771 17.871916,
    -66.399908 17.870594, -66.394154 17.869539, -66.38812 17.869218,
    -66.381019999999992 17.869273, -66.374613 17.870155, -66.372655 17.869954,
    -66.370207 17.870372, -66.36783 17.871088, -66.362625999999992 17.873026,
    -66.366413 17.873026, -66.364772 17.87341, -66.359777 17.875448,
    -66.35851 17.876319, -66.354897999999991 17.876058999999998, -66.349797999999993 17.876248999999998,
    -66.345038 17.876807, -66.341407 17.877484, -66.340356 17.877883999999998,
    -66.33949 17.877530999999998, -66.331698 17.877471, -66.323129999999992 17.877671,
    -66.316481 17.878242, -66.310276 17.879316, -66.306133 17.88048,
    -66.303094 17.882113, -66.284413 17.878517, -66.282814 17.878379,
    -66.272644 17.873026, -66.274382 17.873026, -66.271776 17.871336,
    -66.266963 17.869543999999998, -66.261969999999991 17.867188, -66.257369 17.865593,
    -66.249612 17.864053, -66.249612 17.863861, -66.243313 17.862565999999998,
    -66.235961 17.862242, -66.227330999999992 17.862242, -66.220300999999992 17.862889,
    -66.213273 17.864508999999998, -66.193654 17.873026, -66.18387899999999 17.879341,
    -66.179712 17.882566, -66.175305 17.880937, -66.172439 17.880316,
    -66.167175 17.879507, -66.161402999999993 17.87914, -66.148496999999992 17.877927,
    -66.136642999999992 17.877847, -66.131185 17.877976999999998, -66.125739 17.878472,
    -66.122305 17.879465, -66.114139999999992 17.882628999999998, -66.103687999999991 17.889229999999998,
    -66.09264499999999 17.897458999999998, -66.07743099999999 17.909824999999998, -66.076724 17.910671999999998,
    -66.067291 17.906091999999997, -66.060981 17.903223999999998, -66.056153999999992 17.900282999999998,
    -66.0498 17.897745999999998, -66.041777 17.895774, -66.018892 17.894132,
    -66.00841 17.895951999999998, -66.002808 17.89809, -66 17.899852897499997)), 
  POLYGON((-65.783619038319813 18, -65 18, -65 17.5,
    -66 17.5, -66 17.899852897499997, -65.999608 17.900099,
    -65.991317999999993 17.905991, -65.984522 17.911102, -65.980918 17.914029,
    -65.97549699999999 17.919718, -65.972882 17.923246, -65.96712 17.922296,
    -65.958696 17.921160999999998, -65.95388299999999 17.920824, -65.944611 17.920571,
    -65.940094 17.920863, -65.937006 17.921281999999998, -65.92638 17.921972999999998,
    -65.921748 17.922607, -65.910598 17.924775, -65.904901 17.926261999999998,
    -65.89932 17.927986, -65.888436 17.932572999999998, -65.882307 17.935363,
    -65.877935 17.937634, -65.874606 17.93798, -65.870119 17.937894999999997,
    -65.852532 17.941485999999998, -65.845874 17.94527, -65.834591 17.954385,
    -65.827849 17.962401, -65.822636 17.963424, -65.810983 17.967453,
    -65.805066 17.971211, -65.797639 17.976983, -65.788615 17.989407,
    -65.783619038319813 18)), 
  POLYGON((-65 17.5, -65 18, -64 18,
    -64 17.5, -65 17.5), 
    (-64.908023 17.623958, -64.900361 17.623468, -64.892524999999992 17.62423, 
      -64.885410999999991 17.625951999999998, -64.876796 17.629379999999998, -64.869340999999991 17.634152, 
      -64.865563 17.635329, -64.86258 17.635141, -64.859898 17.633765, 
      -64.851308 17.631166, -64.843842 17.630104, -64.835754 17.630132, 
      -64.83005 17.630663, -64.8257 17.631543999999998, -64.824489 17.631190999999998, 
      -64.82191 17.629576999999998, -64.81759 17.627786999999998, -64.811528 17.625687, 
      -64.801686 17.624695, -64.794596 17.625044, -64.788861 17.625844, 
      -64.781228 17.62779, -64.776248999999993 17.629321, -64.772746 17.629984999999998, 
      -64.767564999999991 17.629931, -64.761006 17.630202999999998, -64.756436 17.630646, 
      -64.752909 17.632051999999998, -64.749589 17.633696, -64.746127 17.634543, 
      -64.740348 17.636342, -64.737459 17.638137999999998, -64.734180999999992 17.638379, 
      -64.724947 17.640722, -64.716559 17.643504, -64.711097999999993 17.646722999999998, 
      -64.708137999999991 17.647537, -64.698902 17.648899, -64.696118 17.649545, 
      -64.687834 17.650803, -64.68096 17.652973, -64.67014 17.657041, 
      -64.661186 17.659314, -64.654733 17.659454, -64.646836 17.660238, 
      -64.633686 17.664642999999998, -64.628011 17.667825999999998, -64.622306999999992 17.670161, 
      -64.618332999999993 17.672912999999998, -64.616151 17.673866999999998, -64.606822999999991 17.677105, 
      -64.600878999999992 17.681116, -64.596445 17.683362, -64.593595999999991 17.685367, 
      -64.58976 17.687034999999998, -64.584974 17.689417, -64.579239 17.692749, 
      -64.576988 17.694516, -64.571317 17.695270999999998, -64.560964 17.697967, 
      -64.548748 17.703038, -64.539776 17.708714, -64.531016 17.715913, 
      -64.524722 17.722703, -64.520297 17.729862999999998, -64.51686 17.738682999999998, 
      -64.514093 17.747061, -64.5136 17.748029, -64.512734 17.752357, 
      -64.512674 17.757141999999998, -64.513178 17.760407, -64.516004999999993 17.771444, 
      -64.520068999999992 17.779576, -64.523061 17.785019, -64.526868999999991 17.789986, 
      -64.529821 17.793003, -64.538884 17.799761999999998, -64.546774 17.804875, 
      -64.553169 17.80848, -64.556191 17.810553, -64.558791 17.814412, 
      -64.563735999999992 17.819608, -64.567115 17.822791, -64.583536999999993 17.833078, 
      -64.59231299999999 17.837342, -64.602154 17.840859, -64.612062 17.843497, 
      -64.615186 17.843985, -64.632284 17.844953999999998, -64.638882 17.844877, 
      -64.647098 17.843731, -64.652346999999992 17.842382999999998, -64.663399 17.837584, 
      -64.668112999999991 17.835227, -64.673278 17.831512, -64.67899899999999 17.826269, 
      -64.682012 17.822715, -64.685442 17.817735, -64.686167 17.815088, 
      -64.690001999999993 17.813249, -64.697524 17.808761, -64.700525 17.809198, 
      -64.704255 17.813584, -64.708833 17.818195, -64.713833 17.822599, 
      -64.71851 17.825924, -64.724385 17.829166999999998, -64.733544999999992 17.832843, 
      -64.743188 17.834753, -64.747072 17.835179, -64.749586 17.835181, 
      -64.758443 17.833928999999998, -64.762647 17.833064999999998, -64.765149 17.833142, 
      -64.767656 17.832644, -64.774903 17.834056, -64.783075 17.834524, 
      -64.789635 17.834175, -64.79807 17.833054999999998, -64.80376 17.831689, 
      -64.809494 17.829444, -64.810898999999992 17.829821, -64.812592999999993 17.829656999999997, 
      -64.819411 17.828091999999998, -64.822243 17.826918, -64.830249 17.823024999999998, 
      -64.834893999999991 17.820303, -64.84122099999999 17.815085, -64.842141 17.81475, 
      -64.843445 17.814754, -64.844775 17.815637, -64.846882 17.816387, 
      -64.851767 17.817687, -64.85619 17.819425, -64.861566 17.820996, 
      -64.869207 17.821628, -64.872486 17.821537, -64.874589 17.821206, 
      -64.87671 17.821234, -64.882729 17.820339, -64.890481999999992 17.818324999999998, 
      -64.899298 17.815462999999998, -64.905885 17.812203, -64.910561 17.809017, 
      -64.916335 17.804544, -64.92019599999999 17.801157, -64.92678699999999 17.794961999999998, 
      -64.931536 17.788973, -64.934196 17.784679, -64.941257 17.768829, 
      -64.943598999999992 17.762612, -64.945446 17.756633, -64.946811 17.748027999999998, 
      -64.945424 17.736822, -64.943731 17.729167, -64.938938 17.717534, 
      -64.93956399999999 17.716694999999998, -64.943502999999993 17.712685999999998, -64.947926999999993 17.706343999999998, 
      -64.9518 17.699281, -64.95447399999999 17.691829, -64.956384 17.683652, 
      -64.956713999999991 17.678977, -64.956488999999991 17.672328999999998, -64.955032 17.666611, 
      -64.953181 17.661022, -64.949714 17.653274, -64.943583 17.644095, 
      -64.935171 17.635186, -64.925838 17.629043, -64.915914 17.625231, 
      -64.908023 17.623958)), 
  POLYGON((-63.049899739095238 17.5, -64 17.5, -64 18,
    -63.2499749 18, -63.2078466 17.9868185, -63.1914127 17.9825579,
    -63.1722399 17.9747296, -63.1341983 17.962471999999998, -63.1075061 17.9465948,
    -63.0865186 17.9354449, -63.0167263 17.9096026, -63 17.9340382,
    -63 17.597721451149425, -63.0013839 17.598806, -63.0578455 17.6231378,
    -63.1170721 17.6648412, -63.1300863 17.6713887, -63.247783 17.6912771,
    -63.2849076 17.67903, -63.2988806 17.6336176, -63.2880396 17.5999938,
    -63.263948 17.560259, -63.2360019 17.546286, -63.1731231 17.5392995,
    -63.131204 17.5567658, -63.0962714 17.5392995, -63.0753118 17.5253264,
    -63.0543522 17.5043668, -63.049899739095238 17.5)), 
  POLYGON((-62.7294788 18, -62 18, -62 17.5,
    -62.9012327 17.5, -62.9012327 17.515635, -62.9061543 17.5349425,
    -62.931388 17.5439509, -63 17.597721451149425, -63 17.9340382,
    -62.9705139 17.9165719, -62.9251015 17.878145999999997, -62.8901689 17.825747,
    -62.8203036 17.825747, -62.7609181 17.8432133, -62.7399585 17.8746527,
    -62.7050259 17.9025988, -62.6875596 17.9340382, -62.7294788 18)), 
  POLYGON((-62 17.5, -62 18, -61 18,
    -61 17.5, -62 17.5), 
    (-61.8894539 17.6742458, -61.8706137 17.6451289, -61.8634842 17.6237405, 
      -61.8741784 17.5809637, -61.8599195 17.5577929, -61.8438782 17.5364045, 
      -61.8046661 17.5132337, -61.765454 17.5132337, -61.7280243 17.509669, 
      -61.6905945 17.5310574, -61.6692061 17.5702695, -61.6674238 17.6290876, 
      -61.7636716 17.7413768, -61.8100132 17.769894700000002, -61.8759608 17.7770241, 
      -61.9169552 17.7217707, -61.906261 17.6985999, -61.8894539 17.6742458)), 
  POLYGON((-61 17.5, -61 18, -60 18,
    -60 17.5, -61 17.5)), 
  POLYGON((-60 17.5, -60 18, -59 18,
    -59 17.5, -60 17.5)), 
  POLYGON((-59 17.5, -59 18, -58 18,
    -58 17.5, -59 17.5)), 
  POLYGON((-58 17.5, -58 18, -57 18,
    -57 17.5, -58 17.5)), 
  POLYGON((-57 17.5, -57 18, -56 18,
    -56 17.5, -57 17.5)), 
  POLYGON((-56 17.5, -56 18, -55 18,
    -55 17.5, -56 17.5)), 
  POLYGON((-85.9942782647015 18.5, -85 18.5, -85 18,
    -85.9987542797838 18, -85.9987208 18.002433, -85.9942782647015 18.5)), 
  POLYGON((-85 18, -85 18.5, -84 18.5,
    -84 18, -85 18)), 
  POLYGON((-84 18, -84 18.5, -83 18.5,
    -83 18, -84 18)), 
  POLYGON((-83 18, -83 18.5, -82 18.5,
    -82 18, -83 18)), 
  POLYGON((-82 18, -82 18.5, -81 18.5,
    -81 18, -82 18)), 
  POLYGON((-81 18, -81 18.5, -80 18.5,
    -80 18, -81 18)), 
  POLYGON((-80 18, -80 18.5, -79 18.5,
    -79 18, -80 18)), 
  POLYGON((-78.1705641 18, -79 18, -79 18.5,
    -78.413902747395184 18.5, -78.4852301 18.4445232, -78.4940939 18.3780445,
    -78.5206854 18.284974300000002, -78.4586386 18.1830403, -78.4320471 18.1520169,
    -78.2946577 18.1032658, -78.2193152 18.0811062, -78.1705641 18)), 
  POLYGON((-76.913255004117858 18.5, -76 18.5, -76 18,
    -76.074268632448835 18, -76.207225999999991 18.2007679, -76.3800706 18.2539509,
    -76.5351877 18.302702, -76.6814408 18.3603168, -76.7390557 18.3957722,
    -76.7833749 18.4179317, -76.8454217 18.4356594, -76.8498536 18.4799785,
    -76.913255004117858 18.5)), 
  POLYGON((-75 18.369949, -75 18, -76 18,
    -76 18.5, -75 18.5, -75 18.4512877,
    -75.0161269 18.4598496, -75.0386021 18.4609199, -75.0632177 18.4534282,
    -75.0803417 18.4320233, -75.0728079 18.4136718, -75.0632177 18.3988456,
    -75.0482343 18.388143200000002, -75.0311104 18.3795812, -75 18.369949)), 
  POLYGON((-74 18.078307, -74 18, -75 18,
    -75 18.369949, -74.9829493 18.3710192, -74.9647551 18.3806515,
    -74.9615444 18.4138291, -74.9658254 18.4330935, -74.9797386 18.443796,
    -75 18.4512877, -75 18.5, -74.582385582572286 18.5,
    -74.6062228 18.3967051, -74.51525199999999 18.2875401, -74.3696986 18.178375,
    -74.1968539 18.1328896, -74 18.078307)), 
  POLYGON((-73 18.1003884, -73 18, -74 18,
    -74 18.078307, -73.9467099 18.0216138, -73.8871872 18.0099996,
    -73.8087914 18.0085478, -73.7695935 18.0107255, -73.5926717 18.0249307,
    -73.558495999999991 18.0334746, -73.5243203 18.05398, -73.4884358 18.0796118,
    -73.4508425 18.1445457, -73.3927438 18.1735951, -73.295343 18.1684687,
    -73.1893983 18.1616336, -73.0902887 18.1360018, -73 18.1003884)), 
  POLYGON((-72 18.1147631, -72 18, -73 18,
    -73 18.1003884, -72.8774219 18.0879792, -72.7905575 18.1034907,
    -72.722307 18.1221045, -72.6695679 18.1345137, -72.5827036 18.1283091,
    -72.5237599 18.1283091, -72.4679186 18.1624344, -72.4089749 18.1500252,
    -72.2569623 18.1500252, -72.1111543 18.1469229, -72 18.1147631)), 
  GEOMETRYCOLLECTION(
  POLYGON((-71.8527291 18, -72 18, -72 18.1147631,
    -71.8527291 18)), 
    POLYGON((-71 18.1947594, -71 18, -71.0705438 18,
      -71.0261015 18.1206888, -71 18.1947594))), 
  POLYGON((-70 18.206611940813044, -70 18, -71 18,
    -71 18.1947594, -70.9109057 18.1801528, -70.7950014 18.2895698,
    -70.7298193 18.3162352, -70.6764885 18.3369749, -70.6438974 18.3517891,
    -70.6231577 18.3429006, -70.6231577 18.3043839, -70.617232 18.2569787,
    -70.6113064 18.2303133, -70.6290833 18.2095736, -70.5431614 18.1473543,
    -70.4927935 18.1562427, -70.3920575 18.1829081, -70.350578 18.1651312,
    -70.2142881 18.1651312, -70.1283662 18.1740197, -70.0000004 18.2066108,
    -70 18.206611940813044)), 
  POLYGON((-69 18.2066108, -69 18, -70 18,
    -70 18.206611940813044, -69.9594854 18.3221609, -69.9387456 18.354751999999998,
    -69.8794892 18.3843802, -69.8054186 18.387342999999998, -69.731348 18.3814174,
    -69.6750544 18.3577148, -69.5724914 18.3541337, -69.396549 18.3666033,
    -69.2780361 18.3754917, -69.180263 18.3458635, -69.0854526 18.304384,
    -69 18.2066108)), 
  POLYGON((-68.251888572156844 18.5, -68 18.5, -68 18,
    -69 18, -69 18.2066108, -68.8158358 18.1118005,
    -68.7269511 18.070321, -68.6765831 18.0643953, -68.5965869 18.070321,
    -68.5254791 18.0880979, -68.4988137 18.1058748, -68.5047394 18.1473544,
    -68.510665 18.1621685, -68.5373304 18.1799454, -68.5906612 18.1888339,
    -68.6054753 18.2125365, -68.5876984 18.2510532, -68.5639958 18.2717929,
    -68.5551074 18.2806814, -68.4839996 18.2806814, -68.4454829 18.2836442,
    -68.4217803 18.3132725, -68.3654867 18.3754917, -68.3180815 18.4317854,
    -68.2617879 18.4703021, -68.251888572156844 18.5)), 
  GEOMETRYCOLLECTION(
  POLYGON((-67.902978153284678 18, -68 18, -68 18.5,
    -67.215620840929233 18.5, -67.216495 18.497996999999998, -67.221627 18.497996999999998,
    -67.222364 18.494062, -67.222720999999993 18.489819999999998, -67.222495999999992 18.486848,
    -67.222467 18.482727, -67.22134299999999 18.476157999999998, -67.217131999999992 18.461412,
    -67.212908 18.451262, -67.207938 18.44228, -67.211975 18.440042,
    -67.216253999999992 18.437079, -67.218246 18.436087999999998, -67.222544 18.433388,
    -67.230535 18.429126999999998, -67.234047 18.427737999999998, -67.236655 18.426395,
    -67.244514 18.421346, -67.248066 18.419909999999998, -67.249151 18.419214,
    -67.249625 18.418574, -67.280492999999993 18.412064, -67.284765999999991 18.4099,
    -67.290106 18.407735, -67.297049 18.404491999999998, -67.302718 18.40156,
    -67.307476 18.397291, -67.314135999999991 18.388806, -67.318671999999992 18.381035999999998,
    -67.321332 18.373001, -67.32323 18.363875999999998, -67.323392 18.357696999999998,
    -67.322032 18.351416999999998, -67.3199 18.343187, -67.316037 18.334217,
    -67.311849 18.327562999999998, -67.306609 18.318087, -67.30051499999999 18.310055,
    -67.299387 18.309012, -67.298863 18.306929999999998, -67.293804 18.296412999999998,
    -67.290184 18.289490999999998, -67.28596499999999 18.282165, -67.281848 18.275779999999997,
    -67.275933999999992 18.267746, -67.269731 18.261392, -67.261333 18.256041,
    -67.253495 18.252302, -67.249625 18.250685, -67.249625 18.250874,
    -67.247453 18.249893999999998, -67.241948 18.248773, -67.237174 18.247522999999997,
    -67.230465 18.246561, -67.230182 18.245105, -67.227992 18.239596,
    -67.227293 18.236955, -67.22697 18.234534, -67.226293 18.231886,
    -67.225061 18.228645999999998, -67.223333 18.224997, -67.221237 18.221216,
    -67.216053 18.213, -67.215186 18.211911, -67.214237 18.211349,
    -67.216584 18.209106, -67.218401 18.20675, -67.220789 18.202645999999998,
    -67.222433 18.201107999999998, -67.225664 18.196154, -67.230032999999992 18.190288,
    -67.232728999999992 18.184642999999998, -67.235175 18.178285, -67.235896 18.175704,
    -67.236096 18.173545, -67.236409 18.172531, -67.237252 18.16765,
    -67.237208999999993 18.164918, -67.236744 18.159632, -67.236261 18.15786,
    -67.235826 18.15689, -67.236204 18.15222, -67.23810499999999 18.147320999999998,
    -67.239015999999992 18.141873, -67.238914 18.138581, -67.238265 18.134793,
    -67.237359 18.131491, -67.235439 18.127492, -67.235146 18.124302,
    -67.235177999999991 18.121304, -67.235374 18.120145, -67.236325999999991 18.118665999999997,
    -67.238557 18.116350999999998, -67.242764999999991 18.115213999999998, -67.246313 18.113264,
    -67.249625 18.110114, -67.253412 18.104647, -67.257426 18.095743,
    -67.260189 18.084802, -67.260614 18.075212, -67.258878 18.066671,
    -67.255365 18.057672, -67.259343 18.049236999999998, -67.261326 18.038702,
    -67.261417 18.025493, -67.259914999999992 18.01313, -67.262761 18.00574,
    -67.264282254025034 18, -67.873264434622456 18, -67.871392 18.000314,
    -67.862583 18.003225999999998, -67.83847999999999 18.013851, -67.830056 18.0196,
    -67.822442 18.026431, -67.81532 18.035763, -67.802838 18.050739,
    -67.797336 18.05865, -67.79356 18.067100999999997, -67.791512 18.074672,
    -67.790449999999993 18.082109, -67.789811 18.090357, -67.79040599999999 18.098468,
    -67.791986999999992 18.107459, -67.794592 18.115875, -67.797832 18.12301,
    -67.799457 18.128059, -67.805323 18.136903, -67.81125999999999 18.143885,
    -67.819071 18.151208, -67.825752 18.157785, -67.834413 18.163653999999998,
    -67.843322 18.167524999999998, -67.852443999999991 18.170111, -67.860822 18.170833,
    -67.86831699999999 18.169963, -67.874631 18.168782, -67.879959 18.166971,
    -67.883944 18.164997, -67.885565 18.163843, -67.892674 18.166785,
    -67.894933999999992 18.166674, -67.89761399999999 18.175946, -67.901491 18.184065,
    -67.908763 18.193113999999998, -67.91617 18.200105, -67.927213999999992 18.206269,
    -67.936942 18.209366, -67.944788 18.210041, -67.955918 18.210228,
    -67.964247 18.207862, -67.973664 18.203601, -67.98254399999999 18.197485,
    -67.988598 18.191447999999998, -67.994508 18.185041, -67.997243 18.179897,
    -67.998691999999991 18.170776, -67.998752 18.159029, -67.997917 18.144012,
    -67.996681 18.135071999999997, -67.995139999999992 18.129206, -67.98989499999999 18.12301,
    -67.988338 18.120905999999998, -67.986149 18.119193, -67.986932 18.112066,
    -67.986921 18.109298, -67.99031699999999 18.104591, -67.993081 18.100054999999998,
    -67.99591199999999 18.094913, -67.997826 18.089334, -67.998297 18.083761,
    -67.996175 18.072292, -67.993422 18.062407999999998, -67.992116 18.059683,
    -67.988005 18.052205, -67.982844 18.045813, -67.976136 18.039593999999997,
    -67.970704 18.035768, -67.965132 18.032211999999998, -67.961283 18.028447,
    -67.953558 18.022638, -67.945835 18.017537, -67.935441 18.012953,
    -67.92942 18.011288, -67.927266 18.008696999999998, -67.92293 18.00608,
    -67.914687 18.002906, -67.902978153284678 18),
    (-67.480902 18.325995, -67.472137 18.326088, -67.460329 18.328734,
      -67.453999 18.331264, -67.445228 18.336444999999998, -67.438436 18.341217999999998,
      -67.431570999999991 18.34733, -67.425624 18.354511, -67.421973999999992 18.361753,
      -67.419918 18.36668, -67.418044 18.373, -67.41757299999999 18.380133999999998,
      -67.418627 18.389964, -67.4213 18.398509, -67.425029999999992 18.406851,
      -67.434096 18.417165, -67.442799 18.424245, -67.451641 18.430585999999998,
      -67.462032 18.435741999999998, -67.473015 18.438136, -67.481416 18.438616,
      -67.491856 18.437306, -67.49963 18.435313, -67.506867 18.432478,
      -67.515363 18.427554999999998, -67.523928 18.421758, -67.531784 18.414483999999998,
      -67.537492 18.406337999999998, -67.541121 18.396447, -67.543025 18.386124,
      -67.54243799999999 18.377651999999998, -67.541581 18.373, -67.540526 18.368767,
      -67.538485999999992 18.364477, -67.535989 18.358279, -67.528322 18.349128999999998,
      -67.523504 18.34457, -67.518334 18.340448, -67.513514 18.337129,
      -67.507218999999992 18.333641999999998, -67.502295 18.33116, -67.494261 18.327896,
      -67.488607 18.326708, -67.480902 18.325995)), 
    POLYGON((-67.889023108433733 18, -67.896644617594248 18, -67.891781 18.000595,
      -67.889023108433733 18)), 
    POLYGON((-67.87463 18, -67.877615415631837 18, -67.87463 18.000826999999997,
      -67.87463 18))), 
  POLYGON((-65 18.252640750111357, -65 18, -65.783619038319813 18,
    -65.782215 18.002976999999998, -65.781458 18.00949, -65.781508 18.014495999999998,
    -65.782383 18.018998, -65.774175 18.021493, -65.765965999999992 18.025033,
    -65.761072 18.027525999999998, -65.756177 18.031523999999997, -65.75209 18.035521,
    -65.749605 18.038791, -65.748573999999991 18.040938999999998, -65.746921 18.046559,
    -65.74622 18.050643, -65.744633999999991 18.052861999999998, -65.742239 18.057363,
    -65.740057 18.062544, -65.739113 18.070778999999998, -65.736361 18.071949999999998,
    -65.731881 18.07534, -65.726874 18.080433, -65.723666 18.085041,
    -65.720812 18.090462, -65.718313 18.097141, -65.716623 18.102318999999998,
    -65.716098 18.106267, -65.713874 18.107129, -65.703921999999991 18.112316,
    -65.699334999999991 18.116114, -65.696018999999993 18.119667, -65.691833 18.123018,
    -65.690489 18.124568, -65.689426 18.126404, -65.684065 18.125947,
    -65.680961 18.125871, -65.677539 18.126065999999998, -65.673975 18.126738,
    -65.670694 18.126932999999998, -65.66653 18.127739, -65.662573999999992 18.129396,
    -65.654727999999992 18.133336, -65.624605 18.150329, -65.623322 18.150743,
    -65.621814 18.151588, -65.618051 18.154180999999998, -65.620398 18.148865999999998,
    -65.623627 18.144787, -65.62859 18.134076, -65.630389999999991 18.128964,
    -65.630974 18.123018, -65.631112 18.116521, -65.630718 18.107476,
    -65.630678 18.098667, -65.629331 18.089254, -65.627492 18.083413,
    -65.624603999999991 18.077472999999998, -65.620251 18.070086, -65.615657 18.064048,
    -65.61053 18.058079, -65.600448 18.049288999999998, -65.594803 18.045082,
    -65.588624 18.041081, -65.574334999999991 18.036334999999998, -65.568787 18.034869,
    -65.556213 18.030217, -65.5497 18.02845, -65.540196999999992 18.027642,
    -65.532063 18.028385999999998, -65.524329 18.029704, -65.514608 18.032484,
    -65.507708999999991 18.035422, -65.50230599999999 18.038016, -65.499603 18.040903999999998,
    -65.49757 18.041867, -65.495864 18.042021, -65.486657 18.038456,
    -65.478227 18.036876, -65.470736 18.036763, -65.464007 18.037036,
    -65.461984 18.037335, -65.45602199999999 18.035985, -65.447962 18.035922,
    -65.440302 18.037174999999998, -65.432232 18.039018, -65.425045 18.042875,
    -65.419209 18.043221, -65.411288 18.045298, -65.4016 18.048958,
    -65.390996 18.056389, -65.386586 18.055636, -65.377274 18.055473,
    -65.374602 18.055588999999998, -65.361535 18.058308, -65.350501999999992 18.062037999999998,
    -65.348153 18.061488999999998, -65.342581 18.06112, -65.333914 18.062189999999998,
    -65.324703 18.064666, -65.318434 18.067171, -65.312226 18.071143,
    -65.30445499999999 18.070733, -65.296069 18.071797, -65.286185 18.074382,
    -65.278258999999991 18.077454, -65.272386 18.080396999999998, -65.269335 18.080299,
    -65.262753 18.081032999999998, -65.255046 18.082428, -65.249601 18.084305999999998,
    -65.243451999999991 18.087881, -65.236269 18.092375, -65.229333 18.098578,
    -65.223248 18.105225, -65.217901 18.116415999999997, -65.216191 18.123019,
    -65.215413 18.129237, -65.21509 18.136257999999998, -65.216256 18.143553,
    -65.21869199999999 18.152444, -65.222909 18.160386, -65.227246 18.165886999999998,
    -65.232013 18.170845, -65.237175 18.174920999999998, -65.242413 18.178047,
    -65.249601 18.184772, -65.252794 18.187165, -65.264399 18.193302,
    -65.271987 18.196244, -65.278317 18.197917999999998, -65.286968 18.198442999999997,
    -65.296736 18.197936, -65.303097 18.199338, -65.309231 18.199761,
    -65.316334 18.198715, -65.320848 18.199769, -65.328785 18.19993,
    -65.331355 18.201145999999998, -65.338974 18.203713, -65.347864 18.205924,
    -65.356776 18.207355, -65.394742999999991 18.211743, -65.406874 18.21099,
    -65.421852 18.209082, -65.431504 18.208672, -65.438102 18.214112,
    -65.448758 18.219151999999998, -65.464424999999991 18.222215, -65.475341 18.221038999999998,
    -65.486486 18.218852, -65.491856 18.216418, -65.510628 18.21391,
    -65.524238 18.208515, -65.53527 18.201294999999998, -65.538913 18.197079,
    -65.554645 18.184768, -65.54903 18.197696, -65.544018999999992 18.205205,
    -65.541735 18.210366999999998, -65.533841 18.218358, -65.520352 18.219988999999998,
    -65.511965 18.223029999999998, -65.499603999999991 18.224451, -65.499603999999991 18.222158,
    -65.48938 18.225299, -65.481261 18.229273, -65.47443 18.233394999999998,
    -65.465369 18.243678, -65.460049 18.254181, -65.458227 18.260614,
    -65.457037 18.268141, -65.456565 18.275164, -65.457725 18.283185,
    -65.461541 18.293300000000002, -65.448947 18.292037999999998, -65.437752 18.290345,
    -65.427712 18.290699, -65.423774999999992 18.291377, -65.416574 18.287001999999998,
    -65.413192 18.282898, -65.403909 18.276434, -65.395304 18.273234,
    -65.383715 18.268580999999998, -65.378052 18.267274, -65.374603999999991 18.264839,
    -65.366547 18.257582, -65.360399 18.253436999999998, -65.348976 18.248013999999998,
    -65.341216 18.245131999999998, -65.333758 18.243983999999998, -65.325938 18.244601,
    -65.320959 18.238782, -65.315578 18.234492, -65.306653 18.230081,
    -65.297224 18.227403, -65.286681 18.225991, -65.273395 18.228202,
    -65.262987 18.231016999999998, -65.25649 18.234794, -65.252926 18.238023,
    -65.249601 18.236881, -65.243028999999993 18.235512999999997, -65.235963 18.235132999999998,
    -65.227576 18.23577, -65.211489 18.241283, -65.205432 18.244702999999998,
    -65.201718 18.248013999999998, -65.200643 18.248013999999998, -65.199463999999992 18.249036,
    -65.194926999999993 18.253839, -65.191135 18.258884, -65.187761999999992 18.266416,
    -65.184857 18.272045, -65.180436 18.279168, -65.175405 18.286683,
    -65.171674 18.294213, -65.168826 18.303517, -65.168504 18.313347999999998,
    -65.171632 18.33344, -65.174668 18.340467999999998, -65.178522 18.347841,
    -65.179116999999991 18.350669, -65.181761999999992 18.357727999999998, -65.185619 18.364387,
    -65.190583 18.370372, -65.192388 18.373010999999998, -65.19514 18.377678,
    -65.199848 18.382566, -65.208484 18.389371, -65.214603 18.392711,
    -65.220944 18.395272, -65.229467 18.397043, -65.235832 18.397734,
    -65.244203 18.397496, -65.249601 18.396369, -65.260766 18.392343,
    -65.2756 18.390556, -65.282245 18.388975, -65.286374999999992 18.387341,
    -65.302922999999993 18.387503, -65.306642 18.390795, -65.311230999999992 18.395488999999998,
    -65.316219 18.399404, -65.322626 18.403295, -65.328797999999992 18.40596,
    -65.334516999999991 18.407975, -65.339609 18.408966, -65.351624 18.415447,
    -65.356386 18.417488, -65.362567 18.419472, -65.369181 18.420643,
    -65.376657999999992 18.421484, -65.384593 18.42103, -65.392904 18.418947,
    -65.400586 18.415599, -65.408244 18.411262999999998, -65.414155999999991 18.406465999999998,
    -65.418609 18.40121, -65.421413 18.396411999999998, -65.424334 18.390867,
    -65.430431 18.391683, -65.439983 18.393285, -65.44494 18.39385,
    -65.456004 18.394413, -65.463978 18.39348, -65.472104 18.398159,
    -65.479126999999991 18.399255, -65.482564 18.401639, -65.487014 18.402945,
    -65.492505 18.403685, -65.499605 18.407166999999998, -65.526933 18.419477,
    -65.548600999999991 18.428653999999998, -65.566293 18.435705, -65.58668 18.443333,
    -65.608280999999991 18.44832, -65.624606 18.451261, -65.626100000000008 18.45106,
    -65.630085 18.450082, -65.635142 18.448152999999998, -65.64030799999999 18.44568,
    -65.64608299999999 18.442221, -65.653189 18.435599, -65.657915 18.428905,
    -65.662076 18.428131, -65.66635 18.426164999999997, -65.673976 18.421927,
    -65.678613 18.418533, -65.681508 18.422214, -65.685715 18.426204,
    -65.691377 18.430740999999998, -65.699102 18.434703, -65.709814 18.438430999999998,
    -65.713802 18.443984999999998, -65.715811 18.447222, -65.723664 18.454449,
    -65.732700999999992 18.459568, -65.743307 18.463805, -65.745581 18.463707,
    -65.749608 18.4647, -65.749608 18.46951, -65.751463 18.470841,
    -65.75358 18.472004, -65.757204 18.473204, -65.766013 18.474799,
    -65.771843 18.475555, -65.77732 18.476605, -65.782729 18.477256999999998,
    -65.787838 18.477515999999998, -65.800912 18.477601, -65.805314 18.478253,
    -65.809058 18.479338, -65.819501 18.482844, -65.832698 18.486204,
    -65.847172 18.487531999999998, -65.85063199999999 18.488069, -65.85772 18.487249,
    -65.866927 18.493375, -65.876995 18.49955, -65.878222909147482 18.5,
    -65 18.5, -65 18.445338689631171, -65.002464 18.444212999999998,
    -65.007769 18.440703, -65.017299 18.449323, -65.030382 18.455375999999998,
    -65.041522 18.459101999999998, -65.060665 18.459747, -65.080567 18.452807,
    -65.091791 18.447101999999997, -65.100413 18.439543999999998, -65.107299 18.428608999999998,
    -65.112214999999992 18.418827, -65.11376 18.406875, -65.113057 18.395758,
    -65.114887 18.391445, -65.11627399999999 18.386603, -65.117426999999992 18.380233,
    -65.122858999999991 18.373010999999998, -65.124597999999992 18.369303, -65.130348 18.361535999999997,
    -65.135139 18.350854, -65.139929999999993 18.336289999999998, -65.139929999999993 18.323666,
    -65.138014 18.311042, -65.142804 18.302303, -65.147596 18.294535,
    -65.153346 18.279971, -65.157179 18.267348, -65.159095 18.254725,
    -65.159095 18.243074, -65.150471 18.231420999999997, -65.13705 18.226566,
    -65.120758 18.225595, -65.100628 18.225593999999997, -65.092962 18.226565,
    -65.083375 18.228507, -65.074749 18.233362, -65.06804 18.240159,
    -65.063988 18.248010999999998, -65.059054 18.254151999999998, -65.051765 18.267927,
    -65.051052 18.269858, -65.048581 18.266776999999998, -65.041061 18.262441,
    -65.031565 18.256277, -65.019821 18.252354999999998, -65.008572 18.251564,
    -65 18.252640750111357)), 
  POLYGON((-65 18.445338689631171, -65 18.5, -64 18.5,
    -64 18, -65 18, -65 18.252640750111357,
    -64.999591999999993 18.252692, -64.997861 18.252492999999998, -64.98924199999999 18.255174999999998,
    -64.98481 18.256161, -64.980414 18.256912, -64.978022 18.256359,
    -64.973942999999991 18.256134, -64.9674 18.256736999999998, -64.960964 18.256059999999998,
    -64.954317 18.256123, -64.948371999999992 18.25683, -64.946472 18.257223,
    -64.945419 18.254556, -64.942061 18.248010999999998, -64.939602 18.244557,
    -64.933714 18.23797, -64.926904 18.232592999999998, -64.921033999999992 18.228842999999998,
    -64.913228 18.225858, -64.906151 18.224193, -64.905239999999992 18.218514,
    -64.90296 18.212995, -64.899774999999991 18.206525, -64.895499 18.200487,
    -64.889752 18.194236, -64.886403 18.191412, -64.874586 18.184525,
    -64.872534 18.183484, -64.86498499999999 18.18091, -64.858034 18.179586999999998,
    -64.849550999999991 18.179683, -64.841667 18.180927, -64.833351999999991 18.183455,
    -64.82485 18.187976, -64.818514999999991 18.192059999999998, -64.811814 18.198712,
    -64.806889 18.205161999999998, -64.80277 18.213371, -64.799962 18.222595,
    -64.799258 18.229319999999998, -64.799511 18.236485, -64.801285 18.245103999999998,
    -64.792514 18.248010999999998, -64.786984 18.250557, -64.782889 18.252692,
    -64.778757 18.255468, -64.773419 18.260506, -64.770741 18.260278,
    -64.764079999999993 18.260431, -64.757769 18.260855, -64.749580999999992 18.262743999999998,
    -64.748736 18.263059, -64.743209999999991 18.260725, -64.740747 18.259470999999998,
    -64.736549 18.258571, -64.729475 18.254130999999997, -64.724198 18.251963,
    -64.720174 18.25069, -64.714838 18.249565, -64.707906999999992 18.248725,
    -64.69448899999999 18.248758, -64.68804 18.249897999999998, -64.681665999999993 18.251575,
    -64.678229 18.252316, -64.671238 18.256154, -64.664259 18.261271,
    -64.661301 18.267359, -64.663856 18.285376, -64.651225 18.300915,
    -64.637734999999992 18.319758999999998, -64.638524 18.320491999999998, -64.639032 18.321597999999998,
    -64.643689 18.344106999999997, -64.639144 18.35519, -64.640436999999991 18.363996999999998,
    -64.645202 18.364456, -64.658667 18.364729999999998, -64.66436 18.365685,
    -64.668481 18.365577, -64.672113 18.366484, -64.675249 18.367463,
    -64.681191 18.371982, -64.683292999999992 18.372567, -64.687522 18.372761,
    -64.691101 18.371883999999998, -64.701604 18.370635999999998, -64.701893 18.371138,
    -64.706147 18.373008, -64.708197 18.373742, -64.71882699999999 18.373884,
    -64.720022 18.374395, -64.727025 18.374724999999998, -64.729829999999993 18.377945,
    -64.73765 18.376217, -64.738812 18.376255, -64.741025 18.377378,
    -64.743804 18.378131, -64.747325 18.377671, -64.749051999999992 18.377069,
    -64.749580999999992 18.377307, -64.751961999999992 18.377449, -64.756923 18.376686,
    -64.778005 18.386209, -64.785001 18.403802, -64.800328999999991 18.407595,
    -64.816673999999992 18.402905999999998, -64.830617 18.395201999999998, -64.841355 18.393708,
    -64.845924 18.393719, -64.85995 18.394225, -64.864976 18.395927,
    -64.87391199999999 18.406254999999998, -64.87384 18.40683, -64.874585 18.409739,
    -64.874585 18.410694, -64.891033 18.46216, -64.895023999999992 18.463364,
    -64.899855 18.46447, -64.907595 18.464983999999998, -64.914498 18.464481,
    -64.921050999999991 18.463805999999998, -64.927641999999992 18.461949, -64.933043 18.459612,
    -64.93848 18.456837, -64.942692 18.454157, -64.946097999999992 18.451304,
    -64.950454999999991 18.446293999999998, -64.959589 18.449075, -64.969182 18.450947,
    -64.974926 18.451279, -64.982389 18.450708, -64.987505 18.449721,
    -64.99959 18.445525999999997, -65 18.445338689631171)), 
  POLYGON((-63.2499749 18, -64 18, -64 18.5,
    -63 18.5, -63 18.37642757338055, -63.0403792 18.3497365,
    -63.1486703 18.290350999999998, -63.211549 18.2938442, -63.2220288 18.3252836,
    -63.2429884 18.3287769, -63.2849076 18.3287769, -63.2953873 18.2728847,
    -63.263948 18.2274723, -63.2185356 18.2100059, -63.2010692 18.1960329,
    -63.2569615 18.1331542, -63.2744278 18.0807552, -63.2499749 18)), 
  POLYGON((-63 18.37642757338055, -63 18.5, -62 18.5,
    -62 18, -62.7294788 18, -62.744713999999995 18.0627574,
    -62.7564012 18.1705397, -62.7667899 18.2289758, -62.7823728 18.287412,
    -62.8122402 18.3354595, -62.8823636 18.3640283, -62.9988791 18.3771685,
    -63 18.37642757338055)), 
  POLYGON((-62 18, -62 18.5, -61 18.5,
    -61 18, -62 18)), 
  POLYGON((-61 18, -61 18.5, -60 18.5,
    -60 18, -61 18)), 
  POLYGON((-60 18, -60 18.5, -59 18.5,
    -59 18, -60 18)), 
  POLYGON((-59 18, -59 18.5, -58 18.5,
    -58 18, -59 18)), 
  POLYGON((-58 18, -58 18.5, -57 18.5,
    -57 18, -58 18)), 
  POLYGON((-57 18, -57 18.5, -56 18.5,
    -56 18, -57 18)), 
  POLYGON((-56 18, -56 18.5, -55 18.5,
    -55 18, -56 18)), 
  POLYGON((-85.989814006321723 19, -85 19, -85 18.5,
    -85.9942782647015 18.5, -85.989814006321723 19)), 
  POLYGON((-85 18.5, -85 19, -84 19,
    -84 18.5, -85 18.5)), 
  POLYGON((-84 18.5, -84 19, -83 19,
    -83 18.5, -84 18.5)), 
  POLYGON((-83 18.5, -83 19, -82 19,
    -82 18.5, -83 18.5)), 
  POLYGON((-82 18.5, -82 19, -81 19,
    -81 18.5, -82 18.5)), 
  POLYGON((-81 18.5, -81 19, -80 19,
    -80 18.5, -81 18.5)), 
  POLYGON((-80 18.5, -80 19, -79 19,
    -79 18.5, -80 18.5)), 
  POLYGON((-78.413902747395184 18.5, -79 18.5, -79 19,
    -78 19, -78 18.581697259789191, -78.126245 18.5420253,
    -78.2015875 18.5508892, -78.3035216 18.5464573, -78.4054556 18.50657,
    -78.413902747395184 18.5)), 
  POLYGON((-78 18.581697259789191, -78 19, -77 19,
    -77 18.524156688171981, -77.0006172 18.5243213, -77.1202004 18.5021381,
    -77.2398621 18.5198658, -77.3728195 18.5464573, -77.4658897 18.5553211,
    -77.5811195 18.5553211, -77.6919174 18.559753, -77.7805557 18.5863445,
    -77.9489684 18.5819126, -77.9993612 18.581898, -78 18.581697259789191)), 
  POLYGON((-77 18.524156688171981, -77 19, -76 19,
    -76 18.5, -76.913255004117858 18.5, -76.93406 18.50657,
    -77 18.524156688171981)), 
  POLYGON((-76 18.5, -76 19, -75 19,
    -75 18.5, -76 18.5)), 
  POLYGON((-74.582385582572286 18.5, -75 18.5, -75 19,
    -74 19, -74 18.8151712771192, -74.0000002 18.8151712,
    -74.205951 18.7787828, -74.2332422 18.7787828, -74.3515044 18.796977,
    -74.4515723 18.7696857, -74.5425432 18.7151032, -74.5607374 18.6059381,
    -74.5789316 18.5149673, -74.582385582572286 18.5)), 
  POLYGON((-74 18.8151712771192, -74 19, -73.6601257 19,
    -73.7032204 18.9296083, -74 18.8151712771192)), 
  POLYGON((-68.562893210988861 19, -68 19, -68 18.5,
    -68.251888572156844 18.5, -68.2499366 18.5058559, -68.226233999999991 18.5532611,
    -68.226233999999991 18.6065919, -68.2469738 18.689550999999998, -68.3210443 18.7369561,
    -68.3625239 18.7902869, -68.4454829 18.8702832, -68.5077022 18.923614,
    -68.562893210988861 19)), 
  POLYGON((-67.215620840929233 18.5, -68 18.5, -68 19,
    -67 19, -67 18.558552590597817, -67.005027 18.560098,
    -67.017094 18.562646, -67.045947 18.56536, -67.052647999999991 18.565642999999998,
    -67.059356 18.566323999999998, -67.06401799999999 18.566579, -67.069762 18.567277,
    -67.08826599999999 18.568, -67.092615999999992 18.567788, -67.096733 18.567244,
    -67.104835 18.566575999999998, -67.112237999999991 18.564883, -67.116891999999993 18.564104999999998,
    -67.121263 18.562962, -67.124625 18.561777, -67.130718 18.560762,
    -67.139162 18.55895, -67.14547 18.556787999999997, -67.151811999999993 18.555167,
    -67.160467 18.552543999999997, -67.168601 18.548468, -67.175508999999991 18.54371,
    -67.184418999999991 18.536396, -67.188559 18.532434, -67.190451 18.531498,
    -67.195151 18.527742, -67.200026 18.523277, -67.205745 18.517129,
    -67.209818 18.511039999999998, -67.214034 18.503636, -67.215620840929233 18.5)), 
  POLYGON((-67 18.558552590597817, -67 19, -66 19,
    -66 18.515726292517009, -66.00402 18.515451, -66.008372 18.514846,
    -66.015563 18.512849, -66.018153 18.511691, -66.021545 18.511717,
    -66.025854 18.512054, -66.028706 18.513220999999998, -66.032305999999991 18.513787999999998,
    -66.041117 18.516651, -66.045985 18.517665, -66.056406 18.518318,
    -66.063121999999993 18.518167, -66.067751 18.518235999999998, -66.074861 18.517616999999998,
    -66.079906 18.518766, -66.084495 18.51944, -66.089686 18.519917,
    -66.094601 18.519617, -66.100151 18.519558, -66.104355 18.51969,
    -66.107646 18.519209, -66.109335 18.520131, -66.115685 18.521592,
    -66.119751999999991 18.521352, -66.120666 18.521964999999998, -66.129348 18.524513,
    -66.149058 18.525181, -66.166011 18.525156, -66.170469 18.525520999999998,
    -66.175462 18.525081999999998, -66.182818 18.526138, -66.194071999999991 18.526512,
    -66.201441 18.525111, -66.207501999999991 18.523466, -66.212111999999991 18.521845,
    -66.218689 18.523063999999998, -66.22547999999999 18.523947, -66.231673 18.524119,
    -66.235533 18.523705, -66.242282 18.526069, -66.249612 18.527665,
    -66.254345 18.528284, -66.260072999999991 18.529443999999998, -66.265526999999992 18.529761999999998,
    -66.271124 18.529709999999998, -66.283098999999993 18.528631999999998, -66.286205999999993 18.527396,
    -66.289898 18.527943999999998, -66.295914 18.528295999999997, -66.30246 18.528281,
    -66.304569 18.528724, -66.310647 18.530452999999998, -66.31325 18.530694999999998,
    -66.318128 18.532792999999998, -66.32245 18.534049, -66.333084 18.538617,
    -66.338809 18.54038, -66.342995 18.540861, -66.353518 18.541155,
    -66.359186999999991 18.541, -66.366305 18.539906, -66.374613 18.541042,
    -66.3767 18.540924, -66.378513 18.540568, -66.383473999999993 18.541852,
    -66.390494 18.543051, -66.39965 18.543761999999997, -66.415984999999992 18.543620999999998,
    -66.419609 18.543279, -66.4237 18.542402, -66.429778 18.541574999999998,
    -66.432985 18.540219999999998, -66.436212 18.540817, -66.440082 18.540746,
    -66.444312 18.540273, -66.452951 18.53896, -66.458041999999992 18.537249,
    -66.463636 18.534802, -66.468237 18.532383, -66.472165 18.529587,
    -66.476381 18.526322, -66.478562 18.524421, -66.489655 18.527024,
    -66.495013 18.527466999999998, -66.497136 18.528391, -66.507485 18.531586,
    -66.515152 18.533752, -66.524523 18.535356, -66.533126 18.535474,
    -66.540088 18.537332, -66.546207 18.538209, -66.548857 18.538964999999997,
    -66.556602 18.540255, -66.562935 18.540931, -66.566441 18.540715,
    -66.573022 18.541290999999998, -66.577552 18.541518, -66.582512 18.54104,
    -66.587086 18.539953999999998, -66.591784 18.541731, -66.596946 18.542904999999998,
    -66.602217 18.543573, -66.607526 18.543838, -66.61231 18.543055,
    -66.614576 18.542932999999998, -66.620342 18.543805, -66.626024 18.544147,
    -66.634761 18.543993, -66.642333999999991 18.54353, -66.648847 18.542692,
    -66.655858 18.541387, -66.660359 18.5398, -66.663083 18.539747,
    -66.665735 18.539994999999998, -66.669311999999991 18.539108, -66.680206 18.539265,
    -66.687992 18.538331, -66.693941 18.537153, -66.699502 18.535738,
    -66.703439 18.53398, -66.705427 18.532614, -66.715803 18.53035,
    -66.720905 18.528696999999998, -66.725681 18.528485, -66.731347 18.527137,
    -66.734607 18.526011999999998, -66.74089 18.528423999999998, -66.745376999999991 18.529453,
    -66.75182 18.533317999999998, -66.755945 18.535159999999998, -66.763457 18.537966,
    -66.770445 18.540029, -66.786020999999991 18.542385, -66.795172 18.543319999999998,
    -66.804219 18.543784, -66.814225 18.543682999999998, -66.822960999999992 18.543104,
    -66.831347 18.541783, -66.836694 18.540205999999998, -66.843932999999993 18.541664,
    -66.851175 18.542517, -66.86196 18.542385, -66.865105 18.542607,
    -66.874623 18.542230999999997, -66.883642999999992 18.541265, -66.888550999999993 18.540214,
    -66.894802 18.538560999999998, -66.899563 18.536756, -66.902086 18.535238,
    -66.908102 18.533296999999997, -66.913254999999992 18.53453, -66.918331 18.53536,
    -66.931786 18.538933999999998, -66.937675 18.540049, -66.94376299999999 18.540452,
    -66.948015 18.540391, -66.951487 18.54001, -66.953964 18.540931,
    -66.957948 18.541617, -66.968657999999991 18.542991999999998, -66.970711999999992 18.543822,
    -66.975039999999993 18.546048, -66.980166 18.547684, -66.984496 18.550010999999998,
    -66.995516 18.554880999999998, -66.999624 18.556196999999997, -66.999624 18.558436999999998,
    -67 18.558552590597817)), 
  POLYGON((-66 18.515726292517009, -66 19, -65 19,
    -65 18.5, -65.878222909147482 18.5, -65.88138 18.501157,
    -65.890132 18.503287, -65.899406 18.504732999999998, -65.907957 18.505374,
    -65.916105 18.505173, -65.924159 18.503550999999998, -65.930866 18.501844,
    -65.932589 18.501044, -65.938881 18.502219999999998, -65.948138 18.502648,
    -65.954467999999991 18.506159, -65.959392 18.508195999999998, -65.972822 18.512328999999998,
    -65.978764 18.513779, -65.983988 18.514594, -65.987675 18.515818,
    -65.993402 18.516598, -65.997507 18.516166, -65.99961 18.515753,
    -66 18.515726292517009)), 
  POLYGON((-65 18.5, -65 19, -64 19,
    -64 18.5, -65 18.5)), 
  POLYGON((-64 18.5, -64 19, -63 19,
    -63 18.5, -64 18.5), 
    (-63.467071999999995 18.6225604, -63.4722796 18.612889, -63.4797192 18.6013577, 
      -63.4704198 18.5883385, -63.4529368 18.5678798, -63.4295023 18.5626721, 
      -63.4042078 18.5708556, -63.3978842 18.5745754, -63.3893287 18.5823869, 
      -63.384865 18.5954061, -63.3911886 18.6069374, -63.3978842 18.6180967, 
      -63.3990002 18.6277681, -63.4042078 18.6329758, -63.4365698 18.6355796, 
      -63.467071999999995 18.6225604)), 
  POLYGON((-63 18.5, -63 19, -62 19,
    -62 18.5, -63 18.5)), 
  POLYGON((-62 18.5, -62 19, -61 19,
    -61 18.5, -62 18.5)), 
  POLYGON((-61 18.5, -61 19, -60 19,
    -60 18.5, -61 18.5)), 
  POLYGON((-60 18.5, -60 19, -59 19,
    -59 18.5, -60 18.5)), 
  POLYGON((-59 18.5, -59 19, -58 19,
    -58 18.5, -59 18.5)), 
  POLYGON((-58 18.5, -58 19, -57 19,
    -57 18.5, -58 18.5)), 
  POLYGON((-57 18.5, -57 19, -56 19,
    -56 18.5, -57 18.5)), 
  POLYGON((-56 18.5, -56 19, -55 19,
    -55 18.5, -56 18.5)), 
  POLYGON((-85.985971169818868 19.5, -85 19.5, -85 19,
    -85.989814006321723 19, -85.9896799 19.01502, -85.9806389 19.3133715,
    -85.985971169818868 19.5)), 
  POLYGON((-85 19, -85 19.5, -84 19.5,
    -84 19, -85 19)), 
  POLYGON((-84 19, -84 19.5, -83 19.5,
    -83 19, -84 19)), 
  POLYGON((-83 19, -83 19.5, -82 19.5,
    -82 19, -83 19)), 
  POLYGON((-82 19, -82 19.5, -81 19.5,
    -81 19, -82 19), 
    (-81.4812661 19.364563999999998, -81.4587152 19.3043274, -81.4587152 19.2868264, 
      -81.4499647 19.2649502, -81.4455894 19.2452615, -81.4368389 19.2386987, 
      -81.4062122 19.2146348, -81.3821483 19.2058843, -81.32527 19.2080719, 
      -81.290268 19.2146348, -81.2202639 19.2211976, -81.161198 19.2299482, 
      -81.1108826 19.2474492, -81.0780682 19.271513, -81.056192 19.2999522, 
      -81.0518167 19.3262037, -81.0693177 19.3743315, -81.0824435 19.3896449, 
      -81.1524475 19.3983954, -81.2027629 19.4027706, -81.290268 19.3983954, 
      -81.307769 19.3874572, -81.3187071 19.383081999999998, -81.3340205 19.383081999999998, 
      -81.3362081 19.3874572, -81.3383958 19.418084, -81.3843359 19.4268345, 
      -81.4193379 19.4268345, -81.4812661 19.364563999999998)), 
  POLYGON((-81 19, -81 19.5, -80 19.5,
    -80 19, -81 19)), 
  POLYGON((-80 19, -80 19.5, -79 19.5,
    -79 19, -80 19)), 
  POLYGON((-79 19, -79 19.5, -78 19.5,
    -78 19, -79 19)), 
  POLYGON((-78 19, -78 19.5, -77 19.5,
    -77 19, -78 19)), 
  POLYGON((-77 19, -77 19.5, -76 19.5,
    -76 19, -77 19)), 
  POLYGON((-76 19, -76 19.5, -75 19.5,
    -75 19, -76 19)), 
  POLYGON((-75 19, -75 19.5, -74 19.5,
    -74 19, -75 19)), 
  POLYGON((-73.6601257 19, -74 19, -74 19.5,
    -73.0945651685952 19.5, -73.0597179 19.3518993, -73.068815 19.1517634,
    -73.1143004 19.0789867, -73.6601257 19)), 
  POLYGON((-69.8315482882353 19.5, -69 19.5, -69 19.133038917005415,
    -69.0528049 19.0705797, -69.0581689 19.0741557, -69.0671089 19.0813077,
    -69.0742609 19.0866716, -69.1207489 19.0902476, -69.1600848 19.0795197,
    -69.1869048 19.0634277, -69.2173007 19.0616397, -69.2441207 19.0795197,
    -69.2870327 19.0902476, -69.3156406 19.0938236, -69.3388846 19.1009756,
    -69.3782205 19.1099156, -69.3889485 19.1170676, -69.3961005 19.1349476,
    -69.3227926 19.1367356, -69.2423327 19.1420996, -69.1976328 19.1707075,
    -69.1636608 19.1867995, -69.1422048 19.2243475, -69.1279009 19.2744114,
    -69.1225369 19.3119594, -69.1422048 19.3387793, -69.1636608 19.3316273,
    -69.1815408 19.3334153, -69.1815408 19.3512953, -69.1779648 19.3602353,
    -69.1779648 19.3638113, -69.1708128 19.3852673, -69.1726008 19.4102992,
    -69.2226647 19.3942073, -69.2745167 19.3727513, -69.3067006 19.3530833,
    -69.3317326 19.3441433, -69.3460366 19.3441433, -69.3639166 19.3620233,
    -69.4032525 19.3673873, -69.4998044 19.3673873, -69.5373523 19.3602353,
    -69.5713243 19.3512953, -69.5945683 19.3405673, -69.6356922 19.3334153,
    -69.6535722 19.3262633, -69.6911202 19.3262633, -69.7376081 19.3369913,
    -69.7662161 19.3566593, -69.787672 19.3799033, -69.80734 19.4085112,
    -69.830584 19.4603632, -69.830584 19.4836071, -69.8315482882353 19.5)), 
  POLYGON((-69 19.133038917005415, -69 19.5, -68 19.5,
    -68 19, -68.562893210988861 19, -68.5628935 19.0000004,
    -68.7309653 19.0562756, -68.8471852 19.0634276, -68.8990371 19.0866716,
    -68.9133411 19.102763500000002, -68.968769 19.1242195, -68.9999 19.1331572,
    -69 19.133038917005415)), 
  POLYGON((-68 19, -68 19.5, -67 19.5,
    -67 19, -68 19)), 
  POLYGON((-67 19, -67 19.5, -66 19.5,
    -66 19, -67 19)), 
  POLYGON((-66 19, -66 19.5, -65 19.5,
    -65 19, -66 19)), 
  POLYGON((-65 19, -65 19.5, -64 19.5,
    -64 19, -65 19)), 
  POLYGON((-64 19, -64 19.5, -63 19.5,
    -63 19, -64 19)), 
  POLYGON((-63 19, -63 19.5, -62 19.5,
    -62 19, -63 19)), 
  POLYGON((-62 19, -62 19.5, -61 19.5,
    -61 19, -62 19)), 
  POLYGON((-61 19, -61 19.5, -60 19.5,
    -60 19, -61 19)), 
  POLYGON((-60 19, -60 19.5, -59 19.5,
    -59 19, -60 19)), 
  POLYGON((-59 19, -59 19.5, -58 19.5,
    -58 19, -59 19)), 
  POLYGON((-58 19, -58 19.5, -57 19.5,
    -57 19, -58 19)), 
  POLYGON((-57 19, -57 19.5, -56 19.5,
    -56 19, -57 19)), 
  POLYGON((-56 19, -56 19.5, -55 19.5,
    -55 19, -56 19)), 
  POLYGON((-86 20, -85 20, -85 19.5,
    -85.985971169818868 19.5, -85.9896799 19.6298049, -86 20)), 
  POLYGON((-85 19.5, -85 20, -84 20,
    -84 19.5, -85 19.5)), 
  POLYGON((-84 19.5, -84 20, -83 20,
    -83 19.5, -84 19.5)), 
  POLYGON((-83 19.5, -83 20, -82 20,
    -82 19.5, -83 19.5)), 
  POLYGON((-82 19.5, -82 20, -81 20,
    -81 19.5, -82 19.5)), 
  POLYGON((-80 19.67053, -80 19.5, -81 19.5,
    -81 20, -80 20, -80 19.76268586801535,
    -80.0000001 19.7626859, -80.0316532 19.738329999999998, -80.0641279 19.7302113,
    -80.1047211 19.7058554, -80.1274534 19.6944892, -80.1371957 19.6766282,
    -80.1371957 19.653896, -80.1219658 19.6371431, -80.1209584 19.6360349,
    -80.0901075 19.6262926, -80.0560092 19.6279163, -80 19.67053)), 
  POLYGON((-80 19.76268586801535, -80 20, -79 20,
    -79 19.5, -80 19.5, -80 19.67053,
    -79.9553379 19.671757, -79.9391006 19.6733807, -79.9228633 19.6701333,
    -79.8871412 19.6490248, -79.857914 19.6506485, -79.8124496 19.6571434,
    -79.7815987 19.6701333, -79.7458766 19.696113, -79.7004121 19.7269639,
    -79.6793036 19.7399537, -79.6793036 19.7691809, -79.7231444 19.7789233,
    -79.7799749 19.7756758, -79.8335581 19.7691809, -79.8660327 19.752943600000002,
    -79.9033785 19.7496961, -79.9455955 19.7496961, -79.9796938 19.756190999999998,
    -80 19.76268586801535)), 
  POLYGON((-78 19.7827176, -78 19.5, -79 19.5,
    -79 20, -78.5492991 20, -78.4992405 19.9725479,
    -78.2482962 19.8671513, -78 19.7827176)), 
  POLYGON((-77 19.6797943, -77 19.5, -78 19.5,
    -78 19.7827176, -77.8605071 19.7481701, -77.5886824 19.6950421,
    -77 19.6797943)), 
  POLYGON((-76 19.740829617230016, -76 19.5, -77 19.5,
    -77 19.6797943, -76.0006646 19.7407444, -76 19.740829617230016)), 
  POLYGON((-75 19.731815343728769, -75 19.5, -76 19.5,
    -76 19.740829617230016, -75.8816769 19.7560014, -75.6104722 19.7944885,
    -75 19.731815343728769)), 
  POLYGON((-74.131338432491944 20, -74 20, -74 19.5,
    -75 19.5, -75 19.731815343728769, -74.9999986 19.7318152,
    -74.7727696 19.7675678, -74.3712403 19.8225718, -74.1842267 19.9600819,
    -74.131338432491944 20)), 
  POLYGON((-73.0945651685952 19.5, -74 19.5, -74 20,
    -73.4508927 20, -73.5964461 19.7976566, -73.6510286 19.6702974,
    -73.4690869 19.5156469, -73.2325626 19.524744, -73.0961063 19.5065498,
    -73.0945651685952 19.5)), 
  POLYGON((-72.394394683982355 20, -72 20, -72 19.897982893204329,
    -72.0920497 19.891727500000002, -72.2338388 19.9292599, -72.3568617 19.9667923,
    -72.3943941 19.999999, -72.394394683982355 20)), 
  POLYGON((-72 19.897982893204329, -72 20, -71 20,
    -71 19.960536877955267, -71.0303305 19.9538506, -71.1537024 19.9574266,
    -71.3164102 19.9484866, -71.454086 19.9520626, -71.5989138 19.9520626,
    -71.8045336 19.9163027, -71.9999999 19.8979829, -72 19.897982893204329)), 
  POLYGON((-71 19.960536877955267, -71 20, -70 20,
    -70 19.739290852886082, -70.1935476 19.7088949, -70.2650675 19.7321389,
    -70.3669834 19.8233268, -70.4295633 19.8233268, -70.4796272 19.8197508,
    -70.5439951 19.8215388, -70.628031 19.8322668, -70.7245829 19.8644507,
    -70.7692829 19.9073627, -70.8247108 19.9270306, -70.9266267 19.9413346,
    -70.9605986 19.9431226, -70.9999999 19.9605369, -71 19.960536877955267)), 
  POLYGON((-70 19.739290852886082, -70 20, -69 20,
    -69 19.5, -69.8315482882353 19.5, -69.832371999999992 19.5140031,
    -69.8431 19.5461871, -69.848463999999993 19.635586999999997, -69.8734959 19.6641949,
    -69.9003159 19.6892269, -69.9003159 19.7017429, -69.9378638 19.7249868,
    -69.9999997 19.7392909, -70 19.739290852886082)), 
  POLYGON((-69 19.5, -69 20, -68 20,
    -68 19.5, -69 19.5)), 
  POLYGON((-68 19.5, -68 20, -67 20,
    -67 19.5, -68 19.5)), 
  POLYGON((-67 19.5, -67 20, -66 20,
    -66 19.5, -67 19.5)), 
  POLYGON((-66 19.5, -66 20, -65 20,
    -65 19.5, -66 19.5)), 
  POLYGON((-65 19.5, -65 20, -64 20,
    -64 19.5, -65 19.5)), 
  POLYGON((-64 19.5, -64 20, -63 20,
    -63 19.5, -64 19.5)), 
  POLYGON((-63 19.5, -63 20, -62 20,
    -62 19.5, -63 19.5)), 
  POLYGON((-62 19.5, -62 20, -61 20,
    -61 19.5, -62 19.5)), 
  POLYGON((-61 19.5, -61 20, -60 20,
    -60 19.5, -61 19.5)), 
  POLYGON((-60 19.5, -60 20, -59 20,
    -59 19.5, -60 19.5)), 
  POLYGON((-59 19.5, -59 20, -58 20,
    -58 19.5, -59 19.5)), 
  POLYGON((-58 19.5, -58 20, -57 20,
    -57 19.5, -58 19.5)), 
  POLYGON((-57 19.5, -57 20, -56 20,
    -56 19.5, -57 19.5)), 
  POLYGON((-56 19.5, -56 20, -55 20,
    -55 19.5, -56 19.5)), 
  LINESTRING(-86 20, -86 20.5), 
  POLYGON((-86 20.5, -85 20.5, -85 20,
    -86 20, -86 20.5)), 
  POLYGON((-85 20, -85 20.5, -84 20.5,
    -84 20, -85 20)), 
  POLYGON((-84 20, -84 20.5, -83 20.5,
    -83 20, -84 20)), 
  POLYGON((-83 20, -83 20.5, -82 20.5,
    -82 20, -83 20)), 
  POLYGON((-82 20, -82 20.5, -81 20.5,
    -81 20, -82 20)), 
  POLYGON((-81 20, -81 20.5, -80 20.5,
    -80 20, -81 20)), 
  POLYGON((-79 20.4768041, -79 20, -80 20,
    -80 20.5, -79.019304700839186 20.5, -79 20.4768041)), 
  POLYGON((-78.5492991 20, -79 20, -79 20.4768041,
    -78.5492991 20)), 
  POLYGON((-74.2301749960743 20.5, -74 20.5, -74 20,
    -74.131338432491944 20, -74.1313383 20.0000001, -74.1017207 20.0673397,
    -74.0384661 20.1498457, -74.0329657 20.2213509, -74.0577175 20.3121075,
    -74.1402235 20.3918633, -74.2224607 20.4872738, -74.2301749960743 20.5)), 
  POLYGON((-73.4508927 20, -74 20, -74 20.5,
    -73 20.5, -73 20.3070936, -73.3508247 20.143345999999998,
    -73.4508927 20)), 
  POLYGON((-73 20.3070936, -73 20.5, -72 20.5,
    -72 20, -72.394394683982355 20, -72.4300711 20.0610916,
    -72.5507791 20.133603, -72.7217601 20.2441151, -72.8698046 20.2753921,
    -73 20.3070936)), 
  POLYGON((-72 20, -72 20.5, -71 20.5,
    -71 20, -72 20)), 
  POLYGON((-71 20, -71 20.5, -70 20.5,
    -70 20, -71 20)), 
  POLYGON((-70 20, -70 20.5, -69 20.5,
    -69 20, -70 20)), 
  POLYGON((-69 20, -69 20.5, -68 20.5,
    -68 20, -69 20)), 
  POLYGON((-68 20, -68 20.5, -67 20.5,
    -67 20, -68 20)), 
  POLYGON((-67 20, -67 20.5, -66 20.5,
    -66 20, -67 20)), 
  POLYGON((-66 20, -66 20.5, -65 20.5,
    -65 20, -66 20)), 
  POLYGON((-65 20, -65 20.5, -64 20.5,
    -64 20, -65 20)), 
  POLYGON((-64 20, -64 20.5, -63 20.5,
    -63 20, -64 20)), 
  POLYGON((-63 20, -63 20.5, -62 20.5,
    -62 20, -63 20)), 
  POLYGON((-62 20, -62 20.5, -61 20.5,
    -61 20, -62 20)), 
  POLYGON((-61 20, -61 20.5, -60 20.5,
    -60 20, -61 20)), 
  POLYGON((-60 20, -60 20.5, -59 20.5,
    -59 20, -60 20)), 
  POLYGON((-59 20, -59 20.5, -58 20.5,
    -58 20, -59 20)), 
  POLYGON((-58 20, -58 20.5, -57 20.5,
    -57 20, -58 20)), 
  POLYGON((-57 20, -57 20.5, -56 20.5,
    -56 20, -57 20)), 
  POLYGON((-56 20, -56 20.5, -55 20.5,
    -55 20, -56 20)), 
  LINESTRING(-86 20.5, -86 21), 
  POLYGON((-86 21, -85 21, -85 20.5,
    -86 20.5, -86 21)), 
  POLYGON((-85 20.5, -85 21, -84 21,
    -84 20.5, -85 20.5)), 
  POLYGON((-84 20.5, -84 21, -83 21,
    -83 20.5, -84 20.5)), 
  POLYGON((-83 20.5, -83 21, -82 21,
    -82 20.5, -83 20.5)), 
  POLYGON((-82 20.5, -82 21, -81 21,
    -81 20.5, -82 20.5)), 
  POLYGON((-81 20.5, -81 21, -80 21,
    -80 20.5, -81 20.5)), 
  POLYGON((-79.019304700839186 20.5, -80 20.5, -80 21,
    -79.5258843218387 21, -79.5258841 20.9999985, -79.3977441 20.8961708,
    -79.2731633 20.7623322, -79.2035929 20.7214347, -79.019304700839186 20.5)), 
  POLYGON((-75.375061400720909 21, -75 21, -75 20.8570704,
    -75.375061400720909 21)), 
  POLYGON((-75 20.8570704, -75 21, -74 21,
    -74 20.5, -74.2301749960743 20.5, -74.2529817 20.537623999999997,
    -74.3574893 20.592627999999998, -74.4436807 20.6686368, -74.5170009 20.7081364,
    -74.7397672 20.798893, -74.8992788 20.8208946, -75 20.8570704)), 
  POLYGON((-73.104241313432865 21, -73 21, -73 20.5,
    -74 20.5, -74 21, -73.7659522 21,
    -73.757403 20.9796245, -73.7471439 20.9693655, -73.7146568 20.9420079,
    -73.7043978 20.9146504, -73.668491 20.8907126, -73.6120661 20.9061012,
    -73.601807 20.9249095, -73.5607707 20.9249095, -73.5317033 20.9061012,
    -73.4633095 20.8941323, -73.3778172 20.9009716, -73.3607188 20.9180701,
    -73.3231022 20.9129406, -73.2906151 20.9214898, -73.2307705 20.930039,
    -73.1640865 20.9334587, -73.1315994 20.9591064, -73.1042419 20.9999991,
    -73.104241313432865 21)), 
  POLYGON((-73 20.5, -73 21, -72 21,
    -72 20.5, -73 20.5)), 
  POLYGON((-72 20.5, -72 21, -71 21,
    -71 20.5, -72 20.5)), 
  POLYGON((-71 20.5, -71 21, -70 21,
    -70 20.5, -71 20.5)), 
  POLYGON((-70 20.5, -70 21, -69 21,
    -69 20.5, -70 20.5)), 
  POLYGON((-69 20.5, -69 21, -68 21,
    -68 20.5, -69 20.5)), 
  POLYGON((-68 20.5, -68 21, -67 21,
    -67 20.5, -68 20.5)), 
  POLYGON((-67 20.5, -67 21, -66 21,
    -66 20.5, -67 20.5)), 
  POLYGON((-66 20.5, -66 21, -65 21,
    -65 20.5, -66 20.5)), 
  POLYGON((-65 20.5, -65 21, -64 21,
    -64 20.5, -65 20.5)), 
  POLYGON((-64 20.5, -64 21, -63 21,
    -63 20.5, -64 20.5)), 
  POLYGON((-63 20.5, -63 21, -62 21,
    -62 20.5, -63 20.5)), 
  POLYGON((-62 20.5, -62 21, -61 21,
    -61 20.5, -62 20.5)), 
  POLYGON((-61 20.5, -61 21, -60 21,
    -60 20.5, -61 20.5)), 
  POLYGON((-60 20.5, -60 21, -59 21,
    -59 20.5, -60 20.5)), 
  POLYGON((-59 20.5, -59 21, -58 21,
    -58 20.5, -59 20.5)), 
  POLYGON((-58 20.5, -58 21, -57 21,
    -57 20.5, -58 20.5)), 
  POLYGON((-57 20.5, -57 21, -56 21,
    -56 20.5, -57 20.5)), 
  POLYGON((-56 20.5, -56 21, -55 21,
    -55 20.5, -56 20.5)), 
  LINESTRING(-86 21, -86 21.5), 
  POLYGON((-86 21.5, -85 21.5, -85 21,
    -86 21, -86 21.5)), 
  POLYGON((-85 21, -85 21.5, -84 21.5,
    -84 21, -85 21)), 
  POLYGON((-83 21.3775852, -83 21, -84 21,
    -84 21.5, -83.232272051496707 21.5, -83.2195528 21.4756675,
    -83.1901281 21.4672604, -83.1326799 21.4294287, -83 21.3775852)), 
  POLYGON((-82 21.4956653, -82 21, -83 21,
    -83 21.3775852, -82.9155839 21.3821633, -82.7750575 21.4037827,
    -82.6561506 21.4362119, -82.5210291 21.4902605, -82.4183367 21.4902605,
    -82.277113724750933 21.5, -82.246613099002744 21.5, -82.1859277 21.4956653,
    -82 21.4956653)), 
  POLYGON((-81.642123180199818 21.5, -81 21.5, -81 21,
    -82 21, -82 21.4956653, -81.9999994 21.4956653,
    -81.932379475470412 21.5, -81.708588455444485 21.5, -81.672466 21.4956653,
    -81.642123180199818 21.5)), 
  POLYGON((-81 21, -81 21.5, -80 21.5,
    -80 21, -81 21)), 
  POLYGON((-79.5258843218387 21, -80 21, -80 21.5,
    -79.660955355940274 21.5, -79.6363347 21.476305, -79.5849206 21.3991838,
    -79.5258843218387 21)), 
  POLYGON((-76.2730824275 21.5, -76 21.5, -76 21.384685632813586,
    -76.0000001 21.3846857, -76.2221252 21.4754423, -76.2730824275 21.5)), 
  POLYGON((-76 21.384685632813586, -76 21.5, -75 21.5,
    -75 21, -75.375061400720909 21, -75.3750635 21.0000008,
    -75.5868289 21.115166, -75.7160883 21.2141733, -75.8123453 21.2609267,
    -75.9058522 21.3214311, -76 21.384685632813586)), 
  POLYGON((-75 21, -75 21.5, -74 21.5,
    -74 21, -75 21)), 
  GEOMETRYCOLLECTION(
  POLYGON((-73.7659522 21, -74 21, -74 21.5,
    -73.114662427887922 21.5, -73.1110813 21.461800999999998, -73.1076616 21.4276041,
    -73.0991124 21.384857999999998, -73.0785942 21.3609201, -73.1042419 21.3386921,
    -73.1230502 21.3113346, -73.1486979 21.2805574, -73.1709259 21.2600392,
    -73.2017031 21.2343916, -73.2290606 21.2207128, -73.2564182 21.2155833,
    -73.2666773 21.2138734, -73.2906151 21.2172931, -73.2974545 21.2207128,
    -73.3231022 21.2378112, -73.324812 21.2429408, -73.3419105 21.2617491,
    -73.3897861 21.2788475, -73.4103043 21.2771377, -73.4239831 21.2702983,
    -73.4274028 21.2566196, -73.435952 21.2531999, -73.4445012 21.2463605,
    -73.4479209 21.2395211, -73.45818 21.2343916, -73.4615997 21.2343916,
    -73.4992163 21.2326817, -73.5180246 21.2395211, -73.5385427 21.2480703,
    -73.5436723 21.2429408, -73.5624806 21.2378112, -73.569319999999991 21.2155833,
    -73.5847086 21.1899356, -73.6069366 21.1779666, -73.6171956 21.1728371,
    -73.6394236 21.1711273, -73.6719107 21.1694174, -73.7112371 21.1642879,
    -73.7232061 21.1574485, -73.7232061 21.1488993, -73.7334651 21.1386402,
    -73.7266258 21.118122, -73.7163667 21.1027334, -73.7078175 21.0873448,
    -73.7112371 21.0719562, -73.7180765 21.063406999999998, -73.7266258 21.0582774,
    -73.7300455 21.0582774, -73.7437242 21.0514381, -73.7642424 21.0394691,
    -73.7710817 21.0172411, -73.7659522 21)), 
    POLYGON((-73 21.1813863, -73 21, -73.104241313432865 21,
      -73.0751745 21.0445987, -73.0649155 21.0702464, -73 21.1813863))), 
  POLYGON((-72.887169146969413 21.5, -72 21.5, -72 21,
    -73 21, -73 21.1813863, -72.981133 21.2412309,
    -72.9828429 21.3010755, -72.9845527 21.3506611, -72.9777133 21.384857999999998,
    -72.9606149 21.4036663, -72.9058998 21.4327337, -72.8836718 21.47377,
    -72.887169146969413 21.5)), 
  POLYGON((-71.09651744245231 21.5, -71 21.5, -71 21,
    -72 21, -72 21.5, -71.552919311652218 21.5,
    -71.5525781 21.4989764, -71.5519274 21.4924693, -71.5610374 21.4794551,
    -71.5655923 21.4768523, -71.5701473 21.4735987, -71.5818601 21.4657902,
    -71.5890179 21.4599339, -71.5994292 21.4592831, -71.6046349 21.4605846,
    -71.6163476 21.4683931, -71.622204 21.4716466, -71.631314 21.472948,
    -71.6521367 21.4703452, -71.6586437 21.4592831, -71.6573423 21.4540775,
    -71.6560409 21.446268999999997, -71.6449788 21.4391112, -71.6293618 21.4345562,
    -71.6150462 21.4339055, -71.6072377 21.4339055, -71.5981278 21.4430154,
    -71.5903193 21.4430154, -71.5681952 21.4384604, -71.5493246 21.4410633,
    -71.5402147 21.4495225, -71.5311047 21.4586324, -71.5265498 21.4644888,
    -71.5252484 21.4742495, -71.4959664 21.4735987, -71.4842537 21.4762016,
    -71.4816508 21.4853115, -71.479047999999992 21.493119999999998, -71.477041314977257 21.5,
    -71.188760619871474 21.5, -71.2017203 21.486648, -71.1851236 21.4517759,
    -71.1675696 21.4239406, -71.1622889 21.4133793, -71.1504074 21.3962172,
    -71.1543679 21.3764147, -71.1543679 21.3750946, -71.1820913 21.3671736,
    -71.2071744 21.361893, -71.2190559 21.3486913, -71.2243366 21.3434107,
    -71.2309374 21.3196477, -71.2335777 21.2998453, -71.236218 21.2760823,
    -71.236218 21.2602404, -71.2454592 21.2351573, -71.2626213 21.216675,
    -71.2692221 21.200833, -71.2705423 21.1863112, -71.259981 21.1691491,
    -71.2375382 21.1704693, -71.2296172 21.1823508, -71.2071744 21.2232758,
    -71.2071744 21.2562799, -71.1939728 21.2906041, -71.1913325 21.3011654,
    -71.179451 21.3156872, -71.1636091 21.3249284, -71.1504074 21.3341695,
    -71.1438066 21.3420905, -71.122683999999992 21.3447308, -71.114763 21.3341695,
    -71.0976009 21.3143671, -71.0791186 21.3170074, -71.0632767 21.3394502,
    -71.0764783 21.3645333, -71.0843993 21.3816954, -71.0843993 21.394897,
    -71.073838 21.4041382, -71.0777985 21.41866, -71.1015614 21.4358221,
    -71.106842 21.4437431, -71.1121227 21.4543044, -71.0976009 21.4701463,
    -71.0949606 21.4859883, -71.09651744245231 21.5), 
    (-71.6806244 21.1070519, -71.6310063 21.1292028, -71.5911345 21.1876814, 
      -71.6141715 21.2151485, -71.6531572 21.2204648, -71.6823965 21.2160346, 
      -71.6956871 21.2240089, -71.7506214 21.1974278, -71.7541656 21.155784, 
      -71.7382169 21.1389492, -71.6806244 21.1070519), 
    (-71.7656841 21.2612225, -71.6974591 21.2514761, -71.6797384 21.2603365, 
      -71.6682199 21.2567923, -71.6540433 21.2647667, -71.648727 21.2656527, 
      -71.6274621 21.2647667, -71.6053111 21.2700829, -71.5902485 21.2860316, 
      -71.5953025 21.297204999999998, -71.5967236 21.321089100000002, -71.6137848 21.3407704, 
      -71.6451829 21.3515984, -71.6673338 21.3489403, -71.6832825 21.3436241, 
      -71.6992312 21.3445101, -71.7045474 21.3400799, -71.740875 21.3303335, 
      -71.7532795 21.318815, -71.787835 21.2807154, -71.7656841 21.2612225)), 
  POLYGON((-71 21, -71 21.5, -70 21.5,
    -70 21, -71 21)), 
  POLYGON((-70 21, -70 21.5, -69 21.5,
    -69 21, -70 21)), 
  POLYGON((-69 21, -69 21.5, -68 21.5,
    -68 21, -69 21)), 
  POLYGON((-68 21, -68 21.5, -67 21.5,
    -67 21, -68 21)), 
  POLYGON((-67 21, -67 21.5, -66 21.5,
    -66 21, -67 21)), 
  POLYGON((-66 21, -66 21.5, -65 21.5,
    -65 21, -66 21)), 
  POLYGON((-65 21, -65 21.5, -64 21.5,
    -64 21, -65 21)), 
  POLYGON((-64 21, -64 21.5, -63 21.5,
    -63 21, -64 21)), 
  POLYGON((-63 21, -63 21.5, -62 21.5,
    -62 21, -63 21)), 
  POLYGON((-62 21, -62 21.5, -61 21.5,
    -61 21, -62 21)), 
  POLYGON((-61 21, -61 21.5, -60 21.5,
    -60 21, -61 21)), 
  POLYGON((-60 21, -60 21.5, -59 21.5,
    -59 21, -60 21)), 
  POLYGON((-59 21, -59 21.5, -58 21.5,
    -58 21, -59 21)), 
  POLYGON((-58 21, -58 21.5, -57 21.5,
    -57 21, -58 21)), 
  POLYGON((-57 21, -57 21.5, -56 21.5,
    -56 21, -57 21)), 
  POLYGON((-56 21, -56 21.5, -55 21.5,
    -55 21, -56 21)), 
  LINESTRING(-86 21.5, -86 22), 
  POLYGON((-86 22, -85 22, -85 21.5,
    -86 21.5, -86 22)), 
  POLYGON((-84 21.844532445285431, -84 21.5, -85 21.5,
    -85 22, -84.897293460798949 22, -84.9502743 21.9670515,
    -84.9767649 21.9372496, -84.9833875 21.9008251, -84.9900101 21.8412214,
    -84.9833875 21.8081082, -84.946963 21.7683724, -84.8608687 21.7385705,
    -84.7416613 21.7485045, -84.6588783 21.8246648, -84.5959633 21.8114195,
    -84.5794067 21.7485045, -84.5429822 21.7153913, -84.4601992 21.7054574,
    -84.3939728 21.7385705, -84.3111899 21.8047969, -84.2151616 21.8312875,
    -84.1091995 21.8246648, -84 21.844532445285431)), 
  POLYGON((-83.232272051496707 21.5, -84 21.5, -84 21.844532445285431,
    -83.9999986 21.8445327, -83.880731825129018 22, -83.706027334522048 22,
    -83.658152 21.9773562, -83.6409477 21.9752709, -83.638341 21.9565026,
    -83.6039325 21.9424263, -83.5580544 21.933563499999998, -83.5017494 21.925221999999998,
    -83.4673409 21.9309568, -83.4412738 21.9273074, -83.4058225 21.9267861,
    -83.36985 21.9267861, -83.331792 21.9273074, -83.3060662 21.9315921,
    -83.2994688 21.9512891, -83.3005115 21.9565026, -83.3062462 21.9638013,
    -83.311981 21.9700574, -83.3229292 21.9742282, -83.3370054 21.9773562,
    -83.3505603 21.9773562, -83.3578591 21.9768349, -83.4021731 21.9757922,
    -83.415728 21.9768349, -83.4313683 21.9789203, -83.4475299 21.9789203,
    -83.4662982 21.983090999999998, -83.4725543 21.9841337, -83.4923653 21.9883044,
    -83.5126976 21.9883044, -83.5231245 21.9924751, -83.5298736473797 22,
    -83.153409643632642 22, -83.2786233 21.9012924, -83.2812045 21.887612999999998,
    -83.293815 21.8722001, -83.2896115 21.8525837, -83.3008209 21.831566,
    -83.2966174 21.799339, -83.2896115 21.7713155, -83.2826056 21.7404896,
    -83.2966174 21.7236755, -83.2826056 21.6732332, -83.2755998 21.6171862,
    -83.2629892 21.5583368, -83.2517798 21.5373192, -83.232272051496707 21.5)), 
  POLYGON((-82.246613099002744 21.5, -82.277113724750933 21.5, -82.2615958 21.5010702,
    -82.246613099002744 21.5)), 
  GEOMETRYCOLLECTION(
  POLYGON((-81.708588455444485 21.5, -81.932379475470412 21.5, -81.9156847 21.5010702,
    -81.8075875 21.5118799, -81.708588455444485 21.5)), 
    POLYGON((-81 21.9287494, -81 21.5, -81.642123180199818 21.5,
      -81.5211299 21.5172848, -81.3213315 21.539437200000002, -81.245482 21.5334994,
      -81.1427896 21.5443091, -81.0779313 21.6470014, -81.0779313 21.7226695,
      -81.1103605 21.7767181, -81.1644091 21.8523861, -81.2725063 21.8361715,
      -81.4130327 21.8307667, -81.5589639 21.8686007, -81.7103 21.9172445,
      -81.7230068 21.9254131, -81.8508264 22, -81.75459699999999 22,
      -81.6234909 21.9647393, -81.500097 21.9621686, -81.3278597 21.9621686,
      -81.0913547 21.9544565, -81.0347991 21.9441737, -81 21.9287494))), 
  POLYGON((-80 21.6845322, -80 21.5, -81 21.5,
    -81 21.9287494, -80.7545921 21.9338908, -80.6442673 21.9245832,
    -80.5283699 21.9261787, -80.4435366 21.9133252, -80.3535619 21.8747646,
    -80.1890366 21.7616535, -80.1042033 21.7128101, -80 21.6845322)), 
  POLYGON((-79.660955355940274 21.5, -80 21.5, -80 21.6845322,
    -79.8815488 21.630855, -79.7226135 21.55934, -79.660955355940274 21.5)), 
  POLYGON((-77.151692579017492 22, -77 22, -77 21.810966102454248,
    -77.0000017 21.8109668, -77.151692579017492 22)), 
  POLYGON((-77 21.810966102454248, -77 22, -76 22,
    -76 21.5, -76.2730824275 21.5, -76.4503918 21.5854503,
    -77 21.810966102454248)), 
  POLYGON((-76 21.5, -76 22, -75 22,
    -75 21.5, -76 21.5)), 
  POLYGON((-75 21.5, -75 22, -74 22,
    -74 21.5, -75 21.5)), 
  POLYGON((-73.114662427887922 21.5, -74 21.5, -74 22,
    -73 22, -73 21.6233098, -73.0324284 21.595169,
    -73.0646549 21.5829752, -73.0820139 21.5472933, -73.0956927 21.5336146,
    -73.1162108 21.5165161, -73.114662427887922 21.5), 
    (-73.8301051 21.736715, -73.8689033 21.7402762, -73.9099508 21.712911300000002, 
      -73.9112628 21.6960424, -73.9112628 21.6838594, -73.903953 21.6639917, 
      -73.893081999999993 21.652371, -73.8685285 21.6433743, -73.8372275 21.6448737, 
      -73.7978669 21.6570568, -73.7894325 21.6937933, -73.7918691 21.7117867, 
      -73.8033024 21.7235948, -73.8301051 21.736715)), 
  POLYGON((-72 21.855852568452583, -72 21.5, -72.887169146969413 21.5,
    -72.8905112 21.5250653, -72.9024801 21.5558426, -72.9571952 21.5763607,
    -72.9640346 21.5849099, -72.9760035 21.6139773, -73 21.6233098,
    -73 22, -72 22, -72 21.979612879470633,
    -72.0000002 21.9796129, -72.0586848 21.9812413, -72.0765975 21.9633286,
    -72.1010238 21.9405307, -72.1221934 21.8981916, -72.1335923 21.880278999999998,
    -72.1498766 21.8672516, -72.1677893 21.8509673, -72.1791882 21.8346831,
    -72.1987293 21.8249125, -72.2296694 21.8297978, -72.2540958 21.8591094,
    -72.281779 21.8705084, -72.3029485 21.8900495, -72.3290033 21.9014485,
    -72.3322601 21.9047053, -72.3404023 21.9014485, -72.3452875 21.8981916,
    -72.3485444 21.8851642, -72.3697139 21.8363115, -72.3745992 21.7972293,
    -72.3794844 21.7793166, -72.3811129 21.7581471, -72.3843697 21.7483766,
    -72.3908834 21.7483766, -72.4267088 21.7532619, -72.4446214 21.750004999999998,
    -72.4609057 21.7418629, -72.4804468 21.7141797, -72.4869605 21.6978954,
    -72.4918457 21.6799827, -72.496731 21.6669554, -72.5097584 21.6474143,
    -72.5179005 21.63113, -72.5146437 21.6083321, -72.5016163 21.5822773,
    -72.4755615 21.5790204, -72.4446214 21.5855341, -72.3937288 21.6361914,
    -72.3876266 21.6734691, -72.3501728 21.7288355, -72.2944743 21.7031305,
    -72.258980999999991 21.7174365, -72.2426968 21.7304639, -72.2036146 21.7255786,
    -72.1623002 21.7392733, -72.1319639 21.7728029, -72.1075375 21.7972293,
    -72.0619417 21.8493389, -72 21.855852568452583)), 
  GEOMETRYCOLLECTION(
  POLYGON((-71.552919311652218 21.5, -72 21.5, -72 21.855852568452583,
    -71.9999997 21.8558526, -71.9805202 21.8493387, -71.9479517 21.8297977,
    -71.9202685 21.811885, -71.887700000000009 21.7972292, -71.8649021 21.7988576,
    -71.8583884 21.7842018, -71.8469894 21.7711744, -71.845361 21.7646607,
    -71.8551315 21.7451196, -71.8665305 21.7272069, -71.8665305 21.7109227,
    -71.8600168 21.6897531, -71.8372189 21.6832394, -71.8111641 21.7027805,
    -71.8013935 21.7174364, -71.7785956 21.7402343, -71.7623114 21.7516333,
    -71.737884999999991 21.758146999999997, -71.7281144 21.746748, -71.7199723 21.7304638,
    -71.7069449 21.7223216, -71.6955459 21.7223216, -71.6890322 21.7288353,
    -71.6760049 21.7239501, -71.6564638 21.7158079, -71.6434364 21.7158079,
    -71.61901 21.7141795, -71.6222668 21.7027805, -71.6369227 21.6946384,
    -71.6580922 21.6848679, -71.672748 21.6750973, -71.6760049 21.6636984,
    -71.6760049 21.6539278, -71.6515785 21.6262446, -71.628708 21.6465835,
    -71.6027257 21.6750973, -71.596212 21.6913816, -71.5717857 21.6913816,
    -71.565272 21.6832394, -71.5408456 21.6783542, -71.5327035 21.6685836,
    -71.5489877 21.6604415, -71.5636436 21.6539278, -71.5913268 21.639272,
    -71.5978405 21.6246162, -71.6027257 21.6115888, -71.5655923 21.5985349,
    -71.5480232 21.5933292, -71.5214335 21.5707367, -71.5154877 21.5588417,
    -71.5154877 21.5503824, -71.5167891 21.5406218, -71.5245976 21.5373683,
    -71.5337076 21.5412725, -71.5434682 21.5406218, -71.5558317 21.5302105,
    -71.5571331 21.5191484, -71.5571331 21.5106892, -71.5545303 21.5048328,
    -71.552919311652218 21.5)), 
    POLYGON((-71.188760619871474 21.5, -71.477041314977257 21.5, -71.474493 21.508737,
      -71.4718902 21.5256555, -71.4718902 21.5360668, -71.4660219 21.5561556,
      -71.4582252 21.5978842, -71.4504167 21.6193576, -71.4432589 21.6421324,
      -71.4426082 21.6629551, -71.4452111 21.6740172, -71.4478139 21.6824764,
      -71.4478139 21.6935384, -71.4465125 21.7137104, -71.4465125 21.728026,
      -71.4504167 21.7345331, -71.4569238 21.7410402, -71.4621295 21.7423416,
      -71.4705887 21.7423416, -71.4810001 21.7390881, -71.4914114 21.7312796,
      -71.5122341 21.737136, -71.5382625 21.7455952, -71.5564823 21.7677193,
      -71.5857643 21.7774799, -71.5961756 21.7703221, -71.6000799 21.7781306,
      -71.6163476 21.7911448, -71.6196011 21.7924462, -71.6293618 21.7878913,
      -71.6339167 21.7930969, -71.637821 21.8100154, -71.6475816 21.8295367,
      -71.6586437 21.8471058, -71.67361 21.8536129, -71.6807678 21.8516607,
      -71.693782 21.8542636, -71.709398999999991 21.8627228, -71.7243653 21.86012,
      -71.7341259 21.8516607, -71.7412837 21.8516607, -71.7673121 21.8523114,
      -71.7848813 21.8536129, -71.7972447 21.8536129, -71.8076561 21.8549143,
      -71.8148139 21.8627228, -71.816766 21.8666271, -71.821321 21.871181999999997,
      -71.8239238 21.879641200000002, -71.8278281 21.8854976, -71.832383 21.8933061,
      -71.8427944 21.8985118, -71.8532057 21.8998132, -71.857109999999992 21.8946075,
      -71.8603635 21.8939568, -71.861665 21.895258300000002, -71.8629664 21.8991625,
      -71.861665 21.9089232, -71.8610142 21.9180331, -71.861665 21.9290951,
      -71.8629664 21.9323487, -71.8720763 21.9401572, -71.8792341 21.9447122,
      -71.8889947 21.9531714, -71.8955018 21.9577263, -71.9104681 21.9642334,
      -71.9176259 21.9668363, -71.9505335 21.9745353, -72 21.979612879470633,
      -72 22, -71 22, -71 21.5,
      -71.09651744245231 21.5, -71.0976009 21.5097512, -71.0923203 21.5348343,
      -71.1187235 21.5506763, -71.1556881 21.5361545, -71.1741704 21.5150319,
      -71.188760619871474 21.5))), 
  POLYGON((-71 21.5, -71 22, -70 22,
    -70 21.5, -71 21.5)), 
  POLYGON((-70 21.5, -70 22, -69 22,
    -69 21.5, -70 21.5)), 
  POLYGON((-69 21.5, -69 22, -68 22,
    -68 21.5, -69 21.5)), 
  POLYGON((-68 21.5, -68 22, -67 22,
    -67 21.5, -68 21.5)), 
  POLYGON((-67 21.5, -67 22, -66 22,
    -66 21.5, -67 21.5)), 
  POLYGON((-66 21.5, -66 22, -65 22,
    -65 21.5, -66 21.5)), 
  POLYGON((-65 21.5, -65 22, -64 22,
    -64 21.5, -65 21.5)), 
  POLYGON((-64 21.5, -64 22, -63 22,
    -63 21.5, -64 21.5)), 
  POLYGON((-63 21.5, -63 22, -62 22,
    -62 21.5, -63 21.5)), 
  POLYGON((-62 21.5, -62 22, -61 22,
    -61 21.5, -62 21.5)), 
  POLYGON((-61 21.5, -61 22, -60 22,
    -60 21.5, -61 21.5)), 
  POLYGON((-60 21.5, -60 22, -59 22,
    -59 21.5, -60 21.5)), 
  POLYGON((-59 21.5, -59 22, -58 22,
    -58 21.5, -59 21.5)), 
  POLYGON((-58 21.5, -58 22, -57 22,
    -57 21.5, -58 21.5)), 
  POLYGON((-57 21.5, -57 22, -56 22,
    -56 21.5, -57 21.5)), 
  POLYGON((-56 21.5, -56 22, -55 22,
    -55 21.5, -56 21.5)), 
  LINESTRING(-86 22, -86 22.5), 
  POLYGON((-86 22.5, -85 22.5, -85 22,
    -86 22, -86 22.5)), 
  POLYGON((-84.897293460798949 22, -85 22, -85 22.5,
    -84.480686463980788 22.5, -84.519803 22.4471927, -84.5396709 22.3644097,
    -84.5496049 22.2584476, -84.5462936 22.1624193, -84.5529162 22.1160609,
    -84.5959633 22.0995043, -84.6721237 22.0928816, -84.7714632 22.0763251,
    -84.8972933 22.0000001, -84.897293460798949 22)), 
  POLYGON((-83.706027334522048 22, -83.880731825129018 22, -83.8807184 22.0000175,
    -83.887340999999992 22.0697023, -83.8343599 22.1160608, -83.6952846 22.1127494,
    -83.5793884 22.1259947, -83.4767376 22.139239999999997, -83.3442848 22.1690418,
    -83.2052095 22.2418909, -83.1191152 22.3412304, -83.0330209 22.420702,
    -83 22.437513045861383, -83 22.139901028852119, -83.0020615 22.1172243,
    -83.1533981 22.0000091, -83.153409643632642 22, -83.5298736473797 22,
    -83.5299019 22.0000315, -83.552786 22.0213648, -83.5956953 22.0309002,
    -83.6243016 22.0390734, -83.6481401 22.0588253, -83.6869628 22.0656363,
    -83.7067147 22.0588253, -83.7135257 22.0438411, -83.7162501 22.0193215,
    -83.7135257 22.0118294, -83.7061155 22.0000417, -83.706027334522048 22)), 
  POLYGON((-82 22.139622493008307, -82 22.058731622880629, -82.1372839 22.0739854,
    -82.2291665 22.1388437, -82.364288 22.1928923, -82.477790099999993 22.241536,
    -82.5858873 22.2685604, -82.6490561 22.3165748, -82.7264137 22.3334187,
    -82.8993692 22.2739652, -82.9966567 22.1766777, -83 22.139901028852119,
    -83 22.437513045861383, -82.9151228 22.4807243, -82.866558807179359 22.5,
    -82.76704697765075 22.5, -82.7449362 22.493551, -82.648907999999992 22.4571265,
    -82.5595024 22.4008341, -82.5065213 22.367721, -82.2548611 22.3776549,
    -82.2051913 22.3842775, -82.1488989 22.3213625, -82.1025404 22.2253343,
    -82.0263801 22.1491739, -82 22.139622493008307)), 
  POLYGON((-81.75459699999999 22, -81.8508264 22, -81.9913526 22.0577708,
    -82 22.058731622880629, -82 22.139622493008307, -81.8343237 22.0796363,
    -81.75459699999999 22)), 
  POLYGON((-77.743157706583418 22.5, -77 22.5, -77 22,
    -77.151692579017492 22, -77.1516929 22.0000004, -77.3222053 22.1877442,
    -77.6494792 22.4132606, -77.743157706583418 22.5)), 
  POLYGON((-77 22, -77 22.5, -76 22.5,
    -76 22, -77 22)), 
  POLYGON((-75.802530565114523 22.5, -75 22.5, -75 22,
    -76 22, -76 22.5, -75.852842279456127 22.5,
    -75.8553298 22.4930063, -75.8744284 22.4786824, -75.8624918 22.4595838,
    -75.8601044 22.430936, -75.8529425 22.3999008, -75.8386186 22.3760276,
    -75.81952 22.3449924, -75.8075834 22.3306685, -75.8004214 22.3091826,
    -75.8051961 22.2781474, -75.7908722 22.2208517, -75.7765482 22.1945912,
    -75.75983699999999 22.170717999999997, -75.7335765 22.1492321, -75.6643442 22.1372955,
    -75.6595695 22.1420702, -75.6547949 22.1731053, -75.6524076 22.1922039,
    -75.6667315 22.2208517, -75.6810554 22.2566615, -75.6810554 22.2805348,
    -75.6953793 22.3115699, -75.7049286 22.3330558, -75.7192525 22.3545417,
    -75.7407384 22.3688656, -75.7502877 22.3736403, -75.7574497 22.3784149,
    -75.7789355 22.3975135, -75.7884848 22.416612, -75.7980341 22.4333233,
    -75.7980341 22.4548092, -75.7956468 22.4810697, -75.802530565114523 22.5)), 
  POLYGON((-74 22.1924167, -74 22, -75 22,
    -75 22.5, -74.402473166666667 22.5, -74.4001907 22.4965763,
    -74.3577498 22.4965763, -74.355782270526433 22.5, -74.3187901030562 22.5,
    -74.3153089 22.4682824, -74.2905518 22.4435252, -74.2516476 22.4293783,
    -74.2253592 22.470423, -74.219233562228254 22.5, -74.0182228 22.5,
    -74.0182228 22.4965763, -74.0323698 22.4647457, -74.0677372 22.4435252,
    -74.1544709 22.4197157, -74.1950598 22.365717, -74.2127435 22.3409598,
    -74.2233537 22.3162026, -74.2375007 22.2737617, -74.2587211 22.2666882,
    -74.2976252 22.2666882, -74.3153089 22.2525413, -74.3329926 22.2277841,
    -74.3400661 22.1994902, -74.3612866 22.1570493, -74.3542131 22.1322921,
    -74.3471396 22.1039982, -74.2976252 22.0969247, -74.1844496 22.1216819,
    -74.0465167 22.1676595, -74 22.1924167)), 
  POLYGON((-73 22.2879087, -73 22, -74 22,
    -74 22.1924167, -73.9156574 22.3197394, -73.8555328 22.3904742,
    -73.8024817 22.4258416, -73.784798 22.461209, -73.7781287465491 22.5,
    -73.194071232856757 22.5, -73.2159198 22.4336511, -73.2153829 22.3551068,
    -73.1941625 22.3303496, -73.1305011 22.291445500000002, -73 22.2879087)), 
  POLYGON((-72.9701884761539 22.5, -72 22.5, -72 22,
    -73 22, -73 22.2879087, -72.9253702 22.2949822,
    -72.9006131 22.2949822, -72.8652457 22.2808352, -72.8263415 22.2454678,
    -72.7662169 22.2242474, -72.7096291 22.2419311, -72.6671882 22.2808352,
    -72.6388943 22.3055924, -72.6742617 22.3904742, -72.7732904 22.4364518,
    -72.8546354 22.4647457, -72.928907 22.4753559, -72.9701884761539 22.5)), 
  POLYGON((-72 22, -72 22.5, -71 22.5,
    -71 22, -72 22)), 
  POLYGON((-71 22, -71 22.5, -70 22.5,
    -70 22, -71 22)), 
  POLYGON((-70 22, -70 22.5, -69 22.5,
    -69 22, -70 22)), 
  POLYGON((-69 22, -69 22.5, -68 22.5,
    -68 22, -69 22)), 
  POLYGON((-68 22, -68 22.5, -67 22.5,
    -67 22, -68 22)), 
  POLYGON((-67 22, -67 22.5, -66 22.5,
    -66 22, -67 22)), 
  POLYGON((-66 22, -66 22.5, -65 22.5,
    -65 22, -66 22)), 
  POLYGON((-65 22, -65 22.5, -64 22.5,
    -64 22, -65 22)), 
  POLYGON((-64 22, -64 22.5, -63 22.5,
    -63 22, -64 22)), 
  POLYGON((-63 22, -63 22.5, -62 22.5,
    -62 22, -63 22)), 
  POLYGON((-62 22, -62 22.5, -61 22.5,
    -61 22, -62 22)), 
  POLYGON((-61 22, -61 22.5, -60 22.5,
    -60 22, -61 22)), 
  POLYGON((-60 22, -60 22.5, -59 22.5,
    -59 22, -60 22)), 
  POLYGON((-59 22, -59 22.5, -58 22.5,
    -58 22, -59 22)), 
  POLYGON((-58 22, -58 22.5, -57 22.5,
    -57 22, -58 22)), 
  POLYGON((-57 22, -57 22.5, -56 22.5,
    -56 22, -57 22)), 
  POLYGON((-56 22, -56 22.5, -55 22.5,
    -55 22, -56 22)), 
  LINESTRING(-86 22.5, -86 23), 
  POLYGON((-85 23, -85 22.5, -86 22.5,
    -86 23, -85 23)), 
  POLYGON((-84.480686463980788 22.5, -85 22.5, -85 23,
    -84 23, -84 22.907466, -84.2141016 22.7865248,
    -84.2449636 22.7220321, -84.2946334 22.6955416, -84.4535766 22.5365983,
    -84.480686463980788 22.5)), 
  POLYGON((-84 23, -83.8674734 23, -84 22.907466,
    -84 23)), 
  POLYGON((-82.76704697765075 22.5, -82.866558807179359 22.5, -82.8244078 22.5167303,
    -82.76704697765075 22.5)), 
  POLYGON((-79.255589560022642 23, -79 23, -79 22.8615335876345,
    -79.0002081 22.8615397, -79.255589560022642 23)), 
  POLYGON((-79 22.8615335876345, -79 23, -78 23,
    -78 22.734999062608978, -78.0000381 22.7350335, -78.2627739 22.8037891,
    -78.6258003 22.8505425, -79 22.8615335876345)), 
  POLYGON((-78 22.734999062608978, -78 23, -77 23,
    -77 22.5, -77.743157706583418 22.5, -77.8722454 22.6195257,
    -78 22.734999062608978)), 
  POLYGON((-77 22.5, -77 23, -76 23,
    -76 22.5, -77 22.5)), 
  POLYGON((-75.852842279456127 22.5, -76 22.5, -76 23,
    -75.738076263859483 23, -75.7336625 22.9968921, -75.7295543 22.9971544,
    -75.723570741096282 23, -75.065097319340069 23, -75 22.9315953,
    -75 22.5, -75.802530565114523 22.5, -75.8051961 22.5073302,
    -75.8099707 22.5192668, -75.8219073 22.5264288, -75.8417521 22.5311802,
    -75.852842279456127 22.5), 
    (-75.8434266 22.9117624, -75.851635 22.9176103, -75.8623271 22.9158283, 
      -75.8739102 22.9095913, -75.8819293 22.8988992, -75.8846023 22.8926621, 
      -75.8854933 22.881079, -75.8748012 22.8659319, -75.8523332 22.8713393, 
      -75.8409429 22.878406, -75.8373788 22.8908801, -75.8355968 22.9006812, 
      -75.8434266 22.9117624)), 
  GEOMETRYCOLLECTION(
  POLYGON((-74.402473166666667 22.5, -75 22.5, -75 22.9315953,
    -74.9413119 22.8714708, -74.9094812 22.8290299, -74.8458199 22.8078094,
    -74.817526 22.8078094, -74.7715484 22.8325666, -74.7397177 22.8608605,
    -74.7432545 22.882081, -74.7456498 22.9224987, -74.775023267284467 23,
    -74 23, -74 22.824986945251748, -74.0394432 22.8219564,
    -74.0642004 22.8254931, -74.110177999999991 22.8290299, -74.1278617 22.8219564,
    -74.1490822 22.8078094, -74.1950598 22.8184197, -74.2162802 22.8396401,
    -74.2516476 22.8573238, -74.2799415 22.8785442, -74.3648233 22.8785442,
    -74.3718968 22.8254931, -74.3754335 22.786589, -74.3365294 22.7193909,
    -74.3294559 22.7017072, -74.3153089 22.691097, -74.3259192 22.6663398,
    -74.3754335 22.6415827, -74.396654 22.623898999999998, -74.4214111 22.5814581,
    -74.4284846 22.5425539, -74.4143377 22.5177968, -74.402473166666667 22.5)), 
    POLYGON((-74.3187901030562 22.5, -74.355782270526433 22.5, -74.3430564 22.5221443,
      -74.3117722 22.5496274, -74.3194762 22.5062511, -74.3187901030562 22.5)), 
    POLYGON((-74.0182228 22.5, -74.219233562228254 22.5, -74.2162802 22.51426,
      -74.219816999999992 22.5319437, -74.2622578 22.5602376, -74.2799415 22.5602376,
      -74.2834783 22.5885316, -74.2764048 22.6132887, -74.2693313 22.623898999999998,
      -74.2410374 22.6309724, -74.2268904 22.623898999999998, -74.1773761 22.623898999999998,
      -74.1420087 22.6345092, -74.1066413 22.6345092, -74.0677372 22.5920683,
      -74.0394432 22.5602376, -74.0182228 22.5354805, -74.0182228 22.5))), 
  POLYGON((-73.194071232856757 22.5, -73.7781287465491 22.5, -73.7751464 22.5173465,
    -73.8130919 22.6168255, -73.8166287 22.6734134, -73.7883348 22.7406114,
    -73.8307756 22.7901258, -73.9333411 22.7795156, -73.9954765 22.8253345,
    -74 22.824986945251748, -74 23, -73 23,
    -73 22.5177968, -73.0809868 22.5567009, -73.1234277 22.5531642,
    -73.1835522 22.5319438, -73.194071232856757 22.5), 
    (-73.6056121 22.5224578, -73.5549099 22.5177968, -73.4470944 22.5290233, 
      -73.4205138 22.5743846, -73.4063669 22.6345092, -73.445270999999991 22.6875603, 
      -73.5053956 22.7123175, -73.5690569 22.7123175, -73.6468652 22.7087808, 
      -73.6822326 22.6910971, -73.703453 22.6486562, -73.6999163 22.6132888, 
      -73.6822326 22.5743846, -73.6327182 22.5319438, -73.6056121 22.5224578)), 
  POLYGON((-73 22.5177968, -73 23, -72 23,
    -72 22.5, -72.9701884761539 22.5, -73 22.5177968)), 
  POLYGON((-72 22.5, -72 23, -71 23,
    -71 22.5, -72 22.5)), 
  POLYGON((-71 22.5, -71 23, -70 23,
    -70 22.5, -71 22.5)), 
  POLYGON((-70 22.5, -70 23, -69 23,
    -69 22.5, -70 22.5)), 
  POLYGON((-69 22.5, -69 23, -68 23,
    -68 22.5, -69 22.5)), 
  POLYGON((-68 22.5, -68 23, -67 23,
    -67 22.5, -68 22.5)), 
  POLYGON((-67 22.5, -67 23, -66 23,
    -66 22.5, -67 22.5)), 
  POLYGON((-66 22.5, -66 23, -65 23,
    -65 22.5, -66 22.5)), 
  POLYGON((-65 22.5, -65 23, -64 23,
    -64 22.5, -65 22.5)), 
  POLYGON((-64 22.5, -64 23, -63 23,
    -63 22.5, -64 22.5)), 
  POLYGON((-63 22.5, -63 23, -62 23,
    -62 22.5, -63 22.5)), 
  POLYGON((-62 22.5, -62 23, -61 23,
    -61 22.5, -62 22.5)), 
  POLYGON((-61 22.5, -61 23, -60 23,
    -60 22.5, -61 22.5)), 
  POLYGON((-60 22.5, -60 23, -59 23,
    -59 22.5, -60 22.5)), 
  POLYGON((-59 22.5, -59 23, -58 23,
    -58 22.5, -59 22.5)), 
  POLYGON((-58 22.5, -58 23, -57 23,
    -57 22.5, -58 22.5)), 
  POLYGON((-57 22.5, -57 23, -56 23,
    -56 22.5, -57 22.5)), 
  POLYGON((-56 22.5, -56 23, -55 23,
    -55 22.5, -56 22.5)), 
  LINESTRING(-86 23, -85 23), 
  MULTILINESTRING((-85 23, -84 23), 
    (-84 23, -84 23.5)), 
  POLYGON((-83.8674734 23, -84 23, -84 23.5,
    -83 23.5, -83 23.202568, -83.291304 23.1425696,
    -83.4436246 23.1061451, -83.8674734 23)), 
  POLYGON((-83 23.202568, -83 23.5, -82.042995143231877 23.5,
    -83 23.202568)), 
  POLYGON((-81.975269417587484 23.5, -81 23.5, -81 23.4830886,
    -81.6940544 23.4107255, -81.8479743 23.4315428, -81.975269417587484 23.5)), 
  POLYGON((-81 23.4830886, -81 23.5, -80 23.5,
    -80 23.274069270424331, -80.0000102 23.2740734, -80.470655999999991 23.3605786,
    -81 23.4830886)), 
  POLYGON((-79.498407114634148 23.5, -79 23.5, -79 23,
    -79.255589560022642 23, -79.2555962 23.0000036, -79.3546035 23.0705585,
    -79.6116027 23.1168226, -80 23.274069270424331, -80 23.5,
    -79.574653942499992 23.5, -79.5680412 23.4932177, -79.5545135 23.4895756,
    -79.5425466 23.4838523, -79.531100000000009 23.4812508, -79.5139301 23.4843726,
    -79.5040444 23.4921771, -79.4972805 23.4979004, -79.498407114634148 23.5)), 
  POLYGON((-79 23, -79 23.5, -78 23.5,
    -78 23, -79 23)), 
  POLYGON((-78 23, -78 23.5, -77 23.5,
    -77 23, -78 23)), 
  POLYGON((-76 23.39114981762464, -76 23, -77 23,
    -77 23.5, -76.034410638982848 23.5, -76.0360545 23.4539715,
    -76.0335353 23.4161831, -76 23.39114981762464)), 
  GEOMETRYCOLLECTION(
  POLYGON((-75.738076263859483 23, -76 23, -76 23.39114981762464,
    -75.9997922 23.3909947, -75.869785999999991 23.3985487, -75.7765748 23.3935103,
    -75.711075 23.4010679, -75.6229023 23.3859526, -75.4289222 23.333049,
    -75.4138069 23.3935103, -75.4389991 23.4413755, -75.4919028 23.4640485,
    -75.6002293 23.4967984, -75.60590485864742 23.5, -75.297512175509738 23.5,
    -75.2914491 23.4939369, -75.2843756 23.4691798, -75.2773021 23.4550328,
    -75.2773021 23.4267389, -75.2313245 23.3666143, -75.1959571 23.3453939,
    -75.1959571 23.3241734, -75.2136408 23.3064897, -75.2624907 23.2703806,
    -75.2737654 23.2003875, -75.2773021 23.1614834, -75.2348612 23.1084323,
    -75.1570529 23.0624547, -75.0650977 23.0000004, -75.065097319340069 23,
    -75.723570741096282 23, -75.7235699 23.0000004, -75.7206157 23.0041272,
    -75.7160841 23.0133523, -75.7157604 23.0183695, -75.7164077 23.0225774,
    -75.7193209 23.0250051, -75.7267658 23.0233867, -75.7312974 23.0207972,
    -75.7368001 23.0109247, -75.7384186 23.0055838, -75.7380774 23.0000008,
    -75.738076263859483 23)), 
    POLYGON((-75.149744030187748 23.5, -75 23.5, -75 23.260962479062222,
      -75.0181097 23.2746786, -75.0247428 23.277773999999997, -75.0884756 23.3912246,
      -75.1126444 23.4298947, -75.1368132 23.4717872, -75.149744030187748 23.5))), 
  POLYGON((-75 23.260962479062222, -75 23.5, -74 23.5,
    -74 23, -74.775023267284467 23, -74.7750985 23.0001985,
    -74.9006603 23.1857235, -75 23.260962479062222)), 
  POLYGON((-74 23, -74 23.5, -73 23.5,
    -73 23, -74 23), 
    (-73.8457664 23.1389836, -73.863987999999992 23.1184843, -73.863987999999992 23.0957073, 
      -73.8571549 23.0660972, -73.8457664 23.0478756, -73.7865463 23.0364871, 
      -73.7478254 23.0250986, -73.6681059 23.0205432, -73.6316627 23.0501533, 
      -73.6020497 23.0906354, -73.6407735 23.1298728, -73.674939 23.1526498, 
      -73.7296038 23.1526498, -73.8024902 23.1480944, -73.8457664 23.1389836)), 
  POLYGON((-73 23, -73 23.5, -72 23.5,
    -72 23, -73 23)), 
  POLYGON((-72 23, -72 23.5, -71 23.5,
    -71 23, -72 23)), 
  POLYGON((-71 23, -71 23.5, -70 23.5,
    -70 23, -71 23)), 
  POLYGON((-70 23, -70 23.5, -69 23.5,
    -69 23, -70 23)), 
  POLYGON((-69 23, -69 23.5, -68 23.5,
    -68 23, -69 23)), 
  POLYGON((-68 23, -68 23.5, -67 23.5,
    -67 23, -68 23)), 
  POLYGON((-67 23, -67 23.5, -66 23.5,
    -66 23, -67 23)), 
  POLYGON((-66 23, -66 23.5, -65 23.5,
    -65 23, -66 23)), 
  POLYGON((-65 23, -65 23.5, -64 23.5,
    -64 23, -65 23)), 
  POLYGON((-64 23, -64 23.5, -63 23.5,
    -63 23, -64 23)), 
  POLYGON((-63 23, -63 23.5, -62 23.5,
    -62 23, -63 23)), 
  POLYGON((-62 23, -62 23.5, -61 23.5,
    -61 23, -62 23)), 
  POLYGON((-61 23, -61 23.5, -60 23.5,
    -60 23, -61 23)), 
  POLYGON((-60 23, -60 23.5, -59 23.5,
    -59 23, -60 23)), 
  POLYGON((-59 23, -59 23.5, -58 23.5,
    -58 23, -59 23)), 
  POLYGON((-58 23, -58 23.5, -57 23.5,
    -57 23, -58 23)), 
  POLYGON((-57 23, -57 23.5, -56 23.5,
    -56 23, -57 23)), 
  POLYGON((-56 23, -56 23.5, -55 23.5,
    -55 23, -56 23)), 
  LINESTRING(-84 23.5, -84 23.8946), 
  POLYGON((-83 23.879163354603467, -83 23.5, -84 23.5,
    -84 23.8946, -83.928400000000011 23.8946, -83.25030000000001 23.8746,
    -83 23.879163354603467)), 
  POLYGON((-82.042995143231877 23.5, -83 23.5, -83 23.879163354603467,
    -82.1533 23.8946, -82.0002563 23.9080114, -82 23.90804152398804,
    -82 23.513299696472384, -82.0000742 23.5133396, -82.042995143231877 23.5)), 
  POLYGON((-81.28411 24, -81 24, -81 23.5,
    -81.975269417587484 23.5, -82 23.513299696472384, -82 23.90804152398804,
    -81.924075 23.9169653, -81.6719599 23.9479022, -81.5592689 23.963299,
    -81.537351 23.9658157, -81.28411 24)), 
  POLYGON((-80.3080659 24, -80 24, -80 23.5,
    -81 23.5, -81 24, -80.4307443 24,
    -80.4595103 23.9791088, -80.484045999999992 23.9706482, -80.5094278 23.9528809,
    -80.5153502 23.939344, -80.5187344 23.9198847, -80.512812 23.9063477,
    -80.4908145 23.9004253, -80.4696173 23.9107836, -80.4468195 23.9308834,
    -80.4451273 23.9418822, -80.4341286 23.9511888, -80.4222838 23.9579573,
    -80.3935178 23.9723403, -80.3740584 23.9723403, -80.3410622 23.9858772,
    -80.3080659 24), 
    (-80.4121311 23.7371361, -80.4239759 23.735443999999998, -80.4332825 23.7312137, 
      -80.4366668 23.7202149, -80.4298983 23.7041398, -80.4239759 23.7016017, 
      -80.4055007 23.6908428, -80.3918257 23.6999095, -80.3875954 23.7092162, 
      -80.3884414 23.7202149, -80.3943639 23.7278295, -80.4121311 23.7371361)), 
  POLYGON((-79.574653942499992 23.5, -80 23.5, -80 24,
    -79 24, -79 23.5, -79.498407114634148 23.5,
    -79.5087271 23.5192327, -79.5279782 23.5311996, -79.5487902 23.5494101,
    -79.5794878 23.587391999999998, -79.6060231 23.6113258, -79.6221524 23.61861,
    -79.6419238 23.6118461, -79.6460862 23.6102852, -79.6445253 23.5931153,
    -79.6382817 23.5764657, -79.6315178 23.5655394, -79.6221524 23.5572146,
    -79.6138276 23.5483695, -79.5883329 23.514029700000002, -79.574653942499992 23.5)), 
  POLYGON((-79 23.5, -79 24, -78 24,
    -78 23.5, -79 23.5)), 
  POLYGON((-77.468210646366444 24, -77 24, -77 23.5,
    -78 23.5, -78 24, -77.8578364 24,
    -77.8388876 23.9768038, -77.8373716 23.9616447, -77.8396455 23.9457276,
    -77.831307999999993 23.9207151, -77.8161489 23.9116196, -77.8290341 23.8979764,
    -77.8404035 23.8911548, -77.847983 23.871447999999997, -77.8499459 23.8606161,
    -77.8373716 23.8479514, -77.8358557 23.8441616, -77.8252444 23.8297605,
    -77.8252444 23.8282446, -77.8275182 23.8229389, -77.831307999999993 23.8161173,
    -77.8373716 23.8115695, -77.8426773 23.80399, -77.8449512 23.7986843,
    -77.84874099999999 23.787315, -77.84874099999999 23.7804934, -77.8472251 23.7668502,
    -77.8343398 23.7592706, -77.8153909 23.7592706, -77.8093273 23.7600286,
    -77.8009898 23.7456274, -77.7934102 23.735774, -77.7827989 23.719856999999998,
    -77.7729455 23.6956024, -77.763849999999991 23.6751376, -77.7608182 23.6668001,
    -77.7585443 23.6584626, -77.7449011 23.6584626, -77.7153409 23.6614944,
    -77.6562204 23.6796853, -77.6410613 23.6910547, -77.6221124 23.6918126,
    -77.6190806 23.6956024, -77.6137749 23.6963604, -77.6046794 23.7001501,
    -77.5970999 23.7054558, -77.5895203 23.7092456, -77.578150999999991 23.7115195,
    -77.5766351 23.7115195, -77.5501066 23.7145513, -77.5341896 23.7592706,
    -77.5235782 23.786557, -77.5190305 23.8017161, -77.5159987 23.8168752,
    -77.5046293 23.8267287, -77.4826486 23.8343082, -77.4781009 23.8555309,
    -77.4781009 23.8737219, -77.4788588 23.8919128, -77.4727952 23.9010083,
    -77.4727952 23.9191992, -77.4811327 23.9298106, -77.4970498 23.939664,
    -77.4970498 23.9480015, -77.4864384 23.9699822, -77.4811327 23.9775617,
    -77.468210646366444 24)), 
  GEOMETRYCOLLECTION(
  POLYGON((-76.034410638982848 23.5, -77 23.5, -77 24,
    -76.3963033 24, -76.3232458 23.9149891, -76.3131689 23.8847584,
    -76.2653037 23.8444508, -76.2275154 23.7991049, -76.1897271 23.7764319,
    -76.1947655 23.7537589, -76.199804 23.7235282, -76.1846886 23.6958168,
    -76.1670541 23.675663, -76.1141504 23.6630669, -76.086439 23.6605477,
    -76.0738429 23.6429131, -76.0486506 23.6277978, -76.0688044 23.6000863,
    -76.0662852 23.5799326, -76.0335353 23.5245097, -76.034410638982848 23.5)), 
    POLYGON((-76.2804191 24, -76 24, -76 23.7411629,
      -76.0461314 23.7991049, -76.131785 23.8696431, -76.2804191 24))), 
  POLYGON((-75.297512175509738 23.5, -75.60590485864742 23.5, -75.6984789 23.5522213,
    -75.8168824 23.6051249, -75.9201705 23.68574, -76 23.7411629,
    -76 24, -75 24, -75 23.5,
    -75.149744030187748 23.5, -75.1545369 23.5104572, -75.1770945 23.5684623,
    -75.218987 23.6087436, -75.2367108 23.6409686, -75.2592683 23.6748049,
    -75.2818258 23.7021962, -75.2963271 23.7183087, -75.3188846 23.726364999999998,
    -75.3559434 23.713475, -75.3672221 23.7005849, -75.3720559 23.6860837,
    -75.3720559 23.6506361, -75.3688334 23.6119661, -75.3704446 23.5813523,
    -75.3511095 23.5732959, -75.3162062 23.5186941, -75.297512175509738 23.5), 
    (-75.1050318 23.9147324, -75.120589 23.9120045, -75.1338474 23.9102751, 
      -75.149988 23.8981696, -75.167858 23.8872171, -75.1851515 23.8779938, 
      -75.1880338 23.8537829, -75.1955276 23.8353365, -75.1724696 23.7984436, 
      -75.1228948 23.7719269, -75.0771906 23.7713498, -75.0338476 23.7894448, 
      -75.0033689 23.8311773, -75.0005554 23.8546225, -75.0099335 23.8710341, 
      -75.0202494 23.8869769, -75.0605752 23.9127666, -75.1050318 23.9147324)), 
  POLYGON((-74.4056574 24, -74 24, -74 23.5,
    -75 23.5, -75 24, -74.579231992870874 24,
    -74.5790386 23.9973173, -74.5692719 23.8912832, -74.5471918 23.8755579,
    -74.5063432 23.8912832, -74.4276823 23.9258939, -74.4056574 24), 
    (-74.997187 23.7402543, -74.997187 23.6710328, -74.9657226 23.6206898, 
      -74.9216725 23.6175434, -74.8618903 23.6238362, -74.7325898 23.6155934, 
      -74.729739999999993 23.7119364, -74.7674973 23.7591329, -74.8272795 23.7937437, 
      -74.9153797 23.781157999999998, -74.997187 23.7402543)), 
  POLYGON((-74 23.5, -74 24, -73 24,
    -73 23.5, -74 23.5)), 
  POLYGON((-73 23.5, -73 24, -72 24,
    -72 23.5, -73 23.5)), 
  POLYGON((-72 23.5, -72 24, -71 24,
    -71 23.5, -72 23.5)), 
  POLYGON((-71 23.5, -71 24, -70 24,
    -70 23.5, -71 23.5)), 
  POLYGON((-70 23.5, -70 24, -69 24,
    -69 23.5, -70 23.5)), 
  POLYGON((-69 23.5, -69 24, -68 24,
    -68 23.5, -69 23.5)), 
  POLYGON((-68 23.5, -68 24, -67 24,
    -67 23.5, -68 23.5)), 
  POLYGON((-67 23.5, -67 24, -66 24,
    -66 23.5, -67 23.5)), 
  POLYGON((-66 23.5, -66 24, -65 24,
    -65 23.5, -66 23.5)), 
  POLYGON((-65 23.5, -65 24, -64 24,
    -64 23.5, -65 23.5)), 
  POLYGON((-64 23.5, -64 24, -63 24,
    -63 23.5, -64 23.5)), 
  POLYGON((-63 23.5, -63 24, -62 24,
    -62 23.5, -63 23.5)), 
  POLYGON((-62 23.5, -62 24, -61 24,
    -61 23.5, -62 23.5)), 
  POLYGON((-61 23.5, -61 24, -60 24,
    -60 23.5, -61 23.5)), 
  POLYGON((-60 23.5, -60 24, -59 24,
    -59 23.5, -60 23.5)), 
  POLYGON((-59 23.5, -59 24, -58 24,
    -58 23.5, -59 23.5)), 
  POLYGON((-58 23.5, -58 24, -57 24,
    -57 23.5, -58 23.5)), 
  POLYGON((-57 23.5, -57 24, -56 24,
    -56 23.5, -57 23.5)), 
  POLYGON((-56 23.5, -56 24, -55 24,
    -55 23.5, -56 23.5)), 
  POLYGON((-81 24.065429563883175, -81 24, -81.28411 24,
    -81.000002999999992 24.065428999999998, -81 24.065429563883175)), 
  POLYGON((-80.4307443 24, -81 24, -81 24.065429563883175,
    -80.9060346 24.0830914, -80.4812755 24.4079368, -80.2797097 24.4958826,
    -80.277086932269867 24.5, -80 24.5, -80 24,
    -80.3080659 24, -80.3004514 24.0095669, -80.2987592 24.0154893,
    -80.3004514 24.0231038, -80.3080659 24.0239498, -80.3199107 24.026488,
    -80.3478306 24.0281801, -80.3689821 24.0231038, -80.3867493 24.0214117,
    -80.4104389 24.012105, -80.4307443 24)), 
  POLYGON((-80 24, -80 24.5, -79 24.5,
    -79 24, -80 24)), 
  POLYGON((-78 24.144791516790796, -78 24, -79 24,
    -79 24.5, -78.397268400764588 24.5, -78.3025799 24.4345071,
    -78.1311988 24.3441532, -78.0967072 24.2546409, -78.0543063 24.2059584,
    -78.0244686 24.1776911, -78.0000003 24.1447916, -78 24.144791516790796)), 
  GEOMETRYCOLLECTION(
  POLYGON((-77.8578364 24, -78 24, -78 24.144791516790796,
    -77.960082 24.1337197, -77.9443779 24.1337197, -77.9333851 24.1211565,
    -77.9475188 24.1101637, -77.9490892 24.0944596, -77.9443779 24.0818964,
    -77.9349555 24.0677628, -77.9239627 24.0614812, -77.8878434 24.0551996,
    -77.8847026 24.0473475, -77.8497155 24.0167547, -77.8419608 24.0130598,
    -77.8432891 24.0073038, -77.8509782 24.0051433, -77.8578364 24)), 
    POLYGON((-77.662171410111611 24.5, -77 24.5, -77 24,
      -77.468210646366444 24, -77.4679485 24.0004552, -77.4746734 24.0096642,
      -77.5062669 24.1120677, -77.5166614 24.1432512, -77.5270559 24.1813644,
      -77.5478449 24.229872, -77.5824932 24.2402665, -77.5755636 24.2783796,
      -77.6102119 24.2991686, -77.6413954 24.3303521, -77.6587195 24.3823246,
      -77.6552547 24.4377619, -77.6587195 24.4654806, -77.662171410111611 24.5), 
      (-77.1963052 24.0275555, -77.1904822 24.0109058, -77.1805726 24.0089239, 
        -77.1657083 24.0099148, -77.1557987 24.0148696, -77.1429163 24.0218063, 
        -77.1389525 24.0356797, -77.149853 24.0495531, -77.1756178 24.0545078, 
        -77.1855274 24.050544, -77.1963052 24.0275555))), 
  GEOMETRYCOLLECTION(
  POLYGON((-76.3963033 24, -77 24, -77 24.5,
    -76.795621453661354 24.5, -76.7943402 24.4969291, -76.7785424 24.4398262,
    -76.7363981 24.3760065, -76.6079178 24.2424877, -76.5474565 24.199661,
    -76.542418 24.1744688, -76.4920336 24.1240843, -76.4441684 24.108969,
    -76.4265338 24.0787383, -76.4189761 24.0485077, -76.3963033 24)), 
    POLYGON((-76.689766972974681 24.5, -76 24.5, -76 24,
      -76.2804191 24, -76.3358418 24.1467573, -76.40638 24.2248532,
      -76.4466876 24.2727184, -76.5071489 24.3155452, -76.5625718 24.3835642,
      -76.6230331 24.4188333, -76.6583022 24.4742561, -76.689766972974681 24.5))), 
  POLYGON((-75.525446717888414 24.5, -75 24.5, -75 24,
    -76 24, -76 24.5, -75.630983094713983 24.5,
    -75.6302139 24.4998077, -75.6065678 24.492532, -75.5883785 24.4779805,
    -75.573827 24.4597912, -75.5629134 24.437963999999997, -75.5592756 24.4161368,
    -75.548362 24.3997664, -75.5356294 24.3743013, -75.5301726 24.3615688,
    -75.5156212 24.3397416, -75.4992508 24.3106387, -75.4883372 24.2979062,
    -75.4737857 24.2851736, -75.4410449 24.2651654, -75.4374071 24.2597086,
    -75.4501396 24.2287867, -75.4574153 24.221511, -75.5065265 24.1796755,
    -75.546543 24.1578484, -75.548362 24.1323833, -75.519258999999991 24.0978236,
    -75.4956129 24.0941857, -75.4737857 24.08691, -75.4246745 24.0887289,
    -75.4173988 24.1014614, -75.3792012 24.1087372, -75.3555551 24.1032804,
    -75.2937114 24.0978236, -75.2700653 24.114193999999998, -75.2664274 24.1523916,
    -75.2664274 24.1796755, -75.2737031 24.2160542, -75.2864357 24.2397003,
    -75.3100818 24.2724411, -75.3282711 24.3070008, -75.3537362 24.3397416,
    -75.3828391 24.3597499, -75.4101231 24.3815771, -75.4374071 24.3924907,
    -75.4592343 24.4070421, -75.4901561 24.4325072, -75.5156212 24.4688858,
    -75.525446717888414 24.5)), 
  POLYGON((-74.579231992870874 24, -75 24, -75 24.5,
    -74 24.5, -74 24, -74.4056574 24,
    -74.3899252 24.0234335, -74.3647537 24.0958015, -74.3836323 24.1587302,
    -74.3993645 24.1807552, -74.5094897 24.1870481, -74.5315148 24.1618766,
    -74.5850041 24.1178265, -74.5850041 24.0800693, -74.579231992870874 24)), 
  POLYGON((-74 24, -74 24.5, -73 24.5,
    -73 24, -74 24)), 
  POLYGON((-73 24, -73 24.5, -72 24.5,
    -72 24, -73 24)), 
  POLYGON((-72 24, -72 24.5, -71 24.5,
    -71 24, -72 24)), 
  POLYGON((-71 24, -71 24.5, -70 24.5,
    -70 24, -71 24)), 
  POLYGON((-70 24, -70 24.5, -69 24.5,
    -69 24, -70 24)), 
  POLYGON((-69 24, -69 24.5, -68 24.5,
    -68 24, -69 24)), 
  POLYGON((-68 24, -68 24.5, -67 24.5,
    -67 24, -68 24)), 
  POLYGON((-67 24, -67 24.5, -66 24.5,
    -66 24, -67 24)), 
  POLYGON((-66 24, -66 24.5, -65 24.5,
    -65 24, -66 24)), 
  POLYGON((-65 24, -65 24.5, -64 24.5,
    -64 24, -65 24)), 
  POLYGON((-64 24, -64 24.5, -63 24.5,
    -63 24, -64 24)), 
  POLYGON((-63 24, -63 24.5, -62 24.5,
    -62 24, -63 24)), 
  POLYGON((-62 24, -62 24.5, -61 24.5,
    -61 24, -62 24)), 
  POLYGON((-61 24, -61 24.5, -60 24.5,
    -60 24, -61 24)), 
  POLYGON((-60 24, -60 24.5, -59 24.5,
    -59 24, -60 24)), 
  POLYGON((-59 24, -59 24.5, -58 24.5,
    -58 24, -59 24)), 
  POLYGON((-58 24, -58 24.5, -57 24.5,
    -57 24, -58 24)), 
  POLYGON((-57 24, -57 24.5, -56 24.5,
    -56 24, -57 24)), 
  POLYGON((-56 24, -56 24.5, -55 24.5,
    -55 24, -56 24)), 
  POLYGON((-80 24.93499, -80 24.5, -80.277086932269867 24.5,
    -80 24.93499)), 
  POLYGON((-79.926988529610341 25, -79 25, -79 24.5,
    -80 24.5, -80 24.93499, -79.926988529610341 25)), 
  POLYGON((-78.397268400764588 24.5, -79 24.5, -79 25,
    -78.2083545 25, -78.2448297 24.9171678, -78.2482998 24.8859371,
    -78.25524 24.8685867, -78.3211714 24.8234758, -78.3419918 24.7748947,
    -78.3589646 24.75717, -78.3730982 24.746177199999998, -78.3856614 24.7367548,
    -78.4148634 24.7228436, -78.442624 24.7159034, -78.487735 24.7228436,
    -78.5259058 24.7124334, -78.5536664 24.7089633, -78.5744868 24.6916129,
    -78.5675467 24.6742625, -78.5361949 24.6218956, -78.5124341 24.5796563,
    -78.397268400764588 24.5)), 
  POLYGON((-77.2203517 25, -77 25, -77 24.5,
    -77.662171410111611 24.5, -77.6621843 24.5001289, -77.6933678 24.5555662,
    -77.6795085 24.5867497, -77.6587195 24.642187, -77.6795085 24.7114836,
    -77.7384106 24.7634561, -77.7591996 24.8188934, -77.7799886 24.867400999999997,
    -77.7772632 24.9130074, -77.8181018 25, -77.6195158 25,
    -77.6083273 24.9778216, -77.5630187 24.9467229, -77.5126626 24.9270717,
    -77.4475682 24.9123333, -77.3738763 24.9172461, -77.3038432 24.9406091,
    -77.253513 24.9651458, -77.2203517 25)), 
  POLYGON((-76.795621453661354 24.5, -77 24.5, -77 25,
    -76.2149192 25, -76.235073 24.9226776, -76.2426306 24.8823701,
    -76.2476691 24.8496202, -76.2552267 24.8395433, -76.3005727 24.8622163,
    -76.3433995 24.8748124, -76.3685917 24.8647355, -76.3635533 24.824427999999997,
    -76.3519392 24.768154, -76.3159624 24.7502285, -76.2627844 24.7211399,
    -76.2451498 24.6606786, -76.1922462 24.620371, -76.1116311 24.6027365,
    -76.1040734 24.6531209, -76.1267464 24.7085438, -76.1519386 24.7463321,
    -76.1368233 24.7916781, -76.0965158 24.8571779, -76.1166695 24.9327545,
    -76.101739751545523 25, -76 25, -76 24.5,
    -76.689766972974681 24.5, -76.7137251 24.5196021, -76.74647499999999 24.572505800000002,
    -76.7641095 24.6354863, -76.7715978 24.7372477, -76.7741864 24.7941973,
    -76.7641095 24.8319856, -76.8018979 24.8748124, -76.8396862 24.8521394,
    -76.8447246 24.7967165, -76.8346477 24.7538898, -76.8396862 24.7387744,
    -76.8396862 24.6858708, -76.8375103 24.6003989, -76.795621453661354 24.5)), 
  POLYGON((-75.630983094713983 24.5, -76 24.5, -76 25,
    -75 25, -75 24.5, -75.525446717888414 24.5,
    -75.5265348 24.5034456, -75.5374484 24.5325485, -75.548362 24.5543757,
    -75.5592756 24.5580135, -75.573827 24.5816597, -75.5883785 24.6016679,
    -75.6083867 24.6635116, -75.6120246 24.6762442, -75.6338518 24.7035282,
    -75.6447654 24.7108039, -75.6593168 24.7180796, -75.6702304 24.7180796,
    -75.7011523 24.7217175, -75.7284363 24.7198985, -75.7320742 24.714441700000002,
    -75.7393499 24.7017092, -75.7520824 24.6962524, -75.7666339 24.6926146,
    -75.7957368 24.672606299999998, -75.8248397 24.6325898, -75.8466669 24.6144005,
    -75.8666752 24.599849, -75.8903213 24.5962111, -75.9176053 24.5943922,
    -75.952165 24.5980301, -75.9758111 24.5980301, -75.9721733 24.5634703,
    -75.9539839 24.545281, -75.903291 24.5333831, -75.8666752 24.5343674,
    -75.8339344 24.5416431, -75.81756399999999 24.5580135, -75.8102883 24.5834786,
    -75.8048315 24.5925733, -75.7757285 24.6216762, -75.7575392 24.6289519,
    -75.6956955 24.6034869, -75.6738683 24.5871165, -75.65386 24.5634703,
    -75.65386 24.5143592, -75.6447654 24.5034456, -75.630983094713983 24.5)), 
  POLYGON((-75 24.5, -75 25, -74 25,
    -74 24.5, -75 24.5)), 
  POLYGON((-73 24.95106, -73 24.5, -74 24.5,
    -74 25, -73.371220982142859 25, -73.06450000000001 24.9452,
    -73 24.95106)), 
  POLYGON((-72 24.953681464602532, -72 24.5, -73 24.5,
    -73 24.95106, -72.7283944 24.948867999999997, -72.0071323 24.9537414,
    -72 24.953681464602532)), 
  POLYGON((-71 24.961721524950331, -71 24.5, -72 24.5,
    -72 24.953681464602532, -71.7171655 24.9513047, -71.4789938 24.9527109,
    -71.1953958 24.9609751, -71 24.961721524950331)), 
  POLYGON((-70 24.963172202553679, -70 24.5, -71 24.5,
    -71 24.961721524950331, -70.6215838 24.9631671, -70 24.963172202553679)), 
  POLYGON((-69 24.963180411509061, -69 24.5, -70 24.5,
    -70 24.963172202553679, -69 24.963180411509061)), 
  POLYGON((-68 24.96318862046444, -68 24.5, -69 24.5,
    -69 24.963180411509061, -68 24.96318862046444)), 
  POLYGON((-67 24.963947600250883, -67 24.5, -68 24.5,
    -68 24.96318862046444, -67.2959475 24.9631944, -67 24.963947600250883)), 
  POLYGON((-66 24.963849557661757, -66 24.5, -67 24.5,
    -67 24.963947600250883, -66.9867194 24.9639814, -66 24.963849557661757)), 
  POLYGON((-65 24.963536354217762, -65 24.5, -66 24.5,
    -66 24.963849557661757, -65.4988831 24.963782600000002, -65.1312441 24.9597865,
    -65 24.963536354217762)), 
  POLYGON((-64 24.967817820784632, -64 24.5, -65 24.5,
    -65 24.963536354217762, -64.8515188 24.9677787, -64.3560055 24.9597865,
    -64.0043509 24.9677788, -64 24.967817820784632)), 
  POLYGON((-63 24.97601406733579, -63 24.5, -64 24.5,
    -64 24.967817820784632, -63.113226 24.9757708, -63 24.97601406733579)), 
  POLYGON((-62 24.9771658, -62 24.5, -63 24.5,
    -63 24.97601406733579, -62.4639392 24.9771658, -62 24.9771658)), 
  POLYGON((-61 24.981923157445244, -61 24.5, -62 24.5,
    -62 24.9771658, -61.7145996 24.9771658, -61 24.981923157445244)), 
  POLYGON((-60 24.9821549, -60 24.5, -61 24.5,
    -61 24.981923157445244, -60.9651901 24.9821549, -60 24.9821549)), 
  POLYGON((-59 24.9821549, -59 24.5, -60 24.5,
    -60 24.9821549, -59 24.9821549)), 
  POLYGON((-58 24.97320157119692, -58 24.5, -59 24.5,
    -59 24.9821549, -58.4500803 24.9821549, -58 24.97320157119692)), 
  POLYGON((-57 24.974755877526455, -57 24.5, -58 24.5,
    -58 24.97320157119692, -57.9346208 24.971901, -57.2328373 24.9738081,
    -57 24.974755877526455)), 
  POLYGON((-56 24.98224181136354, -56 24.5, -57 24.5,
    -57 24.974755877526455, -56.3830519 24.9772672, -56 24.98224181136354)), 
  POLYGON((-55 24.98238040717186, -55 24.5, -56 24.5,
    -56 24.98224181136354, -55.9621162 24.9827338, -55.5083803 24.9772672,
    -55 24.98238040717186)), 
  POLYGON((-79.7575859529123 25.5, -79 25.5, -79 25,
    -79.926988529610341 25, -79.8865592 25.0359986, -79.837217 25.2395849,
    -79.7575859529123 25.5)), 
  POLYGON((-78.2083545 25, -79 25, -79 25.5,
    -78 25.5, -78 25.196609311593384, -78.0071767 25.1976682,
    -78.0342662 25.1953126, -78.0436886 25.1941348, -78.053111 25.1976682,
    -78.059 25.1964904, -78.0684225 25.1847124, -78.0719559 25.1847124,
    -78.0825561 25.1823568, -78.0837339 25.181179, -78.0813783 25.1717565,
    -78.0966897 25.1717565, -78.1061122 25.1800012, -78.1261348 25.1964904,
    -78.130845999999991 25.1964904, -78.1390906 25.1976682, -78.1473353 25.2035572,
    -78.1508687 25.210624, -78.1449797 25.2153353, -78.1261348 25.2188687,
    -78.1178902 25.2259355, -78.1049343 25.2341801, -78.0966897 25.2388913,
    -78.0943341 25.2506693, -78.0955119 25.2612696, -78.0943341 25.2671586,
    -78.0919785 25.2730476, -78.0931563 25.2801144, -78.0966897 25.2966037,
    -78.1037565 25.3036705, -78.1061122 25.3142707, -78.1061122 25.3272265,
    -78.113179 25.3342934, -78.1296682 25.3378268, -78.142624 25.3354712,
    -78.1438018 25.336648999999998, -78.142624 25.3178041, -78.1332016 25.3024927,
    -78.1332016 25.2977815, -78.1343794 25.2930702, -78.1320238 25.2895368,
    -78.1320238 25.2836478, -78.1355572 25.2742254, -78.1355572 25.2624474,
    -78.1438018 25.2612696, -78.1492929 25.2411208, -78.1544021 25.2259355,
    -78.1792417 25.226575, -78.202692 25.2282911, -78.2191812 25.2212243,
    -78.2274259 25.204735, -78.2286037 25.1953126, -78.2238925 25.1764678,
    -78.2274259 25.1623341, -78.2274259 25.1552673, -78.2439151 25.1529117,
    -78.2486263 25.1399559, -78.2509819 25.1246444, -78.2309593 25.1258222,
    -78.2132922 25.1281778, -78.208581 25.1140442, -78.2038698 25.0987328,
    -78.196803 25.0822435, -78.1862028 25.0622209, -78.1791359 25.0304202,
    -78.2083545 25)), 
  GEOMETRYCOLLECTION(
  POLYGON((-77.6195158 25, -77.8181018 25, -77.9270861 25.0292424,
    -77.9376863 25.0304202, -77.9412197 25.0421982, -77.9365085 25.0633987,
    -77.9388641 25.0822435, -77.9659536 25.1069774, -77.9565312 25.1187554,
    -77.9565312 25.1364225, -77.9718426 25.1576229, -77.981265 25.1788234,
    -77.9992144 25.1964934, -78 25.196609311593384, -78 25.5,
    -77.8082069306006 25.5, -77.8094294 25.4908315, -77.8064471 25.4729378,
    -77.8004825 25.4505707, -77.7855711 25.4341681, -77.7810977 25.4222389,
    -77.7975002 25.4207478, -77.8094294 25.4192567, -77.8183763 25.4177655,
    -77.8213585 25.4252212, -77.8303054 25.4460972, -77.83627 25.4490795,
    -77.857146 25.4416238, -77.8631105 25.4326769, -77.867584 25.4282035,
    -77.8899511 25.4311858, -77.9078448 25.4341681, -77.9346854 25.4207478,
    -77.9317031 25.3983806, -77.9212651 25.3819781, -77.8810043 25.3700489,
    -77.8139028 25.3596109, -77.7408368 25.3685578, -77.7095228 25.4028541,
    -77.7005759 25.4267124, -77.6886468 25.4520618, -77.6841734 25.4863581,
    -77.679262314536231 25.5, -77 25.5, -77 25,
    -77.2203517 25, -77.1896467 25.0339249, -77.1613982 25.0462068,
    -77.1306933 25.0548042, -77.0963037 25.071998999999998, -77.072968 25.0805964,
    -77.0582296 25.0891938, -77.0434913 25.1076167, -77.0312093 25.1272679,
    -77.0262965 25.146919, -77.0398067 25.1542882, -77.0520886 25.159201,
    -77.0803372 25.1579728, -77.1159549 25.1579728, -77.1429752 25.1506036,
    -77.1749084 25.140777999999997, -77.2031569 25.1358652, -77.2375464 25.1383216,
    -77.2670232 25.1493754, -77.3102692 25.1443136, -77.359138 25.1432344,
    -77.3886147 25.1395498, -77.4377426 25.1272679, -77.4745885 25.1272679,
    -77.5249446 25.1088449, -77.5654751 25.0990193, -77.5998646 25.084280999999997,
    -77.6170594 25.0634016, -77.626885 25.0339249, -77.6195158 25)), 
    POLYGON((-77.782845227378886 25.5, -77.745722880547987 25.5, -77.7497837 25.4893404,
      -77.7617128 25.4624998, -77.7661862 25.4610087, -77.7721508 25.4803935,
      -77.782845227378886 25.5))), 
  GEOMETRYCOLLECTION(
  POLYGON((-76.2149192 25, -77 25, -77 25.5,
    -76.919003573001717 25.5, -76.9152628 25.4794256, -76.8875513 25.4693487,
    -76.8321285 25.4819448, -76.805045565618727 25.5, -76.767608660183384 25.5,
    -76.7666287 25.4970601, -76.7767056 25.4743871, -76.781744 25.461790999999998,
    -76.8018978 25.4315603, -76.8144939 25.4114066, -76.8447246 25.3761375,
    -76.8472438 25.3761375, -76.8699168 25.3862143, -76.8875513 25.3685798,
    -76.9001475 25.3282722, -76.8875513 25.2854455, -76.8396861 25.3005608,
    -76.8119747 25.3181954, -76.7867825 25.3685798, -76.7414365 25.4088873,
    -76.7313596 25.4189642, -76.668379 25.4088873, -76.6230331 25.4038489,
    -76.5474564 25.3433876, -76.5222642 25.3181954, -76.474398999999991 25.3207146,
    -76.4114184 25.3005608, -76.3711109 25.3005608, -76.3660724 25.2753686,
    -76.3459187 25.2375803, -76.293015 25.1997919, -76.2678228 25.1720805,
    -76.2552267 25.1594844, -76.2174383 25.1242153, -76.1720923 25.1242153,
    -76.1670539 25.1015423, -76.1720923 25.0587155, -76.1771308 25.0335233,
    -76.2149192 25)), 
    POLYGON((-76.589960349410077 25.5, -76 25.5, -76 25,
      -76.101739751545523 25, -76.1015537 25.000837999999998, -76.0763619 25.0536771,
      -76.0688043 25.0864269, -76.0688043 25.1040615, -76.0738427 25.1393306,
      -76.088957999999991 25.1796382, -76.1040734 25.1972727, -76.1267464 25.2199457,
      -76.1695731 25.2451379, -76.2350729 25.2753686, -76.2804189 25.3232338,
      -76.3106495 25.3559837, -76.3308033 25.3761375, -76.3559955 25.3836951,
      -76.4114184 25.3912528, -76.4265338 25.3912528, -76.4466875 25.4013297,
      -76.474398999999991 25.4265219, -76.5021104 25.4441564, -76.5524948 25.4517141,
      -76.5852447 25.4693487, -76.589960349410077 25.5))), 
  POLYGON((-76 25, -76 25.5, -75 25.5,
    -75 25, -76 25)), 
  POLYGON((-74 25.389244775286123, -74 25, -75 25,
    -75 25.5, -74.135614264555088 25.5, -74 25.389244775286123)), 
  POLYGON((-73.371220982142859 25, -74 25, -74 25.389244775286123,
    -73.566 25.0348, -73.371220982142859 25)), 
  POLYGON((-79.604693422702439 26, -79 26, -79 25.5,
    -79.7575859529123 25.5, -79.604693422702439 26), 
    (-79.2916421 25.7722913, -79.2918176 25.7721158, -79.3037218 25.7369985, 
      -79.3114595 25.7215231, -79.3114595 25.7036669, -79.3126499 25.6941436, 
      -79.3090786 25.6822394, -79.2953889 25.6768826, -79.2953889 25.6625976, 
      -79.2959841 25.6477174, -79.2876512 25.6268851, -79.2864608 25.6137906, 
      -79.2888416 25.5989104, -79.2959841 25.598315200000002, -79.3132451 25.600696, 
      -79.3281253 25.59772, -79.3287205 25.5899823, -79.3269349 25.5733165, 
      -79.3197924 25.5506986, -79.3096738 25.5459369, -79.2989601 25.5364136, 
      -79.2936032 25.5227238, -79.2846751 25.5161765, -79.2805087 25.5120101, 
      -79.2623926 25.5004688, -79.2406298 25.5090341, -79.2424154 25.5328424, 
      -79.2489627 25.5572459, -79.2578908 25.5715308, -79.2656285 25.5917679, 
      -79.258485999999991 25.6143858, -79.2543195 25.6304564, -79.2382489 25.6560503, 
      -79.2364633 25.6822394, -79.2352729 25.7024765, -79.2388442 25.7119998, 
      -79.2334873 25.7262848, -79.223964 25.7352129, -79.212655 25.7453314, 
      -79.2352729 25.7744966, -79.2507483 25.8006857, -79.2662237 25.7983049, 
      -79.2781278 25.7858056, -79.2916421 25.7722913)), 
  POLYGON((-78 26, -78 25.5, -79 25.5,
    -79 26, -78 26)), 
  POLYGON((-77.8082069306006 25.5, -78 25.5, -78 26,
    -77.5028984 26, -77.4169476 25.9295561, -77.334578 25.9188122,
    -77.2479671 25.8211173, -77.2094528 25.800113, -77.1555137 25.8328614,
    -77.1189343 25.8881499, -77.1125383 26, -77 26,
    -77 25.5, -77.679262314536231 25.5, -77.6707531 25.5236367,
    -77.6692619 25.5564418, -77.6793225 25.5969854, -77.7021324 25.6907972,
    -77.7718896 25.6800653, -77.7933534 25.706895, -77.7879875 25.7712862,
    -77.8523787 25.8303116, -77.9328678 25.883971, -77.9489656 25.8356775,
    -77.9543316 25.7873841, -77.9221359 25.7337247, -77.8845744 25.6854312,
    -77.8631106 25.6425037, -77.8362809 25.615674, -77.7826215 25.5942103,
    -77.7454142 25.5626967, -77.7378545 25.5415304, -77.7378545 25.5206544,
    -77.745722880547987 25.5, -77.782845227378886 25.5, -77.7900445 25.5131987,
    -77.8064471 25.5131987, -77.8082069306006 25.5)), 
  GEOMETRYCOLLECTION(
  POLYGON((-76.919003573001717 25.5, -77 25.5, -77 26,
    -76 26, -76 25.5, -76.589960349410077 25.5,
    -76.5902832 25.5020985, -76.5902832 25.5398869, -76.6104369 25.5625599,
    -76.6280715 25.5776752, -76.6658598 25.5852329, -76.7540326 25.575156,
    -76.7792248 25.575156, -76.8195324 25.5600406, -76.8648783 25.5600406,
    -76.8850321 25.5550022, -76.9026667 25.5449253, -76.9152628 25.5323292,
    -76.9203012 25.507137, -76.919003573001717 25.5)), 
    POLYGON((-76.767608660183384 25.5, -76.805045565618727 25.5, -76.8018978 25.5020985,
      -76.7792248 25.5121754, -76.7716671 25.5121754, -76.767608660183384 25.5))), 
  POLYGON((-76 25.5, -76 26, -75 26,
    -75 25.5, -76 25.5)), 
  POLYGON((-74.135614264555088 25.5, -75 25.5, -75 26,
    -74.7435140265178 26, -74.64070000000001 25.9125, -74.135614264555088 25.5)), 
  POLYGON((-79.45108362340028 26.5, -79 26.5, -79 26,
    -79.604693422702439 26, -79.5667943 26.1239404, -79.45108362340028 26.5)), 
  POLYGON((-78 26, -79 26, -79 26.5,
    -78.7804026910133 26.5, -78.7767684 26.4949317, -78.7426711 26.4839966,
    -78.731946 26.4773572, -78.7268388 26.4753144, -78.7135602 26.4748037,
    -78.695685 26.4763358, -78.6798528 26.4783787, -78.6568705 26.4839966,
    -78.628064474005242 26.5, -78.002360986736193 26.5, -78 26)), 
  GEOMETRYCOLLECTION(
  LINESTRING(-78 26, -77.5028984 26),
    POLYGON((-77 26.2576539, -77 26, -77.1125383 26,
      -77.1533734 26.0324832, -77.1582332 26.0535424, -77.1582332 26.0746015,
      -77.1549933 26.1053802, -77.1549933 26.1329191, -77.1582332 26.1701775,
      -77.1485136 26.2025762, -77.1290744 26.2268751, -77.0934359 26.2382147,
      -77.0561774 26.2446944, -77.0205389 26.254414, -77 26.2576539))), 
  POLYGON((-76.9444021 26.5, -76 26.5, -76 26,
    -77 26, -77 26.2576539, -76.9768007 26.3078718,
    -76.9638413 26.3386505, -76.9606014 26.3645694, -76.9654612 26.3953481,
    -76.9751808 26.4164072, -76.9719409 26.421267, -76.9622213 26.4536657,
    -76.9444021 26.4795846, -76.9444021 26.5)), 
  POLYGON((-75 26.218282693906634, -75 26, -76 26,
    -76 26.5, -75.331022749476631 26.5, -75 26.218282693906634)), 
  POLYGON((-74.7435140265178 26, -75 26, -75 26.218282693906634,
    -74.7435140265178 26)), 
  POLYGON((-79.148623359799416 27, -79 27, -79 26.5,
    -79.45108362340028 26.5, -79.4498547 26.503994, -79.40044 26.59947,
    -79.3473 26.8064, -79.148623359799416 27)), 
  GEOMETRYCOLLECTION(
  POLYGON((-78.7804026910133 26.5, -79 26.5, -79 27,
    -78.2628696 27, -78.2581833 26.9831419, -78.2511539 26.9761125,
    -78.2394382 26.9679116, -78.2312373 26.9679116, -78.2300657 26.9456518,
    -78.2288941 26.9351077, -78.2136638 26.9421371, -78.2054628 26.9491665,
    -78.1937471 26.966739999999998, -78.1925756 26.974940999999998, -78.1832031 26.9843135,
    -78.1632864 27, -78 27, -78 26.752460725069625,
    -78.0408648 26.7402475, -78.057898 26.7402475, -78.0825963 26.7359892,
    -78.088557999999992 26.7376925, -78.0945196 26.7342858, -78.1013329 26.7240659,
    -78.1115529 26.7181042, -78.1132562 26.7121426, -78.1175145 26.7129942,
    -78.1311412 26.7240659, -78.1413611 26.7317308, -78.1456194 26.7368408,
    -78.1583944 26.7376925, -78.1626527 26.7325825, -78.1652077 26.7223625,
    -78.1728727 26.7257692, -78.1779827 26.7232142, -78.1907576 26.7215109,
    -78.2052359 26.7232142, -78.2214175 26.7240659, -78.2282308 26.7266209,
    -78.2350441 26.7376925, -78.2410058 26.7479124, -78.2546324 26.7504674,
    -78.263149 26.7547258, -78.2691107 26.7623907, -78.27847899999999 26.7675007,
    -78.3074355 26.7632424, -78.3099881 26.7495285, -78.3099905 26.7402475,
    -78.3082872 26.7300275, -78.2989189 26.7206592, -78.2938089 26.7129942,
    -78.2972156 26.7070326, -78.3082872 26.7053293, -78.3159522 26.7036259,
    -78.3210621 26.7078843, -78.3368235 26.7081419, -78.3432054 26.7027743,
    -78.3636453 26.7044776, -78.3687553 26.7121426, -78.3687553 26.7138459,
    -78.3900469 26.7121426, -78.3985635 26.7095876, -78.4011185 26.7181042,
    -78.4070802 26.7249175, -78.4164485 26.7266209, -78.4258168 26.7198075,
    -78.4343061 26.7295508, -78.4334817 26.7419508, -78.4351851 26.7470608,
    -78.4360367 26.7555774, -78.442849999999993 26.7606874, -78.450515 26.7632424,
    -78.4547733 26.7649457, -78.4598833 26.7683524, -78.4598833 26.7760173,
    -78.460735 26.7828307, -78.4692516 26.7853856, -78.4820266 26.787088999999998,
    -78.4845815 26.7904956, -78.4743616 26.7887923, -78.4726583 26.8049739,
    -78.4735099 26.8126389, -78.4786199 26.8254138, -78.4845815 26.8322271,
    -78.4845815 26.8398921, -78.4828782 26.8467054, -78.4956532 26.8526671,
    -78.5271647 26.8637387, -78.5373847 26.8552221, -78.5424947 26.8381888,
    -78.5424947 26.8313755, -78.5445445 26.8202738, -78.546753 26.8049739,
    -78.5535663 26.8032706, -78.564638 26.8092322, -78.5629346 26.8151939,
    -78.5731546 26.8194522, -78.5765612 26.8228588, -78.5833746 26.8228588,
    -78.5867812 26.8211555, -78.5927429 26.8168972, -78.5918912 26.7981606,
    -78.5927429 26.791347299999998, -78.6012595 26.7879406, -78.6072211 26.7811273,
    -78.6072211 26.771759, -78.6038145 26.7683524, -78.6012595 26.7615391,
    -78.6080728 26.7564291, -78.6097761 26.7530224, -78.6097761 26.7504674,
    -78.6157378 26.7462091, -78.6234027 26.7470608, -78.6259577 26.7445058,
    -78.6259577 26.7342858, -78.6285127 26.7317308, -78.6302161 26.7325825,
    -78.6376715 26.732204199999998, -78.6387327 26.7317308, -78.642991 26.7274725,
    -78.6412877 26.7164009, -78.6336227 26.7070326, -78.6234027 26.7053293,
    -78.6106278 26.7104393, -78.6063695 26.698515999999998, -78.5987045 26.698515999999998,
    -78.598244999999991 26.6846357, -78.6037731 26.6846357, -78.6066477 26.6841935,
    -78.609079999999992 26.6784443, -78.6101856 26.672473999999998, -78.609079999999992 26.6673882,
    -78.6097434 26.6638503, -78.6161559 26.6552265, -78.6252219 26.6474873,
    -78.6387103 26.6242695, -78.640036999999992 26.6207316, -78.6413638 26.6185204,
    -78.6451228 26.6202893, -78.6570634 26.6273652, -78.6678983 26.6293553,
    -78.6738686 26.6262596, -78.6782911 26.6205105, -78.6869148 26.607021999999997,
    -78.6917795 26.5988405, -78.6997399 26.5968504, -78.70371999999999 26.5922069,
    -78.7081425 26.5877845, -78.7158817 26.5875633, -78.730918 26.5895534,
    -78.7362249 26.5904379, -78.7402051 26.592427999999998, -78.7494922 26.5948604,
    -78.7552414 26.5948604, -78.7592216 26.5926491, -78.7670643 26.5911508,
    -78.7819962 26.5963544, -78.7922106 26.598907999999998, -78.7881249 26.611676,
    -78.7819963 26.6142296, -78.7748462 26.6331262, -78.7748462 26.6484477,
    -78.775357 26.6601942, -78.7845499 26.6709193, -78.8014036 26.6739836,
    -78.8141715 26.6719407, -78.8228537 26.6668335, -78.8305145 26.6607049,
    -78.8400779 26.6603431, -78.8437931 26.6596835, -78.8397074 26.6438512,
    -78.845835999999991 26.6382333, -78.8570718 26.6407869, -78.8545182 26.6448727,
    -78.8591147 26.6530441, -78.8637111 26.6601942, -78.869329 26.6642799,
    -78.8785219 26.6632585, -78.8795434 26.6612156, -78.8841398 26.6637692,
    -78.8866934 26.6601942, -78.8923113 26.6561085, -78.894556999999992 26.6610773,
    -78.8958863 26.6709193, -78.9009935 26.6734729, -78.9045685 26.6765372,
    -78.9050793 26.6826658, -78.9101864 26.6887944, -78.9163151 26.691347999999998,
    -78.9241495 26.6943159, -78.9301044 26.6928801, -78.9403188 26.6939016,
    -78.9398081 26.6857301, -78.9311259 26.6755157, -78.9244865 26.6693871,
    -78.9459367 26.6780693, -78.952576 26.691347999999998, -78.9602368 26.6964552,
    -78.9699404 26.7061588, -78.9811762 26.7107553, -78.98883699999999 26.7143303,
    -78.9954763 26.7071802, -78.9980299 26.6908373, -78.9949656 26.6857301,
    -78.9878155 26.6775586, -78.9796441 26.6734729, -78.9704511 26.662747799999998,
    -78.9209116 26.6208689, -78.9112079 26.6096331, -78.902014999999992 26.5983974,
    -78.8841399 26.5830758, -78.8739255 26.5728615, -78.8611576 26.5626471,
    -78.8407289 26.548347, -78.8208109 26.5371112, -78.8044679 26.524343299999998,
    -78.7901678 26.5136182, -78.7804026910133 26.5),
    (-78.6149176 26.926295999999997, -78.6165357 26.9255714, -78.6174668 26.9226934,
      -78.6176361 26.9205772, -78.6182286 26.9190535, -78.6182286 26.9165141,
      -78.6166203 26.9146518, -78.6142502 26.912450999999997, -78.6117107 26.9121124,
      -78.6083925 26.9104738, -78.6065472 26.9122817, -78.6056161 26.91389,
      -78.6051928 26.916767999999998, -78.6062933 26.9200693, -78.6081555 26.9229473,
      -78.6102717 26.9248096, -78.6149176 26.926295999999997)), 
    POLYGON((-78.002360986736193 26.5, -78.628064474005242 26.5, -78.6017131 26.5146397,
      -78.590988 26.5202576, -78.5715807 26.5274077, -78.5506413 26.5386434,
      -78.5230625 26.5457935, -78.507741 26.5503899, -78.4919087 26.5544757,
      -78.4801622 26.5621365, -78.4556478 26.5718401, -78.4398155 26.5769473,
      -78.4055974 26.5815437, -78.384658 26.5825652, -78.3657615 26.5876723,
      -78.2988574 26.6070797, -78.2876217 26.6091225, -78.2697465 26.610144,
      -78.2559572 26.6091225, -78.2401249 26.6086118, -78.2202069 26.6111654,
      -78.2023318 26.6142297, -78.1762852 26.6203583, -78.1563672 26.6234226,
      -78.1379814 26.6218905, -78.124192 26.6229119, -78.0904846 26.633637,
      -78.0720988 26.6346584, -78.0434986 26.637211999999998, -78.0322628 26.6361906,
      -78.0148984 26.6310834, -78.0029223 26.6188726, -78.002360986736193 26.5))), 
  GEOMETRYCOLLECTION(
  LINESTRING(-77 26.6286184, -77 27),
    LINESTRING(-77 27, -78 27),
    POLYGON((-78 26.752460725069625, -78 27, -77.9971306 26.7533183,
      -78 26.752460725069625))), 
  POLYGON((-77 27, -76 27, -76 26.5,
    -76.9444021 26.5, -76.9444021 26.5022637, -76.9314426 26.5492417,
    -76.9427821 26.5703008, -76.9735609 26.5719208, -76.9768007 26.5865002,
    -76.970321 26.6091792, -77 26.6286184, -77 27)), 
  POLYGON((-75.331022749476631 26.5, -76 26.5, -76 27,
    -75.943167685589529 27, -75.4826 26.629, -75.331022749476631 26.5)), 
  POLYGON((-79 27.14482569479793, -79 27, -79.148623359799416 27,
    -79 27.14482569479793)), 
  POLYGON((-78.2628696 27, -79 27, -79 27.14482569479793,
    -78.635512202256592 27.5, -78 27.5, -78 27.0977216,
    -78.01418 27.1011821, -78.0245615 27.0959913, -78.0358083 27.0847446,
    -78.0496503 27.0899354, -78.0617622 27.0951262, -78.0652227 27.1055077,
    -78.0583017 27.120214999999998, -78.073874 27.1375176, -78.0920418 27.1444386,
    -78.106749 27.1427084, -78.1220093 27.137013999999997, -78.1236167 27.1288088,
    -78.1093444 27.1106985, -78.0790648 27.087339999999998, -78.0807951 27.0769584,
    -78.0695484 27.0639815, -78.0418642 27.0518696, -78 27.0657117,
    -78 27, -78.1632864 27, -78.1609433 27.0229752,
    -78.1586002 27.0323477, -78.1597717 27.0346909, -78.1730738 27.0421534,
    -78.1960903 27.0300046, -78.2253794 27.0229752, -78.2394382 27.0182889,
    -78.2628696 27), 
    (-78.324761999999993 27.25258, -78.34465999999999 27.2439287, -78.3524462 27.2335471, 
      -78.3541765 27.2205702, -78.3515811 27.2110537, -78.3342785 27.2041327, 
      -78.3135153 27.1989419, -78.2962127 27.1989419, -78.2884265 27.1954814, 
      -78.284966 27.1859649, -78.2789101 27.1781787, -78.2711239 27.1712577, 
      -78.2461173 27.1821265, -78.232193 27.2041327, -78.2538213 27.2075932, 
      -78.2841009 27.2153794, -78.2927522 27.2171096, -78.3091897 27.219704999999998, 
      -78.3057291 27.225761, -78.3014035 27.2387379, -78.324761999999993 27.25258)), 
  POLYGON((-77 27, -78 27, -78 27.0657117,
    -77.9839004 27.067442, -77.9769793 27.0700374, -77.9665978 27.0726328,
    -77.9690895 27.0885772, -77.9778445 27.0951262, -78 27.0977216,
    -78 27.5, -77 27.5, -77 27), 
    (-77.6922701 27.0584454, -77.7374184 27.0729573, -77.7666727 27.0651255, 
      -77.7565373 27.0093812, -77.7178388 27.0229717, -77.6876632 27.0206682, 
      -77.669696 27.0443941, -77.6922701 27.0584454)), 
  POLYGON((-76 27.045779999999997, -76 27, -77 27,
    -77 27.5, -76 27.5, -76 27.045779999999997)), 
  GEOMETRYCOLLECTION(
  POLYGON((-75.943167685589529 27, -76 27, -76 27.045779999999997,
    -75.943167685589529 27)), 
    POLYGON((-76 27.045779999999997, -76 27.5, -75.9992071582174 27.5,
      -76 27.045779999999997))), 
  POLYGON((-78 27.74167, -78 27.5, -78.635512202256592 27.5,
    -78.365000000000009 27.7636, -78 27.74167)), 
  POLYGON((-77 27.7349763, -77 27.5, -78 27.5,
    -78 27.74167, -77.6368166 27.7349763, -77 27.7349763)), 
  POLYGON((-76 27.7299386, -76 27.5, -77 27.5,
    -77 27.7349763, -76.9086666 27.7349763, -76.4313532 27.7299386,
    -76 27.7299386)), 
  POLYGON((-75.9992071582174 27.5, -76 27.5, -76 27.7299386,
    -75.9988058 27.7299386, -75.9992071582174 27.5)))";
            var geom = new WKTReader().Read(WKT);
            Geometry buffer0 = null;
            Assert.DoesNotThrow(() => buffer0 = geom.Buffer(0));
            Assert.NotNull(buffer0);
            Assert.That(buffer0.IsValid, Is.True);
        }

        #region utility

        static void ToImage(int nr, Geometry geom1, Geometry geom2, Geometry geom3)
        {

            //var gpw = new Windows.Forms.GraphicsPathWriter();

            //var extent = new Envelope();
            //if (geom1 != null)
            //    extent.ExpandToInclude(geom1.EnvelopeInternal);
            //if (geom2 != null)
            //    extent.ExpandToInclude(geom2.EnvelopeInternal);
            //if (geom3 != null)
            //    extent.ExpandToInclude(geom3.EnvelopeInternal);

            //extent.ExpandBy(0.05 * extent.Width);

            //using (var img = new Bitmap(ImageWidth, ImageHeight))
            //{
            //    using (var gr = Graphics.FromImage(img))
            //    {
            //        var at = CreateAffineTransformation(extent);
            //        gr.Clear(Color.WhiteSmoke);
            //        gr.SmoothingMode = SmoothingMode.AntiAlias;
            //        //gr.Transform = CreateTransform(extent);

            //        if (geom1 != null)
            //        {
            //            var gp1 = gpw.ToShape(at.Transform(geom1));
            //            if (geom1 is IPolygonal)
            //                gr.FillPath(new SolidBrush(Color.FromArgb(64, Color.Blue)), gp1);
            //            gr.DrawPath(Pens.Blue, gp1);
            //        }

            //        if (geom2 != null)
            //        {
            //            var gp2 = gpw.ToShape(at.Transform(geom2));
            //            if (geom2 is IPolygonal)
            //                gr.FillPath(new SolidBrush(Color.FromArgb(64, Color.OrangeRed)), gp2);
            //            gr.DrawPath(Pens.OrangeRed, gp2);
            //        }

            //        if (geom3 != null)
            //        {
            //            var gp3 = gpw.ToShape(at.Transform(geom3));
            //            if (geom3 is IPolygonal)
            //                gr.FillPath(new SolidBrush(Color.FromArgb(64, Color.Gold)), gp3);
            //            gr.DrawPath(Pens.Gold, gp3);
            //        }

            //    }
            //    var path = System.IO.Path.ChangeExtension(System.IO.Path.GetTempFileName(), "png");
            //    img.Save(path, ImageFormat.Png);
            //    Console.WriteLine("Image for Test {0} written to {1}", nr, new Uri(path).AbsoluteUri);
            //}
        }

        private const int ImageWidth = 640;
        private const int ImageHeight = 480;

        private static AffineTransformation CreateAffineTransformation(Envelope env, int offsetX = 0)
        {
            int imageRatio = ImageWidth / ImageHeight;
            double ratio = env.Width / env.Height;
            if (ratio > imageRatio)
            {
                double growHeight = (env.Width / imageRatio - env.Height) / 2;
                env.ExpandBy(0, growHeight);
            }
            else if (ratio < imageRatio)
            {
                double growWidth = (env.Height * imageRatio - env.Width) / 2;
                env.ExpandBy(growWidth, 0);
            }

            var s1 = new Coordinate(env.MinX, env.MaxY);
            var t1 = new Coordinate(offsetX, 0);
            var s2 = new Coordinate(env.MaxX, env.MaxY);
            var t2 = new Coordinate(offsetX + ImageWidth, 0);
            var s3 = new Coordinate(env.MaxX, env.MinY);
            var t3 = new Coordinate(offsetX + ImageWidth, ImageHeight);

            var atb = new AffineTransformationBuilder(s1, s2, s3, t1, t2, t3);
            return atb.GetTransformation();
        }

        #endregion
    }
}
